#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of Nvidia Corporation
# and its affiliates (the "Company") and all right, title, and interest
# in and to the software product, including all associated intellectual
# property rights, are and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.


# HACK: This module will be exposed on ctx, so don't clutter the "public"
#       namespace.
from cue import exceptions as _exceptions
from connexion import ProblemException as _ProblemException


class _BaseProblem(_ProblemException):
    """
    The base class for the error classes contained here.

    It should not be created directly.
    """

    def __init__(self, **kwargs):
        # 'detail' can (should) change depending on how/where things went
        # wrong. If the caller doesn't set it though, use the title.
        kwargs.setdefault('detail', self.title)
        super().__init__(
            # These properties should not be influenced by instance.
            status=self.code,
            type=self.type_uri,
            title=self.title,
            **kwargs,
        )

    @property
    def type_uri(self):
        return "/errors/" + type(self).__name__


class ItemDoesNotExist(_BaseProblem, _exceptions.NotFound):
    """
    The requested item does not exist right now.

    The URL is valid, so the item may exist later, or it may be created.
    The resource may also exist elswhere. For example, if an interface were
    removed on a 'pending' revision, that interface would still exist on the
    'applied' revision.
    """

    title = "The requested item does not exist."


class OperationalNotSupported(_BaseProblem, _exceptions.NotFound):
    """
    Showing 'operational' is not supported for this resource.

    The URL is valid. The resource is likely configurable via NVUE. But showing
    the 'operational' state is not supported. Future updates to NVUE may add
    support for this resource's 'operational' state.
    """

    title = "Showing 'operational' is not supported for this resource."
