# Copyright (C) 2018, 2019 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


# Job queue
# - Each job is a closure and has a name (for logging)
# - A single daemon thread will run each job in turn.


from enum import IntEnum
from queue import PriorityQueue
from dataclasses import dataclass, field
from threading import Thread, get_ident
import logging


logger = logging.getLogger(__name__)


def __getattr__(name):
    # HACK: if we don't find the attribute in the module, check the
    #       _threaded_job_queue
    return getattr(_threaded_job_queue, name)


class Priority(IntEnum):
    """
    Priority levels for Jobs. A lower number means a higher priority.
    """
    IMMEDIATE = 0
    NORMAL = 10

    # NOTE: Should exhaust the queue before stopping.
    _STOP = 9999


@dataclass(order=True)
class Job:
    """
    A queued Job. To work with the PriorityQueue, it defines its comparison
    methods in terms of its `priority` attribute.
    """

    fn: callable = field(compare=False)
    name: str = field(compare=False)
    priority: Priority = Priority.NORMAL


class ThreadedJobQueue:
    """
    NVUE's job queue.

    At runtime, there's only one instance of this class. Why is it a class
    then? So we can subclass it for unit tests!
    """

    _stop_job = Job(
        fn=object(),
        name="STOP",
        priority=Priority._STOP,
    )
    """Special job that tells the thread to exit."""

    def __init__(self):
        self._queue = None
        self._job_thread = None

    def __next__(self):
        # Exposed for overriding in unit tests
        return self._queue.get()

    def job_main(self):
        """
        Main loop for the job thread.  It blocks on the queue until
        it gets a job, then it runs it.
        """
        logger.info("Started the Job thread")

        while True:
            job = next(self)
            if job is self._stop_job:
                logger.info("Stopped the Job thread")
                return

            logger.info("Running Job %s", job.name)
            job.fn()
            logger.info("Ran Job %s", job.name)

    def _init_thread(self):
        # Exposed for overriding in unit tests
        # Create and start the thread
        self._job_thread = Thread(target=self.job_main, daemon=True)
        self._job_thread.start()

    def _join_thread(self):
        # Exposed for overriding in unit tests
        # Wait for the thread to die
        self._job_thread.join()

    def start(self):
        """
        Start the job thread.
        """
        if self._queue is not None:
            logger.debug("Job thread already started")
            return

        # Create the queue
        self._queue = PriorityQueue()
        self._init_thread()

    def stop(self):
        """
        Stop the job thread.
        """
        # Add the stopper
        self._queue.put(self._stop_job)

        self._join_thread()

        # And cleanup
        self._queue = None
        self._job_thread = None

    def _job_priority(self):
        """Determine the job priority based on which thread we're in."""
        # Exposed for overriding in unit tests
        if get_ident() == self._job_thread.ident:
            # Prioritize jobs that are added by another job.
            return Priority.IMMEDIATE
        else:
            # We're in the main thread. Give it the normal priority.
            return Priority.NORMAL

    def add(self, name, fn):
        """
        Add a new job.

        When the job is run, fn() will be called with no arguments.
        name identifies the job for logging purposes.
        """
        priority = self._job_priority()

        job = Job(
            fn=fn,
            name=name,
            priority=priority,
        )
        logger.info("Queued Job %s", job.name)
        self._queue.put(job)


_threaded_job_queue = ThreadedJobQueue()


# Expose the public API on the module itself.
start = _threaded_job_queue.start
stop = _threaded_job_queue.stop
add = _threaded_job_queue.add
