def format_params_and_convert_to_list_path(openapi_path: str, params: dict):
    """
    Takes an OpenAPI path and a dictionary of path parameters and their
    associated values and attempts to replace the parameters in the path
    with the values from the dictionary and then turn the path into a
    list of tokens.

    Arguments:
        openapi_path -- OpenAPI path of the form /level1/level2/{param1}
        params -- dictionary of parameters that could appear in the
            path and the values to replace them with.
            Example: {param1: value1}

    Returns:
        A list of path tokens.  Given the above example, it would return
        ['level1', 'level2', 'value1']
    """
    list_path = convert_to_list_path(openapi_path)
    formatted_list_path = format_list_path_params(list_path, params)
    return formatted_list_path


def format_list_path_params(list_path: list, path_params: dict):
    """
    Given an OpenAPI path represented as a list and a dictionary of
    path parameters and associated values, replace the path parameters in the
    list with the values provided in `path_params`.  All parameters in
    `path_params` need not be in the `list_path`; this function will replace
    whatever parameters exist in the path and have an entry in `path_params`.

    Arguments:
        list_path -- OpenAPI path represented as a list of tokens
        path_params -- dictionary of parameters that could appear in the
            path and the values to replace them with.
            Example: {param1: value1}

    Returns:
        An openapi list path with the path parameters converted to the
        corresponding values.

        Raises a FormatOpenAPIPathError if a token in the given Openapi list
        path does not have an entry in `path_params`.
    """
    path_params = {
        # Convert python kwargs to path parameter ids.
        key.replace("_", "-"): value
        for key, value in path_params.items()
    }
    try:
        return [segment.format(**path_params) for segment in list_path]
    except KeyError as e:
        bad_param = e.args[0]
        raise FormatOpenAPIPathError(
            bad_param,
            list_path,
            path_params
        )


class FormatOpenAPIPathError(KeyError):
    """
    Exception class for errors in formatting the OpenapiPath due to path
    parameters that haven't been accounted for in the path_params provided.
    """
    def __init__(self, bad_param, list_path, path_params):
        self.bad_param = bad_param
        self.list_path = list_path
        self.path_params = path_params
        self.msg = self._create_exception_msg()

    def __str__(self):  # pragma: no cover
        return self.msg

    def _create_exception_msg(self):
        return (f"'{self.bad_param}' appears in the path {self.list_path} but "
                "could not be found in the given path params: "
                f"{self.path_params}")


def convert_to_list_path(path: str):
    """
    Convert a str path containing '/'.
    """
    path = path.strip('/')
    list_path = path.split('/')
    return list_path
