#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of Nvidia Corporation
# and its affiliates (the "Company") and all right, title, and interest
# in and to the software product, including all associated intellectual
# property rights, are and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
from __future__ import annotations
from cue.utils import fast_merge
from fnmatch import fnmatch
from .json_pointers import convert_json_pointer_to_path


FILTER_PARAM_NAMES = ["include", "omit"]


def filter_response(data, *, include=None, omit=None):
    """
    Filter response data based off filter query param values.

    Args:
        data: The response data to filter
        include: (optional) A list of string JSON pointer patterns
        omit: (optional) A list of string JSON pointer patterns

    Returns:
        A copy of the filtered response data.
    """
    if include is not None:
        data = handle_include(data, include)
    if omit is not None:
        data = handle_omit(data, omit)
    return data


def is_path_match(path: tuple[str, ...], pattern: str) -> bool:
    """
    Does the key path match the JSON Pointer pattern?
    """
    pattern_segments = convert_json_pointer_to_path(pattern)
    if len(pattern_segments) > len(path):
        return False

    for path_token, pattern_token in zip(path, pattern_segments):
        if not fnmatch(path_token, pattern_token):
            return False
    return True


def handle_include(data: dict, include: list[str]) -> dict:

    def _filter_include(path, value):
        is_primitive = value == {} or not hasattr(value, "items")
        if not is_primitive:
            return True
        return any(
            is_path_match(path, pattern)
            for pattern in include
        )

    return fast_merge(data, filter_function=_filter_include)


def handle_omit(data: dict, omit: list[str]) -> dict:
    # Track the parents of matches to make sure we don't drop them. We'll merge
    # this in with the filtered data before returning.
    parents_of_matches = {}

    def _add_match_parent(path):
        cur = parents_of_matches
        for token in path:
            cur = cur.setdefault(token, {})

    def _filter_omit(path, value):
        for pattern in omit:
            if is_path_match(path, pattern):
                _add_match_parent(path[:-1])
                return False
        return True

    data = fast_merge(data, filter_function=_filter_omit)
    return fast_merge(data, parents_of_matches)
