#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of Nvidia Corporation
# and its affiliates (the "Company") and all right, title, and interest
# in and to the software product, including all associated intellectual
# property rights, are and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.

import itertools


__all__ = [
    "sorted_dict",
    "sorted_dict_deep",
    "natural_sort_key",
]


def sorted_dict(obj, **kwargs):
    """
    Return a shallow copy of obj, sorted by keys.

    Any keyword arguments will be used in the call to `sorted`.
    """
    return {key: obj[key] for key in sorted(obj, **kwargs)}


def sorted_dict_deep(obj, **kwargs):
    """
    Return a copy of obj, sorted by keys and recursively sorting all child
    dicts by keys.

    Assumes everything is either a dict or a primitive.

    Any keyword arguments will be used in the call to `sorted`.
    """
    if isinstance(obj, dict):
        return {
            key: sorted_dict_deep(obj[key], **kwargs)
            for key in sorted(obj.keys(), **kwargs)
        }
    else:
        return obj


def natural_sort_key(value: str):
    """
    A number-aware sort key for use as a key in functions like `sorted`.

    Split the string into numeric and non-numeric chunks, treating the numeric
    chunks as numbers and leaving the non-numeric chunks as strings.

    For more info on the basic concept:

    - http://nedbatchelder.com/blog/200712/human_sorting.html
    - https://en.wikipedia.org/wiki/Natural_sort_order
    """
    # HACK: Guarantee that the first item in tuple stays a string. Otherwise
    #       comparison between 'foo1' and '2bar' will end up comparing 'foo'
    #       and 2, generating a TypeError.
    value = " " + value
    return tuple(
        (
            int(''.join(group_chars))
            if isdigit
            else ''.join(group_chars)
        )
        for isdigit, group_chars in itertools.groupby(value, key=str.isdigit)
    )
