# Copyright (C) 2018-2021 Cumulus Networks, inc
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import annotations
from cue import utils, render, issues, exceptions
from cue.events import AbstractEvent
import logging
import re
from datetime import datetime, timedelta, timezone
from hashlib import md5
from cue_config_v1 import repo
import time
import json
from os.path import basename
from textwrap import indent
from cue_config_v1 import schema
from threading import Condition
from pydash import py_


logger = logging.getLogger(__name__)

# MIGRATION: This and supporting code can be removed after 2021/09/01.  It's
# helping us migrate through the name change.
CONFIG_FILE_HEADER_CUE = ("{comment_char} Auto-generated by CUE!\n"
                          "{comment_char} Any local modifications will"
                          " prevent CUE from re-generating this file.\n"
                          "{comment_char} md5sum: {md5sum}\n")

CONFIG_FILE_HEADER_NVUE = ("{comment_char} Auto-generated by NVUE!\n"
                           "{comment_char} Any local modifications will"
                           " prevent NVUE from re-generating this file.\n"
                           "{comment_char} md5sum: {md5sum}\n")

LOG_FILE = "apply_log.txt"
"""
Name of the log file in the repo we'll append to while applying.
"""

RELOAD_JOURNAL = "reload.yml"
"""
Name of the file in the repo where we'll journal the list of reload operations.
"""

CHECKING_TRIES = 3
"""
Number of times we'll wait for warnings to clear while checking.
"""

CHECKING_WAIT_SEC = 2
"""
Number of seconds to wait between checks while waiting on warnings to clear.
"""

AYS_WAIT_SEC = 30
"""
Number of seconds to wait for a user to respond to an AYS.
"""

IGNORE_FAIL_WAIT_SEC = 30
"""
Number of seconds to wait for a user to respond to ignore_fail prompt.
"""


VALID_TRANSITIONS = {
    "apply": ["verifying"],
    "verifying": ["verified", "invalid", "verify_error"],
    "verified": ["readying"],
    "readying": ["ays", "ready", "ready_error"],
    "ays": ["ays_fail", "ays_no", "ays_yes"],
    "ays_yes": ["ready"],
    "ays_no": ["ays_fail"],
    "ready": ["reloading"],
    "reloading": ["reloaded"],
    "reloaded": ["apply_error", "checking"],
    "checking": ["ignore_fail", "checked"],
    "checked": ["confirm", "applied"],
    "ignore_fail": ["apply_fail", "ignore_fail_no", "ignore_fail_yes"],
    "ignore_fail_no": ["apply_fail"],
    "ignore_fail_yes": ["checked"],
    "confirm": ["confirm_fail", "confirm_no", "confirm_yes"],
    "confirm_no": ["confirm_fail"],
    "confirm_yes": ["applied"],
}


APPLY_NAMESPACE = "apply"
"""
Namespace for apply branches/tags.
"""


###############################
# lifecycle event handlers
###############################

def handle_init_revision_transitions(evt):
    evt.add_entry_state("apply")
    evt.update_transitions(VALID_TRANSITIONS)


def handle_cleanup_revisions(evt):
    revisions = evt.versions_v1.getRevisions()
    for rev_id, rev_data in revisions.items():
        # EARLY CONTINUE
        if rev_data["state"] not in VALID_TRANSITIONS:
            # It's in a terminal/unknown state, so we won't touch it.
            continue
        # Put it in a terminal state, so we can still use it.
        evt.versions_v1.updateRevision(rev_id, state="apply_interrupted")


def handle_revision_state(evt):
    """
    Handle transitions to the "apply" state.  That's our cue to call start an
    apply_config() job.
    """
    if evt.new_state == "apply":
        try:
            user = evt.ctx.request.get_username()
        except RuntimeError:
            # This event is not coming from a user's request.
            user = None
        start_apply(
            evt.ctx,
            evt.rev_id,
            evt.progress,
            evt.get_revman,
            evt.auto_prompt,
            evt.state_controls,
            user,
        )
    elif evt.new_state in (
        "ays_yes",
        "ays_no",
        "ignore_fail_yes",
        "ignore_fail_no",
        "confirm_yes",
        "confirm_no",
    ):
        send_client_state_change(evt.rev_id, evt.new_state)


###############################
# FSM implementation
###############################

class WillOverwriteWarning(issues.WarningIssue):
    """
    The file has been locally edited and we will clobber the changes.
    """

    def __init__(self, filename):
        super().__init__("overwriting")
        self.data["filename"] = filename

    def _render_message(self):
        filename = self.data["filename"]
        return (f'{filename} has been manually changed since '
                'the last save.  These changes WILL be overwritten.')


class WillNotOverwriteWarning(issues.WarningIssue):
    """
    The file has been locally edited and we will not clobber the changes.
    """

    def __init__(self, filename):
        super().__init__("not_overwriting")
        self.data["filename"] = filename

    def _render_message(self):
        filename = self.data["filename"]
        return (f'{filename} has been manually changed since '
                'the last save.  These changes WILL NOT be overwritten.')


class UpgradedWarning(issues.WarningIssue):
    """
    The system has been upgraded (or downgraded) since the config was built.
    """

    def __init__(self):
        super().__init__("upgraded")

    def _render_message(self):
        return (
            "This config was originally built with a different version.  "
            "Applying it now may have different results."
        )


class HardwareChanged(issues.WarningIssue):
    """
    The hardware has changed materially since the config was built.
    """

    def __init__(self):
        super().__init__("changed")

    def _render_message(self):
        return (
            "This config was originally built on different hardware.  "
            "Applying it now may have different results."
        )


class ServiceFailed(issues.InternalError):
    """
    A service is in a failed state.  We expected this to work because we
    delivered the systemd unit and declared the config as valid.  Therefore,
    it's an internal error.
    """

    def __init__(self, systemd_unit):
        super().__init__("service_failed")
        self.data["unit"] = systemd_unit

    def _render_message(self):
        fmt = "{unit} service has failed"
        return fmt.format(**self.data)


class ServiceUnexpected(issues.WarningIssue):
    """
    A service is in an unexpected state.  This might be temporary while it
    transitions to the expected state, so this is a warning.
    """

    def __init__(self, systemd_unit, expected, observed):
        super().__init__("service_unexpected")
        self.data["unit"] = systemd_unit
        self.data["observed"] = observed
        self.data["expected"] = expected

    def _render_message(self):
        fmt = "{unit} service is {observed} (expected {expected})"
        return fmt.format(**self.data)


class SystemctlError(issues.InternalError):
    """
    We couldn't perform a systemctl operation.  We expected this to work
    because we delivered the systemd unit and declared the config as valid.
    Therefore, it's an internal error.
    """
    ignore_warnings = [
        ("Warning: The unit file, source configuration file or drop-ins "
         "of ntp@mgmt.service changed on disk. Run "
         "'systemctl daemon-reload' to reload units.\n")
    ]

    def __init__(self, action, systemd_units, reason):
        super().__init__("systemctl")
        self.data["action"] = action
        self.data["units"] = ",".join(systemd_units)
        for w in self.ignore_warnings:
            reason = reason.replace(w, '')
        self.data["reason"] = " ".join(reason.splitlines())
        self.data["indented_reason"] = indent(reason, "  ")

    def _render_message(self):
        fmt = "Unable to {action} services ({units})"
        if self.data["indented_reason"]:
            fmt = fmt + ":\n{indented_reason}"
        return fmt.format(**self.data)


class JournalctlError(issues.InternalError):
    """
    We couldn't perform a systemctl operation and these messages were generated
    in systemd's journal.
    """
    def __init__(self, action, systemd_unit, errors):
        super().__init__("journalctl")
        self.data["action"] = action
        self.data["unit"] = systemd_unit
        self.data["errors"] = errors
        self.data["indented_errors"] = indent("\n".join(errors), "  ")

    def _render_message(self):
        fmt = "During {action} of {unit}:\n{indented_errors}"
        return fmt.format(**self.data)


class RunScriptError(issues.InternalError):
    """
    We couldn't run an apply script.  We expected this to work because we
    delivered the script and declared the config as valid.  Therefore, it's an
    internal error.
    """

    def __init__(self, script_name, reason):
        super().__init__("run_script")
        self.data["script"] = script_name
        self.data["reason"] = reason

    def _render_message(self):
        fmt = "Unable to run '{script}' script: {reason}"
        return fmt.format(**self.data)


class InstallFileError(issues.InternalError):
    """
    We couldn't run a command.  We expected this to work.  Therefore, it's an
    internal error.
    """

    def __init__(self, dst, reason):
        super().__init__("install_file")
        self.data["dst"] = dst
        self.data["reason"] = reason

    def _render_message(self):
        fmt = "Install of '{dst}' failed: {reason}"
        return fmt.format(**self.data)


class RemoveFileError(issues.InternalError):
    """
    We couldn't run a command.  We expected this to work.  Therefore, it's an
    internal error.
    """

    def __init__(self, target, reason):
        super().__init__("remove_file")
        self.data["target"] = target
        self.data["reason"] = reason

    def _render_message(self):
        fmt = "Remove of {target!r} failed: {reason}"
        return fmt.format(**self.data)


class RollbackWarning(issues.WarningIssue):
    """
    Notify the client that we're rolling back to another revision.
    """

    def __init__(self, rollback_target):
        super().__init__("rollback")
        self.data["rollback_target"] = rollback_target

    def _render_message(self):
        fmt = "Rolling back to {rollback_target}"
        return fmt.format(**self.data)


class RollbackFailed(issues.InternalError):
    """
    The rollback failed to apply.

    NVUE won't attempt to rollback again automatically.
    """

    def __init__(self):
        super().__init__("rollback_failed")

    def _render_message(self):
        return "Failed to apply rollback."


class ConfirmDurationInfo(issues.InfoIssue):
    """
    Tell the user when we'll be rolling back (unless they confirm.)
    """

    def __init__(self, end_time):
        super().__init__("confirm_duration")
        self.data["end_time"] = end_time

    def _render_message(self):
        return (
            f"Configuration will be rolled back at {self.data['end_time']}"
        )


class VerifyConfigEvent(AbstractEvent):
    """
    Event object for a verify config operation.  This will be passed with the
    verify_config event.
    """

    def __init__(self, ctx, rev_id, carry_forwards):
        """
        Create a verify config event object for the given global context and
        rev_id.  We'll take what we need from the global context and
        effectively create a streamlined version for verify_config event
        handlers.

        For example, calls to self.config_v1.getFoo(*args) will be proxied to
        ctx.config_v1.getFoo(rev_id, *args).
        """
        self.fileops = ctx.fileops
        self.sh = ctx.sh
        self._ctx = ctx
        self._carry_forwards = carry_forwards

        self.carry_forward = None
        """
        Allows an FUnit to pass data ahead to the `ready_apply` event handler.

        The FUnit cannot assume the `ready_apply` event handler will be called.
        This should be used carefully and primarily in tricky cases where
        efficiency is a concern.  Generally, it's best to avoid maintaining
        state between `verify_config` and `ready_apply` when it's not
        necessary.
        """

        self.config_v1 = utils.PartialObject(ctx.config_v1, rev_id)
        """
        Provides access to the configuration to verify.  Each `operationId`
        in `cue_config_v1/openapi.yaml` is a method on this object.
        """

        self.platform_v1 = ctx.platform_v1
        """
        Provides access to platform data.
        """

        self.issues = []
        """
        List of issues identified by FUnits.
        """

    def report_issue(self, issue):
        """
        Report an issue in the configuration that must be addressed.  The issue
        can be either a simple string or an Issue instance.  If issue is a
        string, it will be converted into a GenericError issue, otherwise issue
        must already be an Issue instance.
        """
        # Normalize into an Issue.
        if isinstance(issue, str):
            issue = issues.GenericError(issue)

        self.issues.append(issue)

    def calling_handler(self, handler_name):
        """
        Before calling a handler, bubble up the handler's cache.
        """
        # Start the context manager as usual.
        stack = super().calling_handler(handler_name)

        # Set this handler's "carry forward" object.
        self.carry_forward = self._carry_forwards.setdefault(handler_name, {})

        return stack


class ReadyApplyEvent(AbstractEvent):
    """
    Event object for a ready apply operation.  This will be passed with the
    ready_apply event.
    """

    def __init__(self, ctx, revman, rev_id, carry_forwards, warnings=None):
        """
        Create a ready apply event for the given global context and rev_id.
        During file operations, the revman will be used as a scratch area.

        This class provides a streamlined API that FUnits can use during the
        `ready_apply` event to interrogate the desired configuration, generate
        their config files, and stage and schedule everything for the apply.

        Staged and scheduled operations will be performed by the CUE core after
        all FUnits are ready.
        """
        self.config_v1 = utils.PartialObject(ctx.config_v1, rev_id)
        self.platform_v1 = ctx.platform_v1
        self.fileops = ctx.fileops
        self._carry_forwards = carry_forwards

        self.carry_forward = None
        """
        Allows an FUnit to pass data ahead from the `verify_config` event
        handler.
        """

        self.sh = ctx.sh
        """
        Provides access to the configuration to ready.  Each `operationId` in
        `cue_config_v1/openapi.yaml` is a method on this object.
        """

        self._revman = revman
        self._ctx = ctx
        self._header_cue = CONFIG_FILE_HEADER_CUE
        self._header = CONFIG_FILE_HEADER_NVUE
        self.warnings = warnings or []
        """
        List of warnings reported by FUnits.
        """

        # Get a list of files that were staged in an earlier run.
        with revman:
            self.stage_orphans = [
                ent.name for ent in revman.list_files()
                if ent.name.endswith(".dst")
            ]

    def _log(self, msg):
        """
        Log a message.  This will append to LOG_FILE.
        """
        logger.dump("ReadyApplyEvent", msg)
        self._revman.append_file(LOG_FILE, [msg, "\n"])

    @property
    def _operational_hostname(self):
        # HACK: This is only needed so env_v1 can warn the user that the
        #       hostname is about to change. We ought to get rid of this
        #       property and make a generic mechanism for warning users about
        #       when an important property is about to change.
        return self._ctx.env_v1.getHostname()

    def calling_handler(self, handler_name):
        """
        Before calling a handler, enter the revision manager and use it as a
        scratch area.  As we ready for the apply, we'll save our work there.

        In practice, each handler's work will show up as a commit on a side
        branch off the main applied branch.  The side branch will be kept
        around for debugging purposes.
        """
        # Start the context manager as usual.
        stack = super().calling_handler(handler_name)

        # Set this handler's "carry forward" object.
        self.carry_forward = self._carry_forwards.setdefault(handler_name, {})

        # Enter the revision manager and start a commit
        stack.enter_context(self._revman)
        self._revman.start_write("Readying " + handler_name)
        self._log("## Readying " + handler_name)

        return stack

    def report_warning(self, issue):
        """
        Report a warning about the configuration.  The issue can be either a
        simple string or an Issue instance.  If issue is a string, it will be
        converted into a GenericWarning issue, otherwise issue must already be
        an Issue instance.
        """
        # Normalize into an Issue.
        if isinstance(issue, str):
            issue = issues.GenericWarning(issue)

        self._log("Warning: " + issue.message)
        self.warnings.append(issue)

    def render(self, pkg, filename=None, template_name=None, **kwargs):
        """
        Render a template, `template_name` and optionally save the result into
        scratch as `filename`.  The template will be loaded from `pkg`.  If
        `template_name` is not provided, it will be assumed to the same as
        `filename`.  If `filename` is not provided, the result will be returned
        instead of saved.  Either `filename` or `template_name` must be
        provided.

        All kwargs are passed into the template.

        Render the template and add it to the scratch space under the same
        name.
        ```
        evt.render(templates, "some.conf", input=obj)
        ```

        Render the template and add it to the scratch space under a
        different filename.
        ```
        evt.render(templates, "scratch_file", "some.conf", input=obj)
        ```

        Render the template and return the output.  Post-process, then save.
        ```
        out = evt.render(templates, template_name="some.conf", input=obj)
        # post-process
        evt.save("scratch_file", out)
        ```
        """
        if template_name is None and filename is None:
            raise ValueError("Either template_name or filename is required")

        # By default, take the template_name from filename.
        if template_name is None:
            template_name = filename

        # Strip the .j2 extension, if any.
        template_name = utils.removesuffix(template_name, '.j2')

        self._log(f"Rendering template '{template_name}' into '{filename}'")

        # Start the render.  out will be an interable stream of strings.
        out = render.render_template(pkg, template_name, **kwargs)

        # Stream it to the scratch area
        if filename is not None:
            filename = utils.removesuffix(filename, '.j2')
            self.save(filename, out)
        else:
            return "".join(out)

    def save(self, name, contents):
        """
        Save a file into scratch given its contents.  It will be saved using
        the name and can be staged and installed later using the same name.

        contents can be a single object or an iterable of objects.  If an
        iterable, each object will be written in turn.  All objects are
        stringified before writing.
        """
        self._log("Saving: " + name)

        # Stream it to the scratch area
        self._revman.save_file(name, contents)

    def load(self, name, default=None):
        """
        Load a file from scratch.
        """
        self._log("Loading: " + name)
        return self._revman.load_file(name)

    def _add_reload_step(self, step_name, step_attrs):
        """
        Add a step to the reload journal.  This will append to RELOAD_JOURNAL.
        """
        reload_step = utils.yaml_dump([{step_name: step_attrs}])
        self._revman.append_file(RELOAD_JOURNAL, reload_step)

    def _strip_header(self, string, comment_char='#'):
        """
        Strips the header off of a string and returns the rest of the string
        and the checksum as a tuple.

        :param str string: A string which contains an auto-generated CUE header
        :param str comment_char: comment character expected in header
        :return: Tuple of (<string without header>, checksum)
                 None if cannot find header at the start of string
        """
        header = self._header.format(comment_char=comment_char, md5sum='')
        header = header.rstrip('\n')
        # Match the whole header, and capture the hex digest in group 1
        pattern = r'{}([a-z0-9]{})\n'.format(header, '{32}')
        match = re.match(pattern, string)
        if not match:
            # Try again with the legacy CUE name
            # MIGRATION: This code can be removed after 2021/09/01
            header = self._header_cue.format(
                comment_char=comment_char, md5sum='')
            header = header.rstrip('\n')
            # Match the whole header, and capture the hex digest in group 1
            pattern = r'{}([a-z0-9]{})\n'.format(header, '{32}')
            match = re.match(pattern, string)
            if not match:
                return
        # Remove the full match to get file contents
        file_contents = string[match.end(0):]
        return file_contents, match.group(1)

    def _check_orig_file(self, src, dst, comment_char="#"):
        """
        Copy the current deployed version of dst into the revman and store it
        as src.orig.  Then check if the orig file is managed by CUE and if that
        file has been manually edited.  It does this by looking for an
        auto-generated header.  If it finds one, it calculates an md5sum and
        checks this against the md5sum stored in the header.

        :param str src: relative path from the revman root directory
        :param str dst: absolute path to the installed location.
        :param str comment_char: comment character for destination file type

        :return: Tuple of (`valid`, `checksum`).  `valid` is True if orig file
            contains header and checksum is valid.  False if not.  `checksum`
            is the `dst` file's current checksum.
        """
        orig = src + ".orig"

        # First, as root, get a copy of the destination file and put it under
        # the revman as an "orig" file.
        try:
            self._revman.copy_file_in(dst, orig)
        except FileNotFoundError:
            # If the file doesn't exist, then it's OK to say CUE can manage it.
            return True, None

        # Now read it in as the cue user.
        orig_str = self._revman.load_file(orig)

        # Get the header.
        res = self._strip_header(orig_str, comment_char=comment_char)

        # If there wasn't a header, then the file is not managed
        if not res:
            return False, None

        # If the header doesn't match, then the file has been manually modified
        file_contents, prev_checksum = res
        new_checksum = md5(file_contents.encode('utf-8')).hexdigest()
        if prev_checksum != new_checksum:
            return False, new_checksum

        # We own it
        return True, new_checksum

    def _can_overwrite(self, dst, controlled, apply_overwrite):
        """
        Return true if dst, the destination file, can be overwritten.
        controlled indicates if we currently control it.  apply_overwrite is
        the value of platform.config.apply_overwrite, as configured by the
        user.

        It will report warnings, as appropriate.

        :param str dst: absolute path to the installed location.
        :param str controlled: If True, the file was last written by us, so we
            control it.
        :param str apply_overwrite: If "all", we're allowed to overwrite files
             we don't control, but with a warning.
        :return: True if we can overrite the file otherwise False.
        """
        if controlled:
            # If we control it, we can always overwrite it.
            return True
        else:
            if apply_overwrite == "all":
                # All files are overwritable, but warn that we're about to
                # clobber local changes.
                self.report_warning(WillOverwriteWarning(dst))
                return True
            else:
                # We cannot overwrite it.
                self.report_warning(WillNotOverwriteWarning(dst))
                return False

    def _check_overwrite(self, src, dst, comment_char):
        """
        Utility to fetch all the controls and call _can_overwrite().

        Returns the result from _can_overwrite().
        """
        # Get the apply config setting.
        settings = self.config_v1.getSystemConfigApply()

        # If it's a file we're supposed to ignore, then do that.
        if dst in settings["ignore"]:
            logger.info('Ignoring %s', dst)
            return False, 0

        # check the original file
        controlled, orig_checksum = self._check_orig_file(
            src, dst, comment_char)

        # Can we overwrite it?
        can_overwrite = self._can_overwrite(
            dst, controlled, settings["overwrite"])

        return can_overwrite, orig_checksum

    def stage_file(self, src, dst, chown="root.", chmod="0644",
                   comment_char="#"):
        """
        Stages a src file in the scratch area to be installed.

        `src` is a relative path and `dst` is an absolute path.  When
        installed, the file will be chown'd and chmod'd.  Returns True if the
        file was successfully staged.

        The file will not be staged if:
        - The destination is in the ignore list.
        - The destination has been changed manually and we're not allowed to
          overwrite it.
        - The file hasn't changed since the last apply.

        To stage the file, we first copy the current deployed version into the
        scratch area and store it as the orig.  If the orig is managed by CUE,
        we create the staged file by attaching a new header.

        :param str src: relative path from the scratch area.
        :param str dst: absolute path to the installed location.
        :param str chown: argument to "chown", setting the destination file's
             ownership.
        :param str chmod: argument to "chmod", setting the destination file's
             permissions.
        :param str comment_char: comment character for destination file type
        :return: True if the file was staged successfully otherwise False.
        """
        staged = src + ".dst"

        can_overwrite, orig_checksum = \
            self._check_overwrite(src, dst, comment_char)

        if can_overwrite:
            # We're allowed to overwrite it, so stage the update

            # Read in the source file
            src_str = self._revman.load_file(src)

            # Build the header.
            digest = md5(src_str.encode('utf-8')).hexdigest()
            header = self._header.format(
                comment_char=comment_char, md5sum=digest)
            staged_contents = header + src_str

            # Stage it
            if orig_checksum and orig_checksum == digest:
                msg = (f"Rendered '{src}' is the same as '{dst}' and does"
                       " not need to be staged.")
                self._log(msg)
                return False
            else:
                self._log(f'Staging {src!r} to {dst!r}')
                self._revman.save_file(staged, staged_contents)

            # Remove it from the orphan list
            try:
                self.stage_orphans.remove(staged)
            except ValueError:
                pass

            # Schedule it to be installed
            self._add_reload_step("install", {
                "src": staged, "dst": dst, "chown": chown, "chmod": chmod})
            return True
        return False

    def schedule_remove_file(self, tgt, comment_char="#"):
        """
        Schedules `tgt` to be removed.

        We'll follow the same rules as stage_file().  If we don't
        own the file we won't touch it.

        :param str tgt: absolute path to the installed location.
        :param str comment_char: comment character for destination file type
        :return: True if the file was scheduled for removal, otherwise False.
        """
        src = basename(tgt)
        can_overwrite, _ = self._check_overwrite(src, tgt, comment_char)

        if can_overwrite:
            # We're allowed to overwrite it, so stage the removal

            # Schedule it to be removed
            self._add_reload_step("remove", {"tgt": tgt})
            return True
        return False

    def schedule_run_script(self, script, user="root"):
        """
        Given the name of a script in the scratch area, schedule it to be run
        as the given user.
        """
        # Make sure the file exists before scheduling it
        logger.info('Schedule run of %s', script)
        self._revman.stat_file(script)
        self._add_reload_step("run_script",
                              {"script": script, "user": user})

    def _schedule_systemctl(self, action, systemd_unit):
        """
        Given a systemctl action and systemd unit name, schedule it to be run.
        """
        self._add_reload_step("systemctl", {
            "action": action, "unit": systemd_unit
        })

    def schedule_reload_or_restart(self, systemd_unit):
        """
        Given a systemd unit name, schedule it to be reloaded or restarted if
        necessary.  If the systemd unit is not already running, it will be
        started.  If it does not support reload, it will be restarted instead.
        """
        self._schedule_systemctl("reload-or-restart", systemd_unit)

    def schedule_reload(self, systemd_unit):
        """
        Given a systemd unit name, schedule it to be reloaded.  If the systemd
        unit is not already running, it will not be started.  It is assumed the
        systemd unit supports reload.
        """
        self._schedule_systemctl("reload", systemd_unit)

    def schedule_restart(self, systemd_unit):
        """
        Given a systemd unit name, schedule it to be restarted.  If the systemd
        unit is not already running, it will be started.
        """
        self._schedule_systemctl("restart", systemd_unit)

    def schedule_stop(self, systemd_unit):
        """
        Given a systemd unit name, schedule it to be stopped.  If the systemd
        unit is already stopped, this is a no-op.
        """
        self._schedule_systemctl("stop", systemd_unit)

    def schedule_enable(self, systemd_unit):
        """
        Given a systemd unit name, schedule it to be enabled.  If the systemd
        unit is already enabled, this is a no-op.
        """
        self._schedule_systemctl("enable", systemd_unit)

    def schedule_disable(self, systemd_unit):
        """
        Given a systemd unit name, schedule it to be disabled.  If the systemd
        unit is already enabled, this is a no-op.
        """
        self._schedule_systemctl("disable", systemd_unit)


class Reloader(object):
    """
    Helper object that runs reload operations and accumulates the results.
    """

    def __init__(self, ctx, revman):
        """
        Create a Reloader for the given global context and revman.
        """
        self._revman = revman
        self._ctx = ctx
        self.errors = []
        self.systemctl = {}

    def run_reload(self, ops, state_controls=None):
        """
        Loop through a list of reload operations and run them.  systemctl
        operations will be accumulated and run at the end.
        """
        reload = py_.get(state_controls, 'reload')
        with self._revman:
            self._revman.start_write("Reloading")
            self._log("## Reloading")

            # Loop through the reload the ops and run each operation.
            for op in ops:
                if "install" in op:
                    self._reload_install(**op["install"])
                elif "remove" in op:
                    self._reload_remove(**op["remove"])
                elif "run_script" in op:
                    self._reload_run_script(**op["run_script"])
                elif "systemctl" in op:
                    self._collect_systemctl(**op["systemctl"])

            # Batch run the systemctl operations
            if reload == 'skip':
                logger.info("Skip reload requested.")
            else:
                self._reload_systemctl()

    def _log(self, msg):
        """
        Log a message.  This will append to LOG_FILE.
        """
        # Strip line feeds for the python logger
        logger.dump(
            "Reloader", msg.replace("\n", ".").replace("\r", ".").strip()
        )
        # But keep them for our log
        self._revman.append_file(LOG_FILE, [msg, "\n"])

    def _report_error(self, error):
        """
        Report an error experienced during the reload.
        """
        self._log("Error: " + error.message)
        self.errors.append(error)

    def _reload_install(self, src, dst, chown, chmod):
        """
        Install a file from our scratch area to the system.
        """
        self._log(f"Installing: {src} to {dst} with {chown} {chmod}")
        try:
            self._revman.copy_file_out(src, dst, chown=chown, chmod=chmod)
        except Exception as ex:
            self._report_error(InstallFileError(dst, str(ex)))

    def _reload_remove(self, tgt):
        self._log(f"Removing: {tgt}")
        try:
            self._ctx.fileops.remove_file(tgt, sudo=True)
        except FileNotFoundError:
            pass
        except Exception as ex:
            self._report_error(RemoveFileError(tgt, str(ex)))

    def _reload_run_script(self, script, user):
        """
        Run a script in our scratch area.
        """
        # Get the shell command.  If we're running as root, it's a simple sudo.
        # Otherwise, we need to sudo to su.
        if user == "root":
            shell = self._ctx.sh.sudo.bash
        else:
            shell = self._ctx.sh.sudo.su.bake("-", user, "-c")

        self._log(f"Running as {user}: {script}")
        try:
            self._revman.run_file(script, shell)
        except Exception as ex:
            self._report_error(RunScriptError(script, str(ex)))

    def _run_journalctl(self, timestamp, unit_name):
        """
        Run journalctl and get important logs since `timestamp` for
        `unit_name`.

        We'll return entries as a list of dictionaries.  `timestamp`
        must be the output of time.time().
        """
        self._log(f"Running journalctl")
        since = time.strftime("%H:%M:%S", time.gmtime(timestamp))
        raw = self._ctx.sh.sudo.journalctl(
            "-p", "3", "-u", unit_name, "-S", since, "-o", "json")
        return [json.loads(raw_line) for raw_line in raw.splitlines()]

    def _get_journalctl_errors(self, timestamp, unit_name):
        """
        Return an array of error strings encountered during a systemctl
        operation that started around `timestamp`.

        It's assumed this is called immediately after a systemctl failure.
        We'll call journalctl and look for errors since `timestamp`.
        """
        entries = self._run_journalctl(timestamp, unit_name)

        errors = []
        for entry in entries:
            errors.append(entry["MESSAGE"])

        return errors

    def _extract_journalctl_issue(self, action, timestamp, reason):
        """
        Given a systemctl reason, look for a journctl reference and digest it
        for the user.
        """
        # Trim the message out of the reason.
        new_reason = "\n".join(
            line.strip() for line in reason.splitlines()
            if "systemctl status" not in line and line.strip()
        )

        # Look for a journalctl message in reason and create issues.
        issues = []
        for m in re.finditer(r'systemctl status ([^"]+)" and "journalctl',
                             reason, re.MULTILINE):
            # Extract the unit name
            unit_name = m.group(1)

            # Get errors out of the journal
            errors = self._get_journalctl_errors(timestamp, unit_name)

            # Create the issue
            if errors:
                issues.append(JournalctlError(action, unit_name, errors))

        return issues, new_reason

    def _collect_systemctl(self, action, unit):
        """
        Collect systemctl operations.  We'll run them as batches later so that
        systemd can control the order.
        """
        # For each action, build a list of units that need it.
        unit_list = self.systemctl.setdefault(action, [])
        if unit not in unit_list:
            unit_list.append(unit)

    def _run_systemctl(self, action, systemd_units):
        """
        Run a systemctl action for the given list of units.  If
        the list is empty, this is a noop.
        """
        if systemd_units:
            self._log(f"systemctl {action}: {systemd_units}")
            start_time = time.time()
            try:
                self._ctx.sh.sudo.systemctl(action, *systemd_units)
            except self._ctx.sh.ErrorReturnCode as ex:
                # Start with the stderr message.
                reason = ex.stderr.decode("UTF-8").strip()

                # If it's telling us a service failed and to go into
                # journalctl, save the user the trip and do it for them.
                jctl_issues, reason = self._extract_journalctl_issue(
                    action, start_time, reason)

                # Report anything else systemctl might have had to say
                self._report_error(
                    SystemctlError(action, systemd_units, reason))

                # Then report all the journalctl issues.
                for issue in jctl_issues:
                    self._report_error(issue)

    def _reload_systemctl_action(self, action):
        """
        Run a specific systemctl action on the collected systemd units.
        """
        self._run_systemctl(action, self.systemctl.get(action, []))

    def _parse_list_units(self, text: str) -> list[tuple[str, str, str, str]]:
        # After the table, there's an empty line and then some information we
        # don't care about. We just want the table
        output_table, *_ = str(text).split("\n\n", 1)

        # EARLY RETURN
        if output_table.startswith("0"):
            # It's saying something like "0 units listed" (ie, there are no
            # items in the table.)
            return []

        header, *row_lines = output_table.splitlines()

        col_count = len(header.split())
        return [
            line.split(maxsplit=col_count)
            for line in row_lines
        ]

    def _systemctl_cleanup_vrf_services_after_ifreload(self):
        # Find the units we care about
        output = self._ctx.sh.systemctl(
            'list-units',
            "*@*",
            '--plain',
            type="service",
            state="failed",
        )
        failed_vrf_services = [
            name
            for name, *_ in self._parse_list_units(str(output))
        ]
        if failed_vrf_services:
            logger.dump(
                "Cleaning up failed VRF services.", failed_vrf_services
            )
            self._run_systemctl('restart', failed_vrf_services)

    def _reload_systemctl(self):
        """
        Run the collected systemctl operations.
        """
        # Order matters.
        self._reload_systemctl_action("stop")
        self._reload_systemctl_action("reload")
        self._reload_systemctl_action("reload-or-restart")
        self._reload_systemctl_action("restart")
        self._reload_systemctl_action("disable")
        self._reload_systemctl_action("enable")

        if (
            "ifreload-nvue.service"
            in self.systemctl.get('reload-or-restart', [])
        ):
            # HACK: Restart any failed vrf services after running
            #       'ifreload-nvue.service'. 'ifreload-nvue.service' is a thin
            #       wrapper around 'ifreload --all', giving us control over
            #       when we run it. Unfortunately, 'ifreload'
            #       starts/creates/enables systemd services for VRFs, which
            #       doesn't seem to work so well when you're running inside a
            #       systemd service.
            self._systemctl_cleanup_vrf_services_after_ifreload()


class CheckHealthEvent(AbstractEvent):
    """
    Event object for a check health operation.  This will be passed with the
    check_health event.
    """

    def __init__(self, ctx, errors=None):
        """
        Create a check health event object for the given global context.
        errors will be seeded with the list provided.
        """
        self._ctx = ctx
        self.fileops = ctx.fileops
        self.sh = ctx.sh

        self.errors = errors or []
        """
        List of errors reported by handlers.
        """

        self.warnings = []
        """
        List of warnings reported by handlers.
        """

    def report_error(self, error):
        """
        Report an error.  Errors are issues that *will not* clear on their own
        and prevent the FUnit's component from functioning.

        The error can be either a simple string or an Issue instance.  If error
        is a string, it will be converted into a GenericError issue, otherwise
        error must already be an Issue instance.
        """
        # Normalize into an Issue.
        if isinstance(error, str):
            error = issues.GenericError(error)

        self.errors.append(error)

    def report_warning(self, warning):
        """
        Report a warning.  Warnings are issues that *should* clear on their
        own, but are currently preventing the FUnit's component from
        functioning.

        The warning can be either a simple string or an Issue instance.  If
        warning is a string, it will be converted into a GenericWarning issue,
        otherwise warning must already be an Issue instance.
        """
        # Normalize into an Issue.
        if isinstance(warning, str):
            warning = issues.GenericWarning(warning)

        self.warnings.append(warning)

    def check_is_active(self, systemd_unit, expected="active"):
        """
        Given a systemd unit name, run `systemd is-active` and compare the
        result against `expected`, which defaults to `active`.  If it matches,
        return True.  If it's `failed`, report an error and return False.
        Otherwise, report a warning and return `False`.
        """
        try:
            action = "is-active"
            observed = str(self._ctx.sh.sudo.systemctl(action, systemd_unit))
            observed = observed.strip()

            # If it's in the state we expect, then we're done.
            if observed == expected:
                return True

            if observed == "failed":
                # Report an error instead of a warning because a systemd
                # failure is permanent.
                self.report_error(ServiceFailed(systemd_unit))
            else:
                self.report_warning(ServiceUnexpected(
                    systemd_unit, expected, observed))
            return False
        except Exception as ex:
            self.report_error(SystemctlError(
                action, [systemd_unit], ex.stderr.decode("UTF-8").strip()))
            return False


class _RevTagger:
    """
    Utility for creating tags on the "applied" revision. Each tag name uses a
    template to make retrieval easy.
    """

    # Before starting an apply
    TMPL_START_APPLY = "{branch_name}/start"
    # After merging an apply
    TMPL_MERGE_APPLY = "{branch_name}/done"
    # After a successful apply
    TMPL_SUCCESS_APPLY = "{branch_name}/success"
    # After a failed apply
    TMPL_FAIL_APPLY = "{branch_name}/failure"

    # Apply number id
    TMPL_APPLY_NUMBER_ID = "n/{apply_number}"

    # For the latest of a given tag type.
    NAME_LATEST = f"{APPLY_NAMESPACE}/latest"

    TARGET = "applied"

    @classmethod
    def _tag(cls, revman, tag_name):
        revman.create_tag(tag_name, cls.TARGET)

    @classmethod
    def tag_start_apply(cls, revman, branch_name):
        tag_name = cls.TMPL_START_APPLY.format(branch_name=branch_name)
        cls._tag(revman, tag_name)

    @classmethod
    def tag_merge_apply(cls, revman, branch_name):
        tag_name = cls.TMPL_MERGE_APPLY.format(branch_name=branch_name)
        cls._tag(revman, tag_name)

    @classmethod
    def tag_apply_number_id(cls, revman, apply_number):
        tag_name = cls.TMPL_APPLY_NUMBER_ID.format(apply_number=apply_number)
        cls._tag(revman, tag_name)

    @classmethod
    def tag_success_apply(cls, revman, branch_name):
        tag_name = cls.TMPL_SUCCESS_APPLY.format(branch_name=branch_name)
        cls._tag(revman, tag_name)
        # Tag it as the latest success too.
        cls._tag(
            revman,
            cls.TMPL_SUCCESS_APPLY.format(branch_name=cls.NAME_LATEST)
        )

    @classmethod
    def tag_fail_apply(cls, revman, branch_name, fail_state):
        tag_name = cls.TMPL_FAIL_APPLY.format(
            fail_state=fail_state,
            branch_name=branch_name,
        )
        cls._tag(revman, tag_name)


def _run_verify_config(ctx, rev_id, carry_forwards):
    """
    Run through the verify config process.  Return the state.
    """
    # First, load the config.  This will give us a "patch valid" config.
    try:
        conf = ctx.config_v1.getFull(rev_id)
    except exceptions.UnprocessableEntity as ex:
        # We can't load a "patch valid" config file.  This indicates a basic
        # syntax error.  Let the user know.
        return "invalid", "Invalid config", [issues.from_http_exception(ex)]

    # Next, validate the config against the schema.  If this passes,
    # we have a "schema valid" config.
    try:
        schema.validate_full_config(ctx, conf)
    except utils.ParseError as ex:
        # The config is not "schema valid".  Let the user know.
        return "invalid", "Invalid config", [
            issues.ConfigParseError(ex.location, ex.reason)]

    # Finally, ask our units to verify they can apply this config.  If so,
    # the schema is "verified" and can be applied.
    try:
        evt = VerifyConfigEvent(ctx, rev_id, carry_forwards)
        ctx.events.verify_config(evt)

        if len(evt.issues) > 0:
            # We're done for now; we can't apply.
            return "invalid", "Invalid config", evt.issues

        # We're good to go.
        return "verified", "Verified config", []

    except Exception:
        # We found a bug.  Log the stack.
        logger.exception("Uncaught exception during verify_config")

        # Tell the user something generic.
        return ("verify_error", "Failed to verify config",
                [issues.InternalError()])


def _start_applying_branch(get_revman, apply_meta):
    """
    Create a new branch to track the apply process.  Returns the new branch's
    name and a revman for it.
    """
    # Generate a unique branch name.  These branches will primarily be around
    # for debugging and eventually we'll want to garbage collect them.  So,
    # make them easy to identify.
    branch_name = datetime.now().strftime(
        "/".join([APPLY_NAMESPACE, utils.generate_branch_id()])
    )

    # Checkout applied and branch from there.
    with get_revman("applied") as revman:
        _RevTagger.tag_start_apply(revman, branch_name)
        revman.start_branch(branch_name)

        # We're now on the new branch.

        # In the first commit message, record what we're trying to apply.
        revman.start_write(
            "Starting apply\n\n---\n" + utils.yaml_dump(apply_meta))

        # Zero out the log files.
        revman.zero_file(LOG_FILE)
        revman.zero_file(RELOAD_JOURNAL)

    return branch_name


_ready_apply_error = ("ready_error", "Failed to prepare to apply",
                      [issues.InternalError()])


def _run_ready_apply(ctx, rev_id, revman, carry_forwards, warnings=None):
    """
    Run through the ready apply process.
    """
    try:
        # Ask our units to prepare for an apply.
        evt = ReadyApplyEvent(ctx, revman, rev_id, carry_forwards,
                              warnings=warnings)
        ctx.events.ready_apply(evt)

        # Remove any orphaned stage files to make sure we can't
        # install them.
        with revman:
            for orphan in evt.stage_orphans:
                revman.remove_file(orphan)

        if len(evt.warnings) > 0:
            # Someone has a concern.  Kick that back to the client.
            return "ays", "Are you sure?", evt.warnings

        # We're good to go.
        return "ready", "Ready to apply", []

    except Exception:
        # We found a bug.  Log the stack.
        logger.exception("Uncaught exception during ready_apply")

        # Tell the user something generic.
        return _ready_apply_error


_apply_error = ("apply_error", "Failed to apply", [issues.InternalError()])


def _run_reload(ctx, revman, state_controls=None):
    """
    Run through the reload tasks.
    """
    try:
        reloader = Reloader(ctx, revman)

        # Load the list of reload tasks and run them.
        ops = utils.yaml_load_safe(revman.load_file(RELOAD_JOURNAL))

        # If the journal was empty, ops will be None.  Normalize it to an
        # empty array.
        ops = ops or []

        # Run the operations.
        reloader.run_reload(ops, state_controls)

        # We're good to go.  We might have run into some errors, but we'll just
        # add those to the check_health errors.
        return "checking", "Checking health", reloader.errors

    except Exception:
        # We found a bug.  Log the stack.
        logger.exception("Uncaught exception during reloading")

        # Tell the user something generic.
        return _apply_error


def _run_check_health(ctx, errors, run_count,
                      state_controls=None):
    """
    Run through the check health process.
    """
    check_health = py_.get(state_controls, 'check-health')
    if check_health == 'skip':
        logger.info('Skip check_health requested')
        return "checked", "Checked", []
    else:
        try:
            # Ask our units to check their health
            evt = CheckHealthEvent(ctx, errors)
            ctx.events.check_health(evt)

            if len(evt.errors) > 0:
                # Someone has a problem.  Kick that back to the client.
                return \
                    "ignore_fail", "Failure during apply. Ignore?", evt.errors

            if len(evt.warnings) > 0:
                if run_count < CHECKING_TRIES:
                    # Someone needs a little more time.
                    return "checking", "Re-checking", evt.warnings
                else:
                    # The warnings aren't clearing.  Treat them as errors.
                    return ("ignore_fail", "Apply remains incomplete. Ignore?",
                            evt.warnings)

            # Good to go
            return "checked", "Checked", []

        except Exception:
            # We found a bug.  Log the stack.
            logger.exception("Uncaught exception during apply_config")

            # Tell the user something generic.
            return _apply_error


def _trigger_rollback(ctx, req):
    """
    Trigger a rollback. Queue up the "apply" job for the rollback revision and
    return an issue that clients can use to follow the rollback.

    Returns:
        A RollbackWarning for the caller to include in the failure transition
        issues.
        Or a RollbackFailed if this rollback is in response to another rollback
        failing.
    """
    if req.user is None:
        # ASSUMPTION: No user means it's a rollback.
        # Don't get stuck in a rollback loop.
        return RollbackFailed()

    target_tag = _RevTagger.TMPL_START_APPLY.format(
        branch_name=req.branch_name
    )
    logger.dump("Rolling back", target_tag)
    ctx.versions_v1.updateRevision(
        target_tag,
        state="apply",
        auto_prompt={
            "ays": "ays_yes",
            "ignore_fail": "ignore_fail_yes",
        },
    )
    return RollbackWarning(target_tag)


def _gen_applied_merge_msg(apply_meta):
    """
    Given a set of metadata about the apply, generate the message we should use
    when merging the working branch into applied.  This is the main message
    users will see when looking at their apply history.
    """
    subject = "{reason} by {user} via {method}".format(**apply_meta)
    body = utils.yaml_dump(apply_meta)

    return subject + "\n\n---\n" + body


def _merge_applied_branch(ctx, branch_name, get_revman, apply_meta):
    """
    Merge the working branch back into the applied branch.
    """
    with get_revman("applied") as revman:
        commit_msg = _gen_applied_merge_msg(apply_meta)
        revman.merge_branch(branch_name, commit_msg)
        _RevTagger.tag_merge_apply(revman, branch_name)

        # Now apply its human-readable apply-id. Count the number of entries
        # we're showing in history and use that. Note that this has already
        # been merged into 'applied', so the length is 1 + the index of the
        # newly created merge commit.
        apply_number = len(ctx.versions_v1.getHistory('applied'))
        _RevTagger.tag_apply_number_id(revman, apply_number)

    # Since the config on applied branch has changed, we need to dump it from
    # the cache.
    repo.clear_cached_config(ctx, "applied")


class UnknownStateError(Exception):
    """
    Indicates the requested state is unknown.  We'll usually take that to mean
    we're at an end state in the FSM and should quit looping.
    """
    pass


class ApplyFSMRequest:
    """
    Convenience object to track our progress through the FSM.
    """

    def __init__(self, rev_id, init_state, auto_prompt=None,
                 state_controls=None, user=None):
        self.rev_id = rev_id
        self.state = init_state
        self.reload_result = None
        self.health_checks = 0
        self._client_lock = Condition()
        self.new_state = None
        if auto_prompt is None:
            auto_prompt = {}
        self.auto_prompt = auto_prompt
        self.state_controls = state_controls or {}
        self.user = user
        self.carry_forwards = {}

    def __str__(self):
        """
        Represent ourselves by dumping our interesting keys
        """
        tolog = []
        for key in ["rev_id", "state", "branch_name"]:
            if hasattr(self, key):
                tolog.append(str(getattr(self, key)))
        return ":: " + " ".join(tolog)

    def wait_client_state_change(self, timeout):
        """
        Block until the client sets a new state or we timeout.  We'll return
        the new state.  If we timeout, new_state will be None.
        """
        # If the client has already specified a flag, then we'll go with that.
        self.new_state = self.auto_prompt.get(self.state)
        # EARLY RETURN
        if self.new_state is not None:
            # Act like the client sent it, so we know we're going through the
            # right steps.
            logger.info("auto_prompt %s", self.new_state)
            send_client_state_change(self.rev_id, self.new_state)
            return self.new_state

        def new_state_set():
            return self.new_state is not None

        # Wait for the client
        with self._client_lock:
            self._client_lock.wait_for(new_state_set, timeout)
            return self.new_state

    def notify_client_state_change(self, new_state):
        """
        Notify the FSM that the client has set a new state.  It is assumed
        the state is a valid transition.
        """
        with self._client_lock:
            self.new_state = new_state
            self._client_lock.notify()


def pretty_duration(td: timedelta) -> str:
    """Convert a timedelta into a human-friendly string."""
    units = ['hours', 'minutes', 'seconds']
    collected_units = {}
    for unit in units:
        conversion_delta = timedelta(**{unit: 1})
        count = td // conversion_delta
        if count:
            if count == 1:
                # Remove the trailing 's'
                unit = unit[:-1]
            collected_units[unit] = count
        td = td % conversion_delta  # noqa: S001
    if collected_units:
        return " ".join([
            f"{count} {unit}"
            for unit, count in collected_units.items()
        ])
    else:
        return '0 seconds'


def _run_apply_fsm(ctx, req):
    """
    Run a single iteration of the apply_config FSM, updating req with the
    current state.  At the end we'll call the progress callback.  If it's a
    state we don't recognize, we'll raise an UnknownStateError, which means the
    FSM should exit.
    """
    # Start each loop by assuming the best.
    req.progress_msg = ""
    req.issues = None

    # apply -> verifying
    # - Start applying this revision.  The client sets this state.
    if req.state == "apply":
        req.state = "verifying"

    # verifying -> verified,invalid,verify_error
    # - CUE is asking each FUnit (via the verify_config event) to verify the
    # revision's configuration.  It will collect "issues" from the FUnits.  If
    # there are any issues, they are reported to the user and the FSM enters
    # the invalid state.  If an internal error occurs, the FSM enters the
    # verify_error state.  Otherwise, CUE proceeds to the verified state.
    elif req.state == "verifying":
        result = _run_verify_config(ctx, req.rev_id, req.carry_forwards)
        req.state, req.progress_msg, req.issues = result

    # verified -> readying
    # - The config has been verified as valid.  CUE proceeds to the readying
    # state.
    elif req.state == "verified":
        # Eventually, we'll support a dry-run that just verifies the config.
        # This is where we would stop.

        # For now, go directly into readying.
        req.state = "readying"

    # readying -> ays, ready, ready_error
    # - CUE is asking each FUnit (via the ready_apply event) to prepare to
    # apply the revision's configuration, usually by generating config files,
    # staging them, and scheduling restarts or reloads.  FUnits can issue
    # "warnings" and CUE will collect them.  If there are any warnings, they
    # are reported to the user and the FSM enters the ays state.  If an
    # internal error occurs, the FSM enters the ready_error state.  Otherwise,
    # CUE proceeds to the ready state.
    elif req.state == "readying":
        # Build a set of metadata about the apply.  This is hardcoded for now,
        # but eventually, we'll get most of this from the caller.
        req.apply_meta = {
            "reason": "Config update",
            "user": req.user,
            "method": "CLI",
            "rev_id": req.rev_id,
            "state_controls": req.state_controls
        }
        logger.dump("run_apply FSM", meta=req.apply_meta)

        # We're going to attempt an apply.  Start a branch.
        req.branch_name = _start_applying_branch(
            req.get_revman, req.apply_meta)
        req.revman = req.get_revman(req.branch_name)

        # Get the config we plan to apply.
        conf = ctx.config_v1.getFull(req.rev_id)

        # Always apply using the current system's immutables and the current
        # build's defaults.  We can get those out of an empty config.
        empty_conf = schema.empty_config()

        # Warn the user if CUE has been upgraded since this config was built.
        # We detect upgrades by comparing the version of the defaults it was
        # originally built against.
        warnings = []
        defaults_version = conf.get("_defaults_version")
        if (defaults_version
                and defaults_version != empty_conf["_defaults"]["x-version"]):
            warnings.append(UpgradedWarning())

        # Warn the user if the hardware has changed.  We detect hardware
        # changes by comparing the immutables.  There might be other hardware
        # changes, but if the immutables didn't change then we don't care.
        if conf["immutables"] != empty_conf["immutables"]:
            warnings.append(HardwareChanged())

        # Update the config with the current immutables and defaults.
        conf["_defaults"] = empty_conf["_defaults"]
        conf["immutables"] = empty_conf["immutables"]

        # Save the config to the apply branch.
        repo.repo_put_internal(ctx, req.branch_name, conf)

        result = _run_ready_apply(ctx, req.rev_id, req.revman,
                                  req.carry_forwards, warnings)
        req.state, req.progress_msg, req.issues = result

    # ays -> ays_fail,ays_no,ays_yes
    # - CUE needs to the client to respond to warnings about this revision's
    # configuration.  If the client is sure it wants to proceed, it should set
    # the state to ays_yes, otherwise ays_no.  If the client does not respond,
    # the FSM eventually times out and goes to ays_fail.
    elif req.state == "ays":
        # Wait for the client to respond to the 'are you sure'.
        new_state = req.wait_client_state_change(AYS_WAIT_SEC)

        # If we timed out waiting on the client, go to ays_fail.
        if new_state is None:
            req.state = "ays_fail"
            req.progress_msg = "Aborted apply after warnings"
            req.issues = [issues.ClientTimeout()]

        # Otherwise, continue to the requested state
        else:
            req.state = new_state

    # ays_yes -> ready
    # - The client sets this state to indicate it DOES want to proceed, despite
    # the warnings.  The FSM will proceed to the ready state.
    elif req.state == "ays_yes":
        req.state = "ready"

    # ays_no -> ays_fail
    # - The client sets this state to indicate it DOES NOT want to proceed,
    # given the warnings.  The FSM will proceed to the ays_fail state.
    elif req.state == "ays_no":
        req.state = "ays_fail"
        req.progress_msg = "Declined apply after warnings"

    # ready -> reloading
    # - CUE and the FUnits are ready to apply this config.  The FSM proceeds to
    # the reloading state.
    elif req.state == "ready":
        req.state = "reloading"

    # reloading -> reloaded
    # - CUE makes the changes that were staged and scheduled by the FUnits
    # during the ready_apply event.  The FSM will proceed to the reloaded
    # state.
    elif req.state == "reloading":
        req.reload_result = _run_reload(ctx, req.revman, req.state_controls)
        logger.dump("run_apply FSM", reload_result=req.reload_result)
        req.state = "reloaded"

    # reloaded -> apply_error,checking
    # - CUE has applied the config changes, but may roll them back.  It adds a
    # commit to the applied branch, indicating that the configuration has
    # changed.  Note that this is done, even if there were internal errors
    # during the reloading state.  As a result, the "applied" branch tracks
    # transitions in the desired config, even if that desired config was
    # short-lived before a rollback.  If an internal error occurs, the FSM
    # enters the apply_error state.  Otherwise, the FSM enters the checking
    # state.
    elif req.state == "reloaded":
        # First, merge what we have.
        _merge_applied_branch(ctx, req.branch_name, req.get_revman,
                              req.apply_meta)

        # Then proceed to whatever state run_apply_config wants us in.
        req.state, req.progress_msg, req.issues = req.reload_result

    # checking -> ignore_fail,checked
    # - CUE is asking each FUnit (via the check_health event) to report on
    # their health.  It will collect "issues" from the FUnits.  If there are
    # any issues, they are reported to the user and the FSM enters the
    # ignore_fail state.  Otherwise, the FSM proceeds to the checked state.
    elif req.state == "checking":
        # If this is a re-check, give FUnits a little extra time to come up.
        if req.health_checks > 0:
            time.sleep(CHECKING_WAIT_SEC)

        # Run the health check.  Seed it with errors from the reload.
        _, _, reload_issues = req.reload_result
        req.health_checks += 1
        result = _run_check_health(
            ctx, reload_issues, req.health_checks, req.state_controls)
        req.state, req.progress_msg, req.issues = result

    # checked -> confirm,applied
    # - CUE has applied the config and it looks good.  If the user asked for a
    # confirmation, the FSM enters the confirm state.  Otherwise, the FSM
    # enters the applied state.
    elif req.state == "checked":

        if req.state_controls.get("confirm"):
            duration_seconds = req.state_controls['confirm']
            # Find the timestamp
            duration = timedelta(seconds=duration_seconds)

            # Trim microseconds. We don't need/want that granularity, and it's
            # a bunch of noise for users.
            now = datetime.now(timezone.utc).replace(microsecond=0)
            end_datetime = now + duration
            # Put the timestamp where it can be found by the next state
            # handler.
            req.end_wait = end_datetime
            # Make the timestamp ISO 8601
            str_stamp = end_datetime.isoformat()
            # Put it in an issue
            req.state = "confirm"
            req.progress_msg = (
                f"Applied configuration will be rolled back after"
                + f" {pretty_duration(duration)} unless confirmed."
            )
            req.issues = [ConfirmDurationInfo(str_stamp)]
        else:
            req.state = "applied"

    # ignore_fail -> apply_fail,ignore_fail_no,ignore_fail_yes
    # - CUE needs to the client to respond to errors during the checking state.
    # If the client wants to ignore the errors, it should set the state to
    # ignore_fail_yes, otherwise ignore_fail_no.  If the client does not
    # respond, the FSM eventually times out and goes to apply_fail.
    elif req.state == "ignore_fail":
        new_state = req.wait_client_state_change(IGNORE_FAIL_WAIT_SEC)

        # If we timed out waiting on the client, go to apply_fail.
        if new_state is None:
            req.state = "apply_fail"
            req.progress_msg = "Aborted apply after failures"
            req.issues = [issues.ClientTimeout()]
        else:
            req.state = new_state

    # ignore_fail_no -> apply_fail
    # - The clients sets this state to indicate it DOES NOT want to ignore the
    # errors.  The FSM will proceed to the apply_fail state.
    elif req.state == "ignore_fail_no":
        req.state = "apply_fail"

    # ignore_fail_yes -> applied
    # - The clients sets this state to indicate it DOES want to ignore the
    # errors.  The FSM will proceed to the applied state.
    elif req.state == "ignore_fail_yes":
        req.state = "checked"

    # confirm -> confirm_fail,confirm_no,confirm_yes
    # - CUE needs the client to respond to confirm the apply looks good.  If
    # the client wants to keep the apply, it should set the state to
    # confirm_yes, otherwise confirm_no.  If the client does not respond, the
    # FSM eventually times out and goes to confirm_fail.
    elif req.state == "confirm":
        end_wait: datetime = req.end_wait
        duration = end_wait - datetime.now(req.end_wait.tzinfo)
        new_state = req.wait_client_state_change(duration.total_seconds())
        # Did we timeout?
        if new_state is None:
            # Default to 'no'.
            req.state = "confirm_fail"
            req.progress_msg = "Aborted apply after lack of confirmation"
            req.issues = [issues.ClientTimeout()]
        else:
            req.state = new_state

    # confirm_no -> confirm_fail
    # - The client sets this state to indicate it DOES NOT want to keep the
    # apply.  The FSM will proceed to the confirm_fail state.
    elif req.state == "confirm_no":
        req.state = "confirm_fail"

    # confirm_yes -> applied
    # - The client sets this state to indicate it DOES want to keep the apply.
    # The FSM will proceed to the applied state.
    elif req.state == "confirm_yes":
        req.state = "applied"

    # By default, exit the FSM.  We're in an end state.
    else:
        raise UnknownStateError()

    # For failure states, queue up the rollback and add an issue so clients can
    # track the rollback.
    if req.state in ("apply_error", "apply_fail", "confirm_fail"):
        # ORDER MATTERS: Queue up the rollback job here. The target revision
        #       MUST be inside the apply_config FSM by the time the current
        #       revision's state shows as failed. Otherwise, there's a window
        #       where clients won't be able to follow the FSM of the rollback
        #       revision.
        rollback_issue = _trigger_rollback(ctx, req)
        req.issues = [*(req.issues or []), rollback_issue]

    # Update the log.  Log the issues, but don't let them fill up the info log.
    # We'll still have the details in apply_log.txt.
    if req.issues:
        issue_msg = "),(".join([issue.message.strip() for issue in req.issues])
        if len(issue_msg) > 200:
            issue_msg = issue_msg[:200] + "..."
        issue_msg = issue_msg.replace("\n", ".").replace("\r", ".")
    else:
        issue_msg = None

    logger.info(
        'Applying %s: %s%s', req.rev_id, req.state,
        " (" + req.progress_msg + ")" if req.progress_msg else ""
    )
    if issue_msg:
        logger.info("Apply Issues: %s", "(" + issue_msg + ")")

    # Update the client
    req.progress(req.state, req.progress_msg, req.issues)


_running_apply_req = None
"""
If there is an apply running, this will point to the job's state object.
"""


def _set_running_apply_req(running_apply_req):
    """
    Set the global apply_req.
    """
    global _running_apply_req
    _running_apply_req = running_apply_req


def _get_running_apply_req():
    """
    Get the global apply_req.
    """
    return _running_apply_req


def run_apply(
    ctx, rev_id, init_state, progress, get_revman, auto_prompt=None,
    state_controls=None, user=None
):
    """
    Apply the config at the given revision, starting with the given
    initial_state.

    progress is a callback through which we can update the revision's state and
    let the user know how things are going.

    get_revman is a callback that can return revision manager contexts for us.
    """
    req = ApplyFSMRequest(rev_id, init_state, auto_prompt=auto_prompt,
                          state_controls=state_controls, user=user)
    req.progress = progress
    req.get_revman = get_revman

    _set_running_apply_req(req)

    # Loop until we exit the FSM.
    try:
        while True:
            logger.dump("run_apply FSM", req)
            _run_apply_fsm(ctx, req)
    except UnknownStateError:
        # Loop exited.
        if req.state == "applied":
            with get_revman("applied") as revman:
                _RevTagger.tag_success_apply(revman, req.branch_name)
        elif req.state in ("apply_error", "apply_fail", "confirm_fail"):
            with get_revman("applied") as revman:
                _RevTagger.tag_fail_apply(revman, req.branch_name, req.state)
        else:
            logger.dump("run_apply FSM exited", req)
    except Exception:
        # This indicates a bug in _run_apply_fsm() itself, not in an FU's event
        # handler.  This should "never happen," but we know better, right?
        logger.exception("Uncaught exception in apply FSM")

        # If we got to the reload, treat it as an apply_error, otherwise a
        # ready_error.
        if req.reload_result:
            rollback_issue = _trigger_rollback(ctx, req)
            *err_progress, err_issues = _apply_error

            req.progress(*err_progress, [*err_issues, rollback_issue])

            with get_revman("applied") as revman:
                _RevTagger.tag_fail_apply(
                    revman, req.branch_name, "unknown_error"
                )
        else:
            req.progress(*_ready_apply_error)

    _set_running_apply_req(None)

    # Return the request object as a convenience for unittesting.
    return req


def start_apply(ctx, rev_id, progress, get_revman, auto_prompt=None,
                state_controls=None, user=None):
    """
    Start an apply of the given rev_id.  This will add a job to the queue and
    return quickly.  The actual apply will be run from the job thread.
    """
    # Define a closure to call apply_config().  The job queue needs
    # something it can run without passing arguments.
    def job_apply_config():
        run_apply(ctx, rev_id, "apply", progress, get_revman, auto_prompt,
                  state_controls, user)

    # Add it to the job queue.
    ctx.jobs.add("apply:" + rev_id, job_apply_config)


def send_client_state_change(rev_id, new_state):
    """
    The client has requested an apply_config state change.  This usually
    happens in response to an interactive step in the FSM.  If the state change
    is valid, pass it along to the running apply job.
    """
    req = _get_running_apply_req()
    if not req:
        logger.info(
            'Applying %s: Unable to update state because apply is not '
            'running', rev_id)
        return

    if req.rev_id != rev_id:
        logger.info(
            'Applying %s: Unable to update state because apply is running '
            'for a different revision', rev_id)
        return

    valid_transitions = {
        "ays": ["ays_yes", "ays_no"],
        "ignore_fail": ["ignore_fail_yes", "ignore_fail_no"],
        "confirm": ["confirm_yes", "confirm_no"],
    }

    # Make sure it's a valid state transition.
    if not (req.state in valid_transitions
            and new_state in valid_transitions[req.state]):
        # If the event state hasn't changed, it's probably a state change we
        # triggered (via an auto_prompt).  In that case, no need to get angry
        # in the logs.
        if req.state != new_state:
            logger.info('Applying %s: Invalid state transition (%s -> %s)',
                        rev_id, req.state, new_state)
        return

    # It's a good state transition.  Pass it on to the job.  It should be
    # blocked and waiting for us.  If the job has moved on (timed out), it
    # won't be listening on this req any more.  But that's OK.  The req is
    # still valid, so the code won't fail.  It'll just become a no-op.
    req.notify_client_state_change(new_state)
