# Copyright (C) 2018, 2019 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from cue import issues
from . import templates
import logging
logger = logging.getLogger(__name__)


class CoppTrapUnsupported(issues.BadInputIssue):
    """
    CoPP/Traps are unsupported on this platform
    """
    def __init__(self):
        super().__init__("copp_trap_unsupported")

    def _render_message(self):
        return "Copp/Traps are unsupported on this platform."


def is_copp_trap_supported(evt):
    """
    Return True if CoPP/Traps are supported.
    """
    hw = evt.platform_v1.getHardware()
    if (hw.get("manufacturer") == "Cumulus" and hw.get("model") == "VX"):
        return False
    else:
        return True


def prepare_unit(ctx):
    """
    Prepare the unit for lifecycle events.
    """
    # Tell the render engine about our templates
    ctx.render.register_package(templates)

    # Register for events
    ctx.events.verify_config.register("copp_trap_v1", handle_verify_config)
    ctx.events.ready_apply.register("copp_trap_v1", handle_ready_apply)
    ctx.events.check_health.register("copp_trap_v1", handle_check_health)


def handle_verify_config(evt):
    """
    Handle a verify_config event.

    Use evt.config_v1 to access the pending configuration's state.
    Use evt.report_issue to report any found issues.

    Args:
        evt: A VerifyConfigEvent instance.
    """
    copp_config = evt.config_v1.getControlPlane()
    if (copp_config['trap'] or copp_config['policer']) and \
       not is_copp_trap_supported(evt):
        evt.report_issue(CoppTrapUnsupported())


default_pol_cfg = {
    'arp': {
        'state': 'on',
        'rate': 800,
        'burst': 800},
    'bfd': {
        'state': 'on',
        'rate': 2000,
        'burst': 2000},
    'pim_ospf_rip': {
        'state': 'on',
        'rate': 2000,
        'burst': 2000},
    'bgp': {
        'state': 'on',
        'rate': 2000,
        'burst': 2000},
    'clag': {
        'state': 'on',
        'rate': 2000,
        'burst': 2000},
    'icmp_def': {
        'state': 'on',
        'rate': 100,
        'burst': 40},
    'dhcp_ptp': {
        'state': 'on',
        'rate': 2000,
        'burst': 2000},
    'igmp': {
        'state': 'on',
        'rate': 1000,
        'burst': 1000},
    'ssh': {
        'state': 'on',
        'rate': 1000,
        'burst': 1000},
    'icmp6_neigh': {
        'state': 'on',
        'rate': 500,
        'burst': 500},
    'icmp6_def_mld': {
        'state': 'on',
        'rate': 300,
        'burst': 100},
    'lacp': {
        'state': 'on',
        'rate': 2000,
        'burst': 2000},
    'lldp': {
        'state': 'on',
        'rate': 200,
        'burst': 200},
    'rpvst': {
        'state': 'on',
        'rate': 2000,
        'burst': 2000},
    'eapol': {
        'state': 'on',
        'rate': 2000,
        'burst': 2000},
    'ip2me': {
        'state': 'on',
        'rate': 1000,
        'burst': 1000},
    'acl_log': {
        'state': 'on',
        'rate': 100,
        'burst': 100},
    'nat': {
        'state': 'on',
        'rate': 200,
        'burst': 200},
    'stp': {
        'state': 'on',
        'rate': 2000,
        'burst': 2000},
    'l3_local': {
        'state': 'on',
        'rate': 400,
        'burst': 100},
    'span_cpu': {
        'state': 'on',
        'rate': 100,
        'burst': 100},
    'catch_all': {
        'state': 'on',
        'rate': 100,
        'burst': 100},
}


def build_copp_strs(copp_config):
    pol_cfg = default_pol_cfg
    for k, v in copp_config['policer'].items():
        k = k.replace('-', '_')
        if 'state' in v.keys():
            pol_cfg[k]['state'] = v['state']
        if 'rate' in v.keys():
            pol_cfg[k]['rate'] = v['rate']
        if 'burst' in v.keys():
            pol_cfg[k]['burst'] = v['burst']
    cs = ['################ COPP CONFIG #################\n']
    ts = ['################ TRAP CONFIG #################\n']
    st = {'on': 'TRUE', 'off': 'FALSE'}
    trap_cfg = copp_config['trap']
    for proto in pol_cfg.keys():
        prefix = 'copp.' + proto
        if 'state' in pol_cfg[proto].keys():
            cs.append(prefix + '.enable = ' + st[pol_cfg[proto]['state']])
        if 'rate' in pol_cfg[proto].keys():
            cs.append(prefix + '.rate = ' + str(pol_cfg[proto]['rate']))
        if 'burst' in pol_cfg[proto].keys():
            cs.append(prefix + '.burst = ' + str(pol_cfg[proto]['burst']))
        cs.append('\n')
    for trap in trap_cfg.keys():
        prefix = 'trap.' + trap
        if 'state' in trap_cfg[trap].keys():
            ts.append(prefix + '.enable = ' + st[trap_cfg[trap]['state']])
        ts.append('\n')
    return cs, ts


def handle_ready_apply(evt):
    """
    Prepare to apply the config.  Generate config files and scripts.  Also,
    look for issues in the config that would prevent us from applying it.

    Use evt.config_v1 to access the pending configuration's state.
    Use evt.render to render a template with the given variables.
    Use evt.stage_file to prepare a rendered file to be copied over to its
        ultimate destination.

    Args:
        evt: A ReadyApplyEvent instance.
    """
    if not is_copp_trap_supported(evt):
        return
    pol_conf = '/etc/cumulus/control-plane/policers.conf'
    trap_conf = '/etc/cumulus/switchd.d/trap.conf'
    copp_config = evt.config_v1.getControlPlane()
    logger.info(copp_config)
    copp_strs, trap_strs = build_copp_strs(copp_config)
    logger.info(copp_strs)
    logger.debug(trap_strs)
    evt.render(templates, "policers.conf", config=copp_strs)
    evt.stage_file("policers.conf", pol_conf)
    evt.render(templates, "trap.conf", config=trap_strs)
    evt.stage_file("trap.conf", trap_conf)
    evt.render(templates, "install_copp.sh")
    evt.schedule_run_script("install_copp.sh")


def handle_check_health(evt):
    """
    Handle a check_health event.

    Use evt.report_error to report any found errors.
    Use evt.report_warning to report any found warnings.

    Warnings *should* clear on their own, given some time, usually indicating
    that a component is still "coming up."  Errors *will not* clear on their
    own.

    Args:
        evt: A CheckHealthEvent instance.
    """
