#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of Nvidia Corporation
# and its affiliates (the "Company") and all right, title, and interest
# in and to the software product, including all associated intellectual
# property rights, are and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
# Copyright (C) 2021 Cumulus Networks, inc.
# All Rights reserved.
"""
# Actions

For when you need one-time operations that can't be sensibly put in a
startup.yaml.

## Defining action runners

Use the `action_runner` decorator on your event handler.

Example:

```python
# lifecycle.py
from cue_cue_v1 import actions

def prepare_unit(ctx):
    ctx.events.action_state.register(
        '<unit-name>',
        handle_action_state_brew,
    )

@actions.action_runner("@brew", "/coffee/{coffee-id}")
def handle_action_state_brew(evt, coffee_id):
    # Handle the "brew" event.
    ...
```

## Advanced

### Preparing the action handling

_Shouldn't be necessary, since it's handled in `cue_v1`'s `prepare_unit`
function_

Actions depend on a couple of things in `ctx`:

- an `ActionStateStore` (in `ctx._action_store`); and
- `handle_action_state_start` registered to `ctx.events.action_state`.
"""

from __future__ import annotations
import sh
import logging
from cue import issues
from cue.events import AbstractEvent
from cue.exceptions import BadRequest
from dataclasses import dataclass, field
from cue.states import validate_transition, TransitionError
from cue.utils.openapi_paths import (
    convert_to_list_path,
    format_params_and_convert_to_list_path,
)
from functools import update_wrapper, wraps
from typing import Callable, Sequence
from inspect import signature


logger = logging.getLogger(__name__)


ACTION_ENTRY_STATES = ["start"]
ACTION_TRANSITIONS = {
    "start": ["running"],
    "running": ["action_error", "action_success"]
}
INACTIVE_STATE = "inactive"


###############################################################################
# lifecycle utilities

def prepare_actions(ctx):
    """
    Prepare the action-specific attributes of ctx.

    Call from `prepare_unit`.
    """
    ctx._action_store = ActionStateStore(ctx)
    ctx.events.action_state.register(
        __name__, handle_action_state_start,
    )


def handle_action_state_start(evt):
    """
    Queue up each started action to be run in the jobs thread.
    """
    if evt.new_state != "start":
        return

    def start_running():
        evt.transition(new_state="running")

    evt._ctx.jobs.add(
        f"running {evt.action_key!r}",
        start_running,
    )


@dataclass
class ActionRunner:
    """
    A convenience wrapper for handling action_state events.

    Instead of directly initializing, use the `action_runner` decorator.

    When an ActionRunner is called with an `evt` object, the wrapped
    `handler_function` will be called with `evt` and any path parameters passed
    in as kwargs.

    The wrapped `handler_function` will NOT be called unless the new_state is
    'running'.

    After calling the `handler_function`, the action's state will automatically
    be transitioned to 'action_success', unless an exception was raised (in
    which case, the state is transitioned to 'action_error'.)
    """

    action_id: str
    path_id: str
    handler_function: Callable

    def __post_init__(self):
        # TODO: make sure signature matches __call__
        update_wrapper(self, self.handler_function)

    def __call__(self, evt):
        # EARLY RETURN
        if (self.action_id, self.path_id) != (evt.action_id, evt.path_id):
            # Not the action we're implementing.
            return
        elif evt.new_state != "running":
            # Not the "running" state of the action.
            return
        try:
            self.handler_function(evt, **evt.action_key.controller_kwargs)
        except sh.ErrorReturnCode as erc:
            msg = f"Non-zero return code '{erc.exit_code}'"\
                  f" while executing '{erc.full_cmd}'"
            logger.error(msg)
            evt.report_issue(msg)
            evt.transition(new_state="action_error")
        except Exception:
            msg = "Uncaught exception during 'running'"
            logger.exception(msg)
            evt.report_issue(msg)
            evt.transition(new_state="action_error")
        else:
            if len(evt.issues) > 0:
                # Errors reported by FUnit
                evt.transition(new_state="action_error")
            else:
                evt.transition(new_state="action_success")

    @classmethod
    def decorate(cls, action_id, path_id):
        return lambda handler_function: cls(
            action_id, path_id, handler_function
        )


action_runner = ActionRunner.decorate


###############################################################################
# controller utilities

def action_get_controller(controller_function=None):
    """
    Decorator for 'GET' controllers to include action information.
    """

    # EARLY RETURN
    if controller_function is None:
        # Support `@action_get_controller()`
        return lambda func: action_get_controller(func)

    controller_signature = signature(controller_function)

    @wraps(controller_function)
    def action_get_controller_wrapper(ctx, rev, *args, **kwargs):
        result = controller_function(ctx, rev, *args, **kwargs)
        # EARLY RETURN
        if rev != "operational":
            # Only add the action data to operational results
            return result

        # Make sure we can associate positional and keyword arguments to
        # argument names.
        bound_args = controller_signature.bind(ctx, rev, *args, **kwargs)

        path_id = ctx.current_operation.path
        path_parameters = get_path_params_from_controller_kwargs(
            path_id, **bound_args.arguments,
        )

        for action_key, action_item in ctx._action_store.search(
            path_id=path_id,
            path_parameters=path_parameters,
        ):
            result[action_key.action_id] = action_item.to_json()

        return result

    return action_get_controller_wrapper


def action_patch(ctx, body, **kwargs):
    """
    Returns:
        The updated state of the action.

    Side-effects:
        Updates `ctx._action_store` and emits an `action_state` event.

    Raises:
        BadRequest if the patch isn't valid.
    """
    ###########################################################################
    # Validate the action request
    if len(body.keys()) != 1:
        raise BadRequest(
            "An action request should have only one top-level property."
        )
    action_ids = [
        action_id
        for action_id in body.keys()
        if action_id.startswith("@")
    ]
    if len(action_ids) != 1:
        raise BadRequest(
            "An action request's top-level property must be an action id."
        )

    ###########################################################################
    # Update the _action_store with the action request.
    action_id = action_ids[0]
    action_key = ActionKey.from_controller_kwargs(
        action_id, ctx.current_operation.path, **kwargs
    )

    patch_value = body[action_id]

    try:
        updated = ctx._action_store.update(action_key, **patch_value)
    except TransitionError:
        # Idempontency: don't bother clients with transition errors.
        updated = ctx._action_store[action_key]

    return {action_key.action_id: updated.to_json()}


###############################################################################
# Action event implementation

def get_path_params_from_controller_kwargs(
    path_id: str, **kwargs,
) -> tuple[str, ...]:
    """
    Return a tuple of string path param values, in the order specified by
    `path_id`.

        >>> get_path_params_from_controller_kwargs(
        ...     '/foo/{foo-id}/bar/{bar-id}',
        ...     foo_id="FOO", bar_id="BAR", extra="[ignore me!]"
        ... )
        ('FOO', 'BAR')

    Returns:
        An n-tuple of string param values.
    """
    path_id_segments = convert_to_list_path(path_id)
    path_value_segments = format_params_and_convert_to_list_path(
        path_id,
        kwargs
    )
    return tuple(
        value_segment
        for id_segment, value_segment in zip(
            path_id_segments, path_value_segments
        )
        if id_segment != value_segment
    )


@dataclass(frozen=True)
class ActionKey:
    """
    A hashable key identifying a specific action on a specific endpoint.
    """

    action_id: str
    """The action type (e.g., '@clear')"""

    path_id: str
    """The OpenAPI path id for the operation (e.g., '/coffee/{coffe-id}')"""

    path_parameters: tuple[str, ...]
    """The path parameters (in order) for the given request"""

    @classmethod
    def from_controller_kwargs(cls, action_id, path_id, **kwargs):
        """
        Determine the ActionKey's path_parameters tuple from `kwargs`.
        """
        return cls(
            action_id=action_id,
            path_id=path_id,
            path_parameters=get_path_params_from_controller_kwargs(
                path_id, **kwargs,
            ),
        )

    @property
    def controller_kwargs(self):
        """`path_parameters` as a `kwargs` dict"""
        keywords = [
            segment.strip("{}").replace("-", "_")
            for segment in convert_to_list_path(self.path_id)
            if segment.startswith("{")
        ]
        return dict(zip(keywords, self.path_parameters))

    def __repr__(self):
        # Get a more readable repr.
        cls_name = type(self).__name__
        action_id = self.action_id
        path_id = self.path_id
        path_parameters = self.path_parameters

        return f"{cls_name}({action_id!r}, {path_id!r}, {path_parameters!r})"


@dataclass
class ActionStateItem:
    """The current state of an action in `ActionStateStore`"""

    state: str
    issue: list = field(default_factory=list)

    def to_json(self):
        return vars(self)


class ActionStateStore:
    """
    An in-memory store of the actions that are being run or have been run since
    starting nvued.

    Maps `ActionKey` instances to `ActionStateItem` instances.
    """

    def __init__(self, ctx):
        self._ctx = ctx
        self._data = {}

    def __getitem__(self, action_key: ActionKey) -> ActionStateItem:
        try:
            return self._data[action_key]
        except KeyError:
            return ActionStateItem(INACTIVE_STATE)

    def search(self, **attr_matchers):
        """
        Yield all ActionStateItem instances whose attributes match the
        `attr_matchers`.
        """
        for key, item in self._data.items():
            all_matches = all(
                getattr(key, match_prop) == match_value
                for match_prop, match_value in attr_matchers.items()
            )
            if all_matches:
                yield key, item

    def update(
        self,
        action_key: ActionKey,
        state: str,
        *,
        issues: Sequence = (),
    ) -> ActionStateItem:
        """
        Update the `action_key`'s value in store, emitting an
        `ActionStateEvent` to trigger all registered handlers.

        Side-effects:
            - Emits an ActionStateEvent for the transition.
            - Updates the stored ActionStateItem for action_key, upon entering
              the emitted event's context.

        Raises:
            cue.states.TransitionError if transitioning to `state` is invalid.
        """
        old_state = self[action_key].state
        # This will raise an error if the transition isn't valid.
        evt = ActionStateEvent(
            self._ctx,
            old_state=old_state,
            new_state=state,
            action_key=action_key,
        )

        action_state_obj = ActionStateItem(state, issue=list(issues))

        @evt.before_emit
        def update_stored_state():
            # Only update the value when we're actually handling the event.
            self._data[action_key] = action_state_obj

        # For cases where no events are in the queue (ie, when a client starts
        # an action), this will immediately handle the event and return.
        self._ctx.events.action_state(evt)

        # NOTE: Because of multithreading, it's possible that the state has
        #       moved farther along by this point. That's fine. I'd rather be
        #       consistent with return values than be "correct" in surprising
        #       ways.
        return action_state_obj


class ActionStateEvent(AbstractEvent):
    """
    This should never be initialized directly. Instead, it's emitted as part of
    updating an action's state in the ActionStateStore.
    """

    def __init__(
        self,
        ctx,
        *,
        old_state: str,
        new_state: str,
        action_key: ActionKey,
    ):
        """
        Raises:
            cue.states.TransitionError if it's not a valid transition.
        """
        # Make it an error to express an invalid transition. We'll let the
        # caller decide whether or not it makes sense to suppress the error.
        validate_transition(
            ACTION_ENTRY_STATES,
            ACTION_TRANSITIONS,
            old_state,
            new_state,
        )
        self._ctx = ctx
        self.sh = ctx.sh
        self.fileops = ctx.fileops
        self._action_store = ctx._action_store
        self.old_state = old_state
        self.new_state = new_state

        self.action_key = action_key

        self._before_emit_callbacks = []
        self.issues = []

    def report_issue(self, issue):
        """
        Report an issue in the configuration that must be addressed.  The issue
        can be either a simple string or an Issue instance.  If issue is a
        string, it will be converted into a GenericError issue, otherwise issue
        must already be an Issue instance.
        """
        if isinstance(issue, str):
            issue = issues.GenericError(issue).as_dict()

        if issue not in self.issues:
            self.issues.append(issue)

    def before_emit(self, callback):
        """Add a function to be called upon entering the event's context."""
        self._before_emit_callbacks.append(callback)

    @property
    def action_id(self) -> str:
        """The action_id of this event's action."""
        return self.action_key.action_id

    @property
    def path_id(self) -> str:
        """The path_id of this event's action."""
        return self.action_key.path_id

    def transition(self, *, new_state):
        """
        Update the stored state for this action, emitting a new action_state
        event.
        """
        self._action_store.update(
            self.action_key, state=new_state, issues=self.issues
        )

    def __repr__(self):
        # Don't show un-useful attrs in the repr.
        ignore_attrs = {
            "_ctx", "sh", "fileops", "_action_store", "_before_emit_callbacks"
        }

        attrs_repr = ', '.join([
            f"{key}={value!r}"
            for key, value in vars(self).items()
            if key not in ignore_attrs
        ])

        return f"<{type(self).__name__} {attrs_repr}>"

    def __enter__(self):
        """
        Enter the event's context.

        Called before the first handler is called.
        """
        for callback in self._before_emit_callbacks:
            callback()
