#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of Nvidia Corporation
# and its affiliates (the "Company") and all right, title, and interest
# in and to the software product, including all associated intellectual
# property rights, are and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
# Copyright (C) 2021 Cumulus Networks, inc.
# All Rights reserved.

from cue.morpher import expand_name_range, ExpansionError
from cue.utils import reusable_controller
from cue_cue_v1.actions import action_patch
from cue.exceptions import BadRequest


@reusable_controller
def base_patch(ctx, rev, body, **kwargs):
    """Handle a basic PATCH request to the NVUE object model."""
    allows_action = "action" in ctx.current_operation.tags

    if allows_action:
        if rev == "operational":
            # Attempt to treat the body as an action
            return action_patch(ctx, body, **kwargs)
        elif any(key.startswith("@") for key in body.keys()):
            # Don't allow actions on non-operational revisions.
            formatted_action_keys = ", ".join([
                repr(key)
                for key in body.keys()
                if key.startswith("@")
            ])
            raise BadRequest(
                "Actions are only allowed on 'operational'."
                + f" (Found action keys: {formatted_action_keys})"
            )

    key_path = _oai_path_to_key_path(ctx.current_operation.path, kwargs)

    # Convert the body into a PATCH relative to root
    for key in reversed(key_path):
        body = {key: body}

    root = ctx.cue_v1.updateRoot(rev, body)
    return _expansion_aware_deep_get(root, key_path)


@reusable_controller
def base_delete(ctx, rev, **kwargs):
    """Handle a basic DELETE request to the NVUE object model."""
    return base_patch(ctx, rev, None, **kwargs)


def _expansion_aware_deep_get(obj, path):
    for key in path:
        try:
            obj = obj[key]
        except KeyError:
            try:
                # HACK: Try it as a range and get the first item
                # TODO: CUE-2604
                obj = obj[expand_name_range(key)[0]]
            except (KeyError, ExpansionError):
                # EARLY RETURN
                # We did our best and couldn't find it. Return an empty dict to
                # minimize chances of unexpected failure.
                # TODO: Revisit whether this is the right thing to do when the
                #       requested object doesn't exist anymore.
                return {}

    return obj


def _oai_path_to_key_path(oai_path, param_values):
    """
    Fill an oai path template with param values from this request.

    Returns:
        A tuple of path keys representing the OM location of this request.
    """
    tokens = oai_path.split("/")[1:]
    key_path = []
    for token in tokens:
        if token.startswith("{"):
            param_id = token.strip("{}").replace("-", "_")
            key_path.append(param_values[param_id])
        else:
            key_path.append(token)
    return tuple(key_path)
