# Copyright (C) 2018-2021 Cumulus Networks, inc
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from cue.exceptions import NotFound
from cue import utils


###############################
# Bridge
###############################

def bridge_get(ctx, rev):
    if rev != "operational":
        return ctx.config_v1.getBridge(rev)
    else:
        return ctx.netlink_v1.getBridge()


###############################
# Bridge Domains
###############################

def domains_get(ctx, rev):
    output = bridge_get(ctx, rev)
    try:
        return output["domain"]
    except KeyError:
        raise NotFound


###############################
# Bridge Domain
###############################

def domain_get(ctx, rev, domain_id):
    domains = domains_get(ctx, rev)
    try:
        return domains[domain_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# Bridge Domain STP
###############################

def stp_get(ctx, rev, domain_id):
    return domain_get(ctx, rev, domain_id)["stp"]


###############################
# STP State
###############################

def stp_state_get(ctx, rev, domain_id):
    return stp_get(ctx, rev, domain_id)["state"]


###############################
# Bridge Domain Multicast
###############################

def multicast_get(ctx, rev, domain_id):
    return domain_get(ctx, rev, domain_id)["multicast"]


###############################
# Bridge Domain Snooping
###############################

def snooping_get(ctx, rev, domain_id):
    return multicast_get(ctx, rev, domain_id)["snooping"]


###############################
# Bridge Domain Querier
###############################

def querier_get(ctx, rev, domain_id):
    return snooping_get(ctx, rev, domain_id)["querier"]


###############################
# Bridge Domain VLANs
###############################
def vlans_get(ctx, rev, domain_id):
    changed = False
    if rev == 'operational':
        operational_output = domain_get(ctx, rev, domain_id)
        if 'vlan' in operational_output:
            operational_output = operational_output["vlan"]
            changed = True
        else:
            operational_output = {}
        # Bridge VLAN config has to be patched with operational revision
        rev = 'applied'

    applied = domain_get(ctx, rev, domain_id)
    try:
        output = applied['vlan']
    except KeyError:
        raise NotFound

    if changed:
        return utils.fast_merge(output, operational_output)

    return output


###############################
# Bridge Domain VLAN
###############################

def vlan_get(ctx, rev, domain_id, vid):
    vlans = vlans_get(ctx, rev, domain_id)
    try:
        return vlans[vid]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# Bridge Domain VNIs
###############################

def vnis_get(ctx, rev, domain_id, vid):
    output = vlan_get(ctx, rev, domain_id, vid)
    try:
        return output["vni"]
    except KeyError:
        raise NotFound


###############################
# Bridge Domain VNI
###############################

def vni_get(ctx, rev, domain_id, vid, vni_id):
    vnis = vnis_get(ctx, rev, domain_id, vid)
    try:
        return vnis[vni_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# VNI Flooding
###############################

def vni_flooding_get(ctx, rev, domain_id, vid, vni_id):
    output = vni_get(ctx, rev, domain_id, vid, vni_id)
    try:
        return output["flooding"]
    except KeyError:
        raise NotFound


###############################
# VNI Flooding head-end-replication
###############################

def vni_flooding_hreps_get(ctx, rev, domain_id, vid, vni_id):
    output = vni_flooding_get(ctx, rev, domain_id, vid, vni_id)
    try:
        return output["head-end-replication"]
    except KeyError:
        raise NotFound


###############################
# VNI Flooding head-end-replication-id
###############################

def vni_flooding_hrep_get(ctx, rev, domain_id, vid, vni_id, hrep_id):

    hreps = vni_flooding_hreps_get(ctx, rev, domain_id, vid, vni_id)
    try:
        return hreps[hrep_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# VLAN PTP
###############################

def vlan_ptp_get(ctx, rev, domain_id, vid):
    if rev == 'operational':
        # PTP under bridge.vlan is purely a config concept.  Return the last
        # applied config as the operational state.
        rev = 'applied'

    return vlan_get(ctx, rev, domain_id, vid)["ptp"]


# ###############################
# # PTP Timers
# ###############################
#
# def vlan_ptp_timers_get(ctx, rev, domain_id, vid):
#    return vlan_ptp_get(ctx, rev, domain_id, vid)["timers"]
#
#
###############################
# VLAN Multicast
###############################

def vlan_multicast_get(ctx, rev, domain_id, vid):
    return vlan_get(ctx, rev, domain_id, vid)["multicast"]


###############################
# VLAN Snooping
###############################

def vlan_snooping_get(ctx, rev, domain_id, vid):
    return vlan_multicast_get(ctx, rev, domain_id, vid)["snooping"]


###############################
# VLAN Querier
###############################

def vlan_querier_get(ctx, rev, domain_id, vid):
    return vlan_snooping_get(ctx, rev, domain_id, vid)["querier"]


###############################
# Bridge Domain MAC Table
###############################
def mac_table_get(ctx, rev, domain_id):
    # MAC Table does not have a config side
    if rev != "operational":
        raise NotFound

    res = ctx.netlink_v1.getMacTable(domain_id)
    if not res:
        raise NotFound

    return res


###############################
# Bridge Domain Mdb
###############################
def mdb_get(ctx, rev, domain_id):
    return domain_get(ctx, rev, domain_id).get("mdb", {})


###############################
# Bridge Domain Router Ports
###############################
def rport_get(ctx, rev, domain_id):
    return domain_get(ctx, rev, domain_id).get("router-port", {})
