# Copyright (C) 2018-2021 Cumulus Networks, inc
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


from cue_cue_v1.actions import action_get_controller
from pydash import py_
from cue.exceptions import NotFound
from cue import utils


###############################
# Interfaces
###############################

def interfaces_get(ctx, rev):
    if rev != "operational":
        return ctx.config_v1.getInterfaces(rev)

    interfaces = ctx.netlink_v1.getInterfaces()
    lldps = ctx.lldp_v1.getNbrs()
    for intf in interfaces:
        if intf not in lldps:
            continue
        interfaces[intf]['lldp'] = lldps[intf]

    return interfaces


###############################
# Interface
###############################

def interface_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id)

    iface = ctx.netlink_v1.getInterface(interface_id)

    # Get FEC encoding
    fec = ctx.netlink_v1.getInterfaceLinkFec(interface_id)
    if fec and 'link' in iface:
        iface = py_.merge({'link': {'fec': fec}},
                          iface)

    # Depending on the type, we can fill in more from lldpctl
    # Skip fetching lldp info for logical interfaces
    if iface["type"] in ["eth", "swp", "bond"]:
        lldp = lldp_get(ctx, rev, interface_id, iface["type"])
        if lldp:
            iface['lldp'] = lldp

    return iface


def interface_pluggable_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)

    return ctx.netlink_v1.getInterfacePluggable(interface_id)


###############################
# Bridge
###############################

def bridge_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, "bridge")

    try:
        bridge = ctx.netlink_v1.getInterface(interface_id)["bridge"]

        # Hack!  Assume one bridge domain.
        stp = {"domain": {"br_default": {
            "stp": ctx.netlink_v1.getInterfaceBridgeStp(interface_id)
        }}}

        vlan = ctx.netlink_v1.getInterfaceBridgeVlan(interface_id)

        full_bridge = py_.merge({}, bridge, vlan, stp)
        return full_bridge
    except KeyError:
        raise NotFound


###############################
# Bridge Domains
###############################

def domains_get(ctx, rev, interface_id):
    return bridge_get(ctx, rev, interface_id)["domain"]


###############################
# Bridge Domain
###############################

def domain_get(ctx, rev, interface_id, domain_id):
    domains = domains_get(ctx, rev, interface_id)
    try:
        return domains[domain_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# Bridge domain STP
###############################

def stp_get(ctx, rev, interface_id, domain_id):
    return domain_get(ctx, rev, interface_id, domain_id)["stp"]


###############################
# Bridge Domain VLAN
###############################

def vlans_get(ctx, rev, interface_id, domain_id):
    try:
        return domain_get(ctx, rev, interface_id, domain_id)["vlan"]
    except KeyError:
        raise NotFound


###############################
# Bridge Domain VLAN
###############################

def vlan_get(ctx, rev, interface_id, domain_id, vid):
    vlans = vlans_get(ctx, rev, interface_id, domain_id)
    try:
        return vlans[vid]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# IP
###############################

def ip_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, "ip")

    return ctx.netlink_v1.getInterfaceIp(interface_id)


###############################
# IP Addresses
###############################

ip_address_path = ("ip", "address")


def ip_addresses_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *ip_address_path)

    ip = ctx.netlink_v1.getInterfaceIp(interface_id)

    # netlink_v1 always fills in "address" for us, even if it is empty.
    return ip["address"]


###############################
# IP Address
###############################

def ip_address_get(ctx, rev, interface_id, ip_prefix_id):
    addresses = ip_addresses_get(ctx, rev, interface_id)
    try:
        return addresses[ip_prefix_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# IP Neighbors
###############################

def ip_neighbors_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)

    return {
        "ipv4": ctx.netlink_v1.getInterfaceIpv4Neighbors(interface_id),
        "ipv6": ctx.netlink_v1.getInterfaceIpv6Neighbors(interface_id),
    }


def ipv4_neighbors_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)

    return ctx.netlink_v1.getInterfaceIpv4Neighbors(interface_id)


def ipv4_neighbor_get(ctx, rev, interface_id, neighbor_id):
    neighbors = ipv4_neighbors_get(ctx, rev, interface_id)
    try:
        return neighbors[neighbor_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


def ipv6_neighbors_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)

    return ctx.netlink_v1.getInterfaceIpv6Neighbors(interface_id)


def ipv6_neighbor_get(ctx, rev, interface_id, neighbor_id):
    neighbors = ipv6_neighbors_get(ctx, rev, interface_id)
    try:
        return neighbors[neighbor_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# VRR
###############################

def vrr_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, "ip", "vrr")

    return ctx.netlink_v1.getInterfaceVrr(interface_id)


###############################
# VRR Addresses
###############################

vrr_address_path = ("ip", "vrr", "address")


def vrr_addresses_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *vrr_address_path)

    vrr = ctx.netlink_v1.getInterfaceVrr(interface_id)
    try:
        return vrr["address"]
    except KeyError:
        return {}


###############################
# VRR Address
###############################

def vrr_address_get(ctx, rev, interface_id, ip_prefix_id):
    addresses = vrr_addresses_get(ctx, rev, interface_id)
    try:
        return addresses[ip_prefix_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# VRR State
###############################

vrr_state_path = ("ip", "vrr", "state")


def vrr_state_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *vrr_state_path)

    return ctx.netlink_v1.getInterfaceVrr(interface_id)["state"]


###############################
# IP Gateways
###############################

ip_gateway_path = ("ip", "gateway")


def ip_gateways_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *ip_gateway_path)

    # TODO CUE-1900
    return {}


###############################
# IP Gateway
###############################

def ip_gateway_get(ctx, rev, interface_id, ip_address_id):
    gateways = ip_gateways_get(ctx, rev, interface_id)
    try:
        return gateways[ip_address_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# IPv4
###############################

ipv4_path = ("ip", "ipv4")


def ipv4_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *ipv4_path)

    ip = ctx.netlink_v1.getInterfaceIp(interface_id)

    # TODO CUE-1928
    return ip.get('ipv4', {})


###############################
# IPv6
###############################

ipv6_path = ("ip", "ipv6")


def ipv6_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *ipv6_path)

    ip = ctx.netlink_v1.getInterfaceIp(interface_id)

    # TODO CUE-1928
    return ip.get('ipv6', {})


###############################
# IGMP
###############################

igmp_path = ("ip", "igmp")


def igmp_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *igmp_path)

    return {}


###############################
# IGMP Static groups
###############################

igmp_st_grps_path = ("ip", "igmp", "static-group")


def igmp_st_grps_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id,
                                          *igmp_st_grps_path)

    return {}


###############################
# IGMP Static group
###############################

def igmp_st_grp_get(ctx, rev, interface_id, static_group_id):
    static_rts = igmp_st_grps_get(ctx, rev, interface_id)
    try:
        return static_rts[static_group_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# VRRP
###############################

vrrp_path = ("ip", "vrrp")


def vrrp_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *vrrp_path)

    raise ctx.errors.OperationalNotSupported


###############################
# VRRP Virtual Routers
###############################

def vrrp_vrs_get(ctx, rev, interface_id):
    return vrrp_get(ctx, rev, interface_id)["virtual-router"]


###############################
# VRRP Virtual router
###############################

def vrrp_vr_get(ctx, rev, interface_id, virtual_router_id):
    vrs = vrrp_vrs_get(ctx, rev, interface_id)
    try:
        return vrs[virtual_router_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# VRRP Virtual router addresses
###############################

def vrrp_vr_addrs_get(ctx, rev, interface_id, virtual_router_id):
    return vrrp_vr_get(ctx, rev, interface_id, virtual_router_id)["address"]


###############################
# VRRP Virtual router address
###############################

def vrrp_vr_addr_get(ctx, rev, interface_id, virtual_router_id, ip_address_id):
    vr_addrs = vrrp_vr_addrs_get(ctx, rev, interface_id, virtual_router_id)
    try:
        return vr_addrs[ip_address_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# Link
###############################

def link_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, "link")

    iflink = {}

    # getInterface() fills in part of the link object.
    iface = ctx.netlink_v1.getInterface(interface_id)

    # Get FEC encoding
    fec = ctx.netlink_v1.getInterfaceLinkFec(interface_id)
    if fec:
        iflink['fec'] = fec

    # Depending on the type, we can fill in more from ifLinkSettings.
    if iface["type"] in ["eth", "swp", "bond"]:
        py_.merge(iflink, ctx.netlink_v1.getIfLinkSettings(interface_id))

    return py_.merge({}, iface["link"], iflink)


###############################
# Router
###############################
router_path = ("router",)


def router_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *router_path)

    # We will not show the operational state as of now.
    raise ctx.errors.OperationalNotSupported


###############################
# PBR
###############################

pbr_path = ('router', 'pbr')


def pbr_get(ctx, rev, interface_id):
    if rev != 'operational':
        return ctx.config_v1.getInterface(rev, interface_id, *pbr_path)

    # Operational data not supported yet
    raise ctx.errors.OperationalNotSupported


###############################
# OSPF
###############################

ospf_path = ('router', 'ospf')


def ospf_get(ctx, rev, interface_id):
    if rev != 'operational':
        return ctx.config_v1.getInterface(rev, interface_id, *ospf_path)

    # Operational data not supported yet
    raise ctx.errors.OperationalNotSupported


###############################
# OSPF/timers
###############################

def ospf_timers_get(ctx, rev, interface_id):
    return ospf_get(ctx, rev, interface_id)["timers"]


###############################
# OSPF/authentication
###############################

def ospf_authentication_get(ctx, rev, interface_id):
    return ospf_get(ctx, rev, interface_id)["authentication"]


###############################
# OSPF/bfd
###############################

def ospf_bfd_get(ctx, rev, interface_id):
    return ospf_get(ctx, rev, interface_id)["bfd"]


###############################
# PIM
###############################

pim_path = ('router', 'pim')


def pim_get(ctx, rev, interface_id):
    if rev != 'operational':
        return ctx.config_v1.getInterface(rev, interface_id, *pim_path)

    # Operational data not supported yet
    raise ctx.errors.OperationalNotSupported


###############################
# PIM/timers
###############################

def pim_timers_get(ctx, rev, interface_id):
    return pim_get(ctx, rev, interface_id)["timers"]


###############################
# PIM/bfd
###############################

def pim_bfd_get(ctx, rev, interface_id):
    return pim_get(ctx, rev, interface_id)["bfd"]


###############################
# PIM/address-family
###############################

def pim_af_get(ctx, rev, interface_id):
    return pim_get(ctx, rev, interface_id)["address-family"]


###############################
# PIM/address-family/ipv4-unicast
###############################

def pim_af_v4_unicast_get(ctx, rev, interface_id):
    return pim_af_get(ctx, rev, interface_id)["ipv4-unicast"]


###############################
# PIM/address-family/ipv4-unicast/allow-rp
###############################

def pim_af_v4_allow_rp_get(ctx, rev, interface_id):
    return pim_af_v4_unicast_get(ctx, rev, interface_id)["allow-rp"]


###############################
# Bond
###############################

bond_path = ("bond",)


def bond_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *bond_path)

    try:
        return ctx.netlink_v1.getInterface(interface_id)["bond"]
    except KeyError:
        raise NotFound


###############################
# Bond Members
###############################

bond_member_path = ("bond", "member")


def bond_members_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *bond_member_path)

    # netlink_v1 always fills in "member" for us, if we have a bond.
    return bond_get(ctx, rev, interface_id)["member"]


###############################
# Bond Member
###############################

def bond_member_get(ctx, rev, interface_id, member_id):
    members = bond_members_get(ctx, rev, interface_id)
    try:
        return members[member_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# Mlag
###############################

def mlag_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, "bond", "mlag")

    try:
        return ctx.netlink_v1.getInterface(interface_id)["bond"]["mlag"]
    except KeyError:
        raise NotFound


###############################
# Link Dot1x
###############################

link_dot1x_path = ("link", "dot1x")


def dot1x_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *link_dot1x_path)

    # TODO CUE-1903
    return {}


###############################
# Link State
###############################

link_state_path = ("link", "state")


def link_state_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, *link_state_path)

    return ctx.netlink_v1.getInterface(interface_id)["link"]["state"]


###############################
# Link traffic engineering
###############################


def te_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)

    return ctx.frr_v1.getInterface(interface_id)["link"]["traffic-engineering"]


###############################
# Link flags
###############################


def link_flags_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)

    return ctx.frr_v1.getInterface(interface_id)["link"]["flag"]


###############################
# LLDP
###############################

def lldp_get(ctx, rev, interface_id, if_type=None):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)

    # getNbr() will always return, even if the interface doesn't support LLDP.
    # So, we need to check the type before continuing.
    if not if_type:
        if_type = ctx.netlink_v1.getInterface(interface_id)["type"]

    if if_type in ["swp", "eth", "bond"]:
        return ctx.lldp_v1.getNbr(interface_id)
    else:
        # TODO: Consider something like NotSupported?
        raise ctx.errors.ItemDoesNotExist


def lldp_neighbors_get(ctx, rev, interface_id):
    # We always get a neighbor object from lldp_v1.
    return lldp_get(ctx, rev, interface_id)


def lldp_neighbor_get(ctx, rev, interface_id, neighbor_id):
    neighbors = lldp_neighbors_get(ctx, rev, interface_id)
    try:
        return neighbors[neighbor_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


def lldp_neighbor_bridge_get(ctx, rev, interface_id, neighbor_id):
    neighbor = lldp_neighbor_get(ctx, rev, interface_id, neighbor_id)
    try:
        return neighbor["bridge"]
    except KeyError:
        return {}


def lldp_neighbor_bridge_vlans_get(ctx, rev, interface_id, neighbor_id):
    dot1tlv = lldp_neighbor_bridge_get(ctx, rev, interface_id, neighbor_id)
    try:
        return dot1tlv["vlan"]
    except KeyError:
        return {}


def lldp_neighbor_bridge_vlan_get(ctx, rev, interface_id, neighbor_id, vid):
    vlans = lldp_neighbor_bridge_vlans_get(ctx, rev, interface_id, neighbor_id)
    try:
        return vlans[vid]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# PTP
###############################

def ptp_get(ctx, rev, interface_id):
    if rev != "operational":
        iface = interface_get(ctx, rev, interface_id)
        return iface["ptp"]

    if_ptps = ctx.linuxptp_v1.getCuev1Ptp4lConfIfPtps()

    try:
        ptp_iface = if_ptps[interface_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist

    if_stats = ctx.linuxptp_v1.getCuev1Ptp4lCliStatsIfPtp(interface_id)

    return utils.fast_merge(
        ptp_iface["ptp"],
        ctx.linuxptp_v1.getCuev1Ptp4lCliPortIfPtp(interface_id),
        {"counters": if_stats},
    )


###############################
# PTP timers
###############################

def ptp_timers_get(ctx, rev, interface_id):
    ptp = ptp_get(ctx, rev, interface_id)
    return ptp["timers"]


###############################
# PTP stats
###############################

def ptp_counters_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.linuxptp_v1.getCuev1Ptp4lCliStatsIfPtp(interface_id)


###############################
# Stats
###############################

def stats_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    swp = ctx.netlink_v1.getInterface(interface_id)
    stats = py_.get(swp, 'link.stats')
    if stats:
        return stats
    else:
        return {}


def qos_stats_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    stats = ctx.qos_v1.getInterfaceQosStats(interface_id)
    return stats


def qos_port_stats_get(ctx, rev, interface_id):
    qos_stats = qos_stats_get(ctx, rev, interface_id)
    return qos_stats['port-stats']


def qos_port_stats_rx_get(ctx, rev, interface_id):
    port_stats = qos_port_stats_get(ctx, rev, interface_id)
    return port_stats['rx-stats']


def qos_port_stats_tx_get(ctx, rev, interface_id):
    port_stats = qos_port_stats_get(ctx, rev, interface_id)
    return port_stats['tx-stats']


def qos_egr_queue_stats_get(ctx, rev, interface_id):
    qos_stats = qos_stats_get(ctx, rev, interface_id)
    return qos_stats['egress-queue-stats']


def qos_ing_buffer_stats_get(ctx, rev, interface_id):
    qos_stats = qos_stats_get(ctx, rev, interface_id)
    return qos_stats['ingress-buffer-stats']


def qos_pfc_stats_get(ctx, rev, interface_id):
    qos_stats = qos_stats_get(ctx, rev, interface_id)
    return qos_stats['pfc-stats']


@action_get_controller()
def roce_stats_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    stats = ctx.qos_v1.getInterfaceRoceStats(interface_id)
    return stats


def roce_status_tc_map_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    status = ctx.qos_v1.getInterfaceRoceStatusTcMap(interface_id)
    return status


def roce_status_prio_map_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    status = ctx.qos_v1.getInterfaceRoceStatusPrioMap(interface_id)
    return status


def roce_status_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    status = ctx.qos_v1.getInterfaceRoceStatus(interface_id)
    return status


###############################
# RoCE system status
###############################

def roce_system_pool_get(ctx, rev, interface_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    pool_info = ctx.qos_v1.getRoceInterfacePools(interface_id)
    return pool_info


###############################
# EVPN
###############################

def evpn_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(rev, interface_id, "evpn")

    raise ctx.errors.OperationalNotSupported


###############################
# EVPN MH
###############################

def evpn_mh_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(
            rev, interface_id, "evpn", "multihoming")

    return ctx.frr_v1.getInterface(interface_id)["evpn"]["multihoming"]


###############################
# EVPN MH Segment
###############################

def evpn_mh_segment_get(ctx, rev, interface_id):
    if rev != "operational":
        return ctx.config_v1.getInterface(
            rev, interface_id, "evpn", "multihoming", "segment")

    return ctx.frr_v1.getInterface(interface_id)["evpn"]["multihoming"][
        "segment"]


###############################
# ACLs
###############################

def acls_get(ctx, rev, interface_id):
    if rev == 'operational':
        # Add the operational state when FUnit is ready
        rev = 'applied'

    res = ctx.config_v1.getInterface(rev, interface_id)
    return res['acl']


###############################
# ACL
###############################

def acl_get(ctx, rev, interface_id, acl_id):
    acls = acls_get(ctx, rev, interface_id)
    try:
        return acls[acl_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# ACL inbound
###############################

def acl_inbound_get(ctx, rev, interface_id, acl_id):
    return acl_get(ctx, rev, interface_id, acl_id)["inbound"]


###############################
# ACL inbound control plane
###############################

def acl_in_control_plane_get(ctx, rev, interface_id, acl_id):
    return acl_inbound_get(ctx, rev, interface_id, acl_id)["control-plane"]


###############################
# ACL outbound
###############################

def acl_outbound_get(ctx, rev, interface_id, acl_id):
    return acl_get(ctx, rev, interface_id, acl_id)["outbound"]


###############################
# ACL inbound control plane
###############################

def acl_out_control_plane_get(ctx, rev, interface_id, acl_id):
    return acl_outbound_get(ctx, rev, interface_id, acl_id)["control-plane"]
