# Copyright (C) 2018-2021 Cumulus Networks, inc
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


import re
from cue.utils import cue_format_checker, clean_none
from cue.utils.format_checker import patterns
from cue.morpher import (
    Morpher,
    expand_number_ids,
    expand_name_ids,
    expand_feature_on,
    expand_side_feature_on,
    make_state_preparer,
)
from pydash import py_


###################
# Morphers
###################
root_patch_expand = Morpher()
"""
Morpher that takes a rooted patch directly from the client and expands globs,
derives fields, and does any other magic we want done on a set/PATCH.  The
result is an expanded patch.  The patch is morphed in place.
"""


root_patch_prepare = Morpher()
"""
Morpher that takes a copy of the expanded patch (generated by
root_patch_expand) and creates a prepare patch.  The prepare patch will be
merged with the config before the expanded patch.  The prepare patch is morphed
in place.

We start with the expanded patch so that, in the prepare morpher, we don't need
to worry about globs and derivations.  Those will already be resolved.
"""


@root_patch_expand.add_step("platform.hardware.component.{component_id}")
def _derive_module_type(body, component_id):
    if clean_none(body) is None:
        return body

    if component_id[:6] == "module":
        body["type"] = "linecard"

    return body


###############################
# Interface
###############################

@root_patch_expand.add_step("interface.{interface_id}")
def _derive_if_type_and_base_interface(body, interface_id):
    # If it's empty except nulls, don't populate.
    if clean_none(body) is None:
        return body

    # Derive the type if we can.
    if "." in interface_id:
        # Smells like a sub interface.  This check should improve with
        # CUE-2095.
        body["type"] = "sub"
        tokens = interface_id.split('.')
        body["base-interface"] = '.'.join(tokens[0:-1])
        body["vlan"] = int(tokens[-1])
    elif interface_id[:3] == "swp":
        body["type"] = "swp"
    elif interface_id[:3] == "eth":
        body["type"] = "eth"
    elif interface_id[:4] == "bond":
        body["type"] = "bond"
    elif interface_id[:2] == "lo":
        body["type"] = "loopback"
    elif interface_id[:4] == "vlan":
        body["type"] = "svi"
        m = re.match(patterns['svi'], interface_id)
        if m:
            # If the suffix of "vlan" is an integer, set the vlan
            try:
                body["vlan"] = int(m.group(1))
            except ValueError:
                pass
    elif interface_id[:8] == "peerlink":
        body["type"] = "peerlink"

    return body


@root_patch_expand.add_step("interface.{_}.bridge.domain.{_}.vlan")
def expand_interface_bridge_vlan(collection):
    # Don't try to expand "all".  It's not a glob.
    if "all" in collection:
        return

    expand_number_ids(collection)


# If the link state is set, clear the current state
root_patch_prepare.add_step("interface.{_}.link")(make_state_preparer())

# If the vrr state is set, clear the current state
root_patch_prepare.add_step("interface.{_}.ip.vrr")(make_state_preparer())

# If igmp is touched, enable it.
root_patch_expand.add_step("interface.{_}.ip.igmp")(expand_feature_on)

# If vrr is touched, enable it.
root_patch_expand.add_step("interface.{_}.ip.vrr")(expand_feature_on)

# Expand interface globs
root_patch_expand.add_step("interface")(expand_name_ids)

# Expand bond member globs
root_patch_expand.add_step("interface.{_}.bond.member")(expand_name_ids)

# If interface.ip.vrr is touched, enable it globally.
root_patch_expand.add_step(
    "interface.{_}.ip.vrr", side_feature="router.vrr"
)(expand_side_feature_on)

###############################
# Interface EVPN
###############################

# If evpn.multihoming.segment is touched, enable it
root_patch_expand.add_step(
    "interface.{_}.evpn.multihoming.segment")(expand_feature_on)


###############################
# Interface PTP
###############################

# If interface.service.ptp is touched, enable it
root_patch_expand.add_step("interface.{_}.ptp")(expand_feature_on)


###############################
# Interface MLAG
###############################

# If interface.service.bond.mlag is touched, enable it
root_patch_expand.add_step("interface.{_}.bond.mlag")(expand_feature_on)


###############################
# Interface OSPFv2
###############################

# If router.ospf is touched, enable it
root_patch_expand.add_step(
    "interface.{_}.router.ospf")(expand_feature_on)

# If router.ospf.authentication is touched, enable it
root_patch_expand.add_step(
    "interface.{_}.router.ospf.authentication")(expand_feature_on)

# If router.ospf.bfd is touched, enable it
root_patch_expand.add_step(
    "interface.{_}.router.ospf.bfd")(expand_feature_on)


def _ospf_dead_interval_prepare(timers):
    """
    Clear hello-interval or hello-multiplier appropriately.
    """
    dead = timers.get("dead-interval")

    if dead:
        # If we're setting dead-interval to minimal, clear hello-interval
        if "minimal" == dead:
            timers["hello-interval"] = None
        # Otherwise, clear hello-multiplier
        else:
            timers["hello-multiplier"] = None


root_patch_prepare.add_step(
    "interface.{_}.router.ospf.timers")(_ospf_dead_interval_prepare)


###############################
# Interface PIM
###############################

# If router.pim is touched, enable it
root_patch_expand.add_step(
    "interface.{_}.router.pim")(expand_feature_on)

# If interface.router.pim is touched, enable it globally.
root_patch_expand.add_step(
    "interface.{_}.router.pim", side_feature="router.pim"
)(expand_side_feature_on)

# If router.pim's bfd is touched, enable it
root_patch_expand.add_step(
    "interface.{_}.router.pim.bfd")(expand_feature_on)

# If pim's allow-rp is touched, enable it
root_patch_expand.add_step(
    "interface.{_}.router.pim.address-family.ipv4-unicast.allow-rp"
)(expand_feature_on)

###############################
# Interface VRRP
###############################

# If vrrp is touched, enable it
root_patch_expand.add_step("interface.{_}.ip.vrrp")(expand_feature_on)

# If interface.ip.vrrp is touched, enable it globally.
root_patch_expand.add_step(
    "interface.{_}.ip.vrrp", side_feature="router.vrrp"
)(expand_side_feature_on)

###############################
# Bridge
###############################

# Expand vlan id globs
root_patch_expand.add_step("bridge.domain.{_}.vlan")(expand_number_ids)

# If the stp state is set, clear the current state
root_patch_prepare.add_step("bridge.domain.{_}.stp")(make_state_preparer())


def _flooding_prepare(vni):
    flooding = vni.get("flooding", None)
    if flooding:
        if "head-end-replication" in flooding:
            flooding["multicast-group"] = None
        elif "multicast-group" in flooding:
            flooding["head-end-replication"] = None
        # Always return None, since we're mutating the input.
        return None


# If flooding is set, clear the other value.  We can't use the state_preparer()
# because we don't want to lose current settings if the value hasn't changed.
root_patch_prepare.add_step(
    "bridge.domain.{_}.vlan.{_}.vni.{_}")(_flooding_prepare)

# If flooding is touched, enable it
root_patch_expand.add_step(
    "bridge.domain.{_}.vlan.{_}.vni.{_}.flooding"
)(expand_feature_on)

# Touching a VNI enables VXLAN
root_patch_expand.add_step(
    "bridge.domain.{_}.vlan.{_}.vni",
    side_feature="nve.vxlan"
)(expand_side_feature_on)

# If bridge.domain.multicast.snooping is touched, enable it
root_patch_expand.add_step(
    "bridge.domain.{_}.multicast.snooping"
)(expand_feature_on)

# If bridge.domain.multicast.snooping.querier is touched, enable it
root_patch_expand.add_step(
    "bridge.domain.{_}.multicast.snooping.querier"
)(expand_feature_on)


###############################
# MLAG
###############################

# If MLAG is enabled, create peerlink and peerlink.4094 if they don't exist
@root_patch_expand.add_step("mlag", interfaces="interface")
def handle_mlag(mlag, interfaces):
    """
    Prepare MLAG - Ensure all MLAG objects are present if 'mlag' is.

    If 'mlag' is configured, configure 'interface.peerlink' and
    'interface.peerlink.4094' if they aren't already present.
    """
    # If it's empty except nulls, don't populate.
    if clean_none(mlag) is None:
        return

    py_.set_(interfaces, ["peerlink", "type"], "peerlink")
    py_.set_(interfaces, ["peerlink.4094", "type"], "sub")
    py_.set_(interfaces, ["peerlink.4094", "base-interface"],
             "peerlink")
    py_.set_(interfaces, ["peerlink.4094", "vlan"],
             4094)


# If mlag is touched, enable it
root_patch_expand.add_step("mlag")(expand_feature_on)


###############################
# Router
###############################

# If bgp is touched, enable it
root_patch_expand.add_step("router.bgp")(expand_feature_on)

# If pbr is touched, enable it
root_patch_expand.add_step("router.pbr")(expand_feature_on)

# If ospf is touched, enable it
root_patch_expand.add_step("router.ospf")(expand_feature_on)

# If pim is touched, enable it
root_patch_expand.add_step("router.pim")(expand_feature_on)

# If igmp is touched, enable it
root_patch_expand.add_step("router.igmp")(expand_feature_on)

# If vrrp is touched, enable it
root_patch_expand.add_step("router.vrrp")(expand_feature_on)

# If vrr is touched, enable it
root_patch_expand.add_step("router.vrr")(expand_feature_on)

###############################
# System
###############################

# If port-mirroring span is touched, enable it
root_patch_expand.add_step("system.port-mirror.session.{_}.span")(
    expand_feature_on)

root_patch_expand.add_step("system.port-mirror.session.{_}.erspan")(
    expand_feature_on)

root_patch_expand.add_step("system.port-mirror.session.{_}.span.truncate")(
    expand_feature_on)

root_patch_expand.add_step("system.port-mirror.session.{_}.erspan")(
    expand_feature_on)

root_patch_expand.add_step("system.port-mirror.session.{_}.erspan.truncate")(
    expand_feature_on)

###############################
# VRF Router
###############################


def _derive_address_family(body, route_id):
    """
    Utility to set route address-family based on the route id.
    """
    # If it's empty except nulls, don't populate.
    if clean_none(body) is None:
        return body

    # Derive the address-family if we can.
    if cue_format_checker.conforms(route_id, "ipv4-prefix"):
        body["address-family"] = "ipv4-unicast"
    elif cue_format_checker.conforms(route_id, "ipv6-prefix"):
        body["address-family"] = "ipv6-unicast"
    elif route_id == "default":
        body["address-family"] = "ipv4-unicast"
    elif route_id == "default6":
        body["address-family"] = "ipv6-unicast"

    return body


root_patch_expand.add_step(
    "vrf.{_}.router.static.{route_id}"
)(_derive_address_family)


def _derive_via_type(body, via_id):
    """
    Utility to set via types based on the via id.
    """
    # If it's empty except nulls, don't populate.
    if clean_none(body) is None:
        return body

    # Derive the type if we can.
    if via_id == "blackhole":
        body["type"] = "blackhole"
    elif via_id == "reject":
        body["type"] = "reject"
    elif cue_format_checker.conforms(via_id, "ipv4"):
        body["type"] = "ipv4-address"
    elif cue_format_checker.conforms(via_id, "ipv6"):
        body["type"] = "ipv6-address"
    # Look for interface last because it's the most permissive pattern
    elif cue_format_checker.conforms(via_id, "interface-name"):
        body["type"] = "interface"

    return body


# Expand via types
root_patch_expand.add_step(
    "vrf.{_}.router.static.{_}.via.{via_id}"
)(_derive_via_type)

# Expand via types under paths
root_patch_expand.add_step(
    "vrf.{_}.router.static.{_}.distance.{_}.via.{via_id}"
)(_derive_via_type)


###############################
# Vxlan
###############################

# If vxlan is touched, enable it
root_patch_expand.add_step("nve.vxlan")(expand_feature_on)

# If flooding is set, clear the other value.  We can't use the state_preparer()
# because we don't want to lose current settings if the value hasn't changed.
root_patch_prepare.add_step("nve.vxlan")(_flooding_prepare)

# If flooding is touched, enable it
root_patch_expand.add_step("nve.vxlan.flooding")(expand_feature_on)


###############################
# EVPN
###############################

# If evpn is touched, enable it
root_patch_expand.add_step("evpn")(expand_feature_on)

# Enable VXLAN for EVPN.
root_patch_expand.add_step(
    "evpn", side_feature="nve.vxlan"
)(expand_side_feature_on)

# Enable BGP for EVPN.
root_patch_expand.add_step(
    "evpn", side_feature="router.bgp"
)(expand_side_feature_on)

# If evpn.dad is touched, enable it
root_patch_expand.add_step("evpn.dad")(expand_feature_on)

# If evpn.multihoming is touched, enable it
root_patch_expand.add_step("evpn.multihoming")(expand_feature_on)


###############################
# VRF evpn
###############################

# If vrf.evpn is touched, enable it
root_patch_expand.add_step("vrf.{_}.evpn")(expand_feature_on)


###############################
# VRF OSPFv2
###############################

# If vrf.router.ospf is touched, enable it
root_patch_expand.add_step("vrf.{_}.router.ospf")(expand_feature_on)

# If vrf.router.ospf is touched, enable it globally.
root_patch_expand.add_step(
    "vrf.{_}.router.ospf", side_feature="router.ospf"
)(expand_side_feature_on)

# If ospf redistribute is touched, enable it.  There are a few different types.
# Each is enabled independently.
root_patch_expand.add_step(
    "vrf.{_}.router.ospf.redistribute.{_}"
)(expand_feature_on)


###############################
# VRF PIM
###############################

# If vrf.router.pim is touched, enable it
root_patch_expand.add_step("vrf.{_}.router.pim")(expand_feature_on)

# If pim ecmp is touched, enable it.
root_patch_expand.add_step("vrf.{_}.router.pim.ecmp")(expand_feature_on)


###############################
# VRF BGP
###############################

# If vrf.router.bgp is touched, enable it
root_patch_expand.add_step("vrf.{_}.router.bgp")(expand_feature_on)

# If vrf.router.bgp is touched, enable it globally.
root_patch_expand.add_step(
    "vrf.{_}.router.bgp", side_feature="router.bgp"
)(expand_side_feature_on)

# If an vrf.bgp.address-family is touched, enable it.  There are a few
# different families.  Each is enabled independently.
root_patch_expand.add_step(
    "vrf.{_}.router.bgp.address-family.{_}"
)(expand_feature_on)

# If an vrf.bgp.address-family.route-export.to-evpn is touched, enable it.
# There are a few different families.  Each is enabled independently.
root_patch_expand.add_step(
    "vrf.{_}.router.bgp.address-family.{_}.route-export.to-evpn"
)(expand_feature_on)

# If an vrf.bgp.address-family.route-import.from-vrf is touched, enable it.
# There are a few different families.  Each is enabled independently.
root_patch_expand.add_step(
    "vrf.{_}.router.bgp.address-family.{_}.route-import.from-vrf"
)(expand_feature_on)

# If bgp redistribute is touched, enable it.  There are a few different types.
# Each is enabled independently.
root_patch_expand.add_step(
    "vrf.{_}.router.bgp.address-family.{_}.redistribute.{_}"
)(expand_feature_on)


###############################
# VRF BGP Peer
###############################

# If an vrf.bgp.neighbor.address-family is touched, enable it, but this doesn't
# apply to ipv4-unicast because it's always on.
root_patch_expand.add_step(
    "vrf.{_}.router.bgp.neighbor.{_}.address-family.ipv6-unicast"
)(expand_feature_on)

root_patch_expand.add_step(
    "vrf.{_}.router.bgp.neighbor.{_}.address-family.l2vpn-evpn"
)(expand_feature_on)


# If vrf.router.bgp.peer.address-family is touched, enable it globally.
def _expand_vrf_side_feature_on(body, vrf_id, side_feature):
    expand_side_feature_on(body, side_feature)


root_patch_expand.add_step(
    "vrf.{vrf_id}.router.bgp.neighbor.{_}.address-family.l2vpn-evpn",
    side_feature="vrf.{vrf_id}.router.bgp.address-family.l2vpn-evpn"
)(_expand_vrf_side_feature_on)

root_patch_expand.add_step(
    "vrf.{vrf_id}.router.bgp.neighbor.{_}.address-family.ipv6-unicast",
    side_feature="vrf.{vrf_id}.router.bgp.address-family.ipv6-unicast"
)(_expand_vrf_side_feature_on)

# If bgp vrf.bgp.address-family.conditional-advertise is touched, enable it.
# It is enabled ipv4 and ipv6 unicast address-family under neighbor and
# peer-group. Each is enabled independently.
root_patch_expand.add_step(
  "vrf.{_}.router.bgp.neighbor.{_}.address-family.{_}.conditional-advertise"
)(expand_feature_on)

root_patch_expand.add_step(
   "vrf.{_}.router.bgp.peer-group.{_}.address-family.{_}.conditional-advertise"
)(expand_feature_on)


def _derive_neighbor_type(body, neighbor_id, interfaces):
    """
    Utility to set neighbor types based on the neighbor id.
    """
    # If it's empty except nulls, don't populate.
    if clean_none(body) is None:
        return None

    if cue_format_checker.conforms(neighbor_id, "ipv4"):
        body["type"] = "numbered"
    elif cue_format_checker.conforms(neighbor_id, "ipv6"):
        body["type"] = "numbered"
    # When in doubt, assume neighbor_id is an interface, which makes it an
    # unnumbered neighbor.
    else:
        body["type"] = "unnumbered"

        # Add BGP neighbor "swp" interface to interfaces
        if neighbor_id[:3] == "swp" and "." not in neighbor_id:
            py_.set_(interfaces, [neighbor_id, "type"], "swp")

    return body


root_patch_expand.add_step(
    "vrf.{_}.router.bgp.neighbor.{neighbor_id}", interfaces="interface"
)(_derive_neighbor_type)


###############################
# QOS
###############################

root_patch_expand.add_step(
    "qos.roce"
)(expand_feature_on)


###############################
# Service
###############################

# If a ptp instance is touched, enable it.
root_patch_expand.add_step("service.ptp.{_}")(expand_feature_on)

###############################
# Port-mirror
###############################

root_patch_expand.add_step(
    "system.port-mirror.session.{_}.span")(expand_feature_on)
root_patch_expand.add_step(
    "system.port-mirror.session.{_}.erspan")(expand_feature_on)
