#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of Nvidia Corporation
# and its affiliates (the "Company") and all right, title, and interest
# in and to the software product, including all associated intellectual
# property rights, are and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
# Copyright (C) 2021 Cumulus Networks, inc.
# All Rights reserved.


from cue import utils, exceptions


###############################
# PTPs
###############################

def ptps_get(ctx, rev):
    if rev != "operational":
        return ctx.config_v1.getPtps(rev)

    ptp = ctx.linuxptp_v1.getCuev1Ptp4lConfPtp()
    if not ptp:
        # No instances
        return {}

    return {
        "1": utils.fast_merge(ptp, ctx.linuxptp_v1.getCuev1Ptp4lServicePtp())
    }


###############################
# PTP instance
###############################

def ptp_get(ctx, rev, instance_id):
    if rev != "operational":
        return ctx.config_v1.getPtp(rev, instance_id)

    ptps = ptps_get(ctx, rev)

    try:
        return ptps[instance_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# PTP acceptable masters
###############################

def ptp_acc_masters_get(ctx, rev, instance_id):
    if rev != "operational":
        ptp = ctx.config_v1.getPtp(rev, instance_id)
        return ptp["acceptable-master"]

    ptp = ctx.linuxptp_v1.getCuev1Ptp4lConfPtp()
    if not ptp:
        # Not configured
        raise exceptions.NotFound

    return ptp["acceptable-master"]


###############################
# PTP acceptable master
###############################

def ptp_acc_master_get(ctx, rev, instance_id, clock_id):
    masters = ptp_acc_masters_get(ctx, rev, instance_id)
    try:
        return masters[clock_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# PTP monitor
###############################

def ptp_monitor_get(ctx, rev, instance_id):
    if rev != "operational":
        ptp = ctx.config_v1.getPtp(rev, instance_id)
        return ptp["monitor"]

    ptp = ctx.linuxptp_v1.getCuev1Ptp4lConfPtp()
    if not ptp:
        # Not configured
        raise exceptions.NotFound

    return ptp["monitor"]


###############################
# PTP timestamp logs
###############################
def ptp_ts_logs_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.linuxptp_v1.getCuev1Ptp4lCliTsLogs()


###############################
# PTP violations
###############################

def ptp_violations_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.linuxptp_v1.getCuev1Ptp4lCliViolations()


###############################
# PTP violations log
###############################

def ptp_vlog_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)

    # We will not show the operational state as of now.
    raise ctx.errors.OperationalNotSupported


###############################
# PTP acceptable master violations
###############################

def ptp_vlog_ams_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.linuxptp_v1.getCuev1Ptp4lCliAmVioLog()


###############################
# PTP forced master violations
###############################

def ptp_vlog_fms_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.linuxptp_v1.getCuev1Ptp4lCliFmVioLog()


###############################
# PTP violations max offsets log
###############################

def ptp_vlog_max_offsets_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.linuxptp_v1.getCuev1Ptp4lCliMaxOffsetVioLog()


###############################
# PTP violations min offsets log
###############################

def ptp_vlog_min_offsets_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.linuxptp_v1.getCuev1Ptp4lCliMinOffsetVioLog()


###############################
# PTP violations path delays log
###############################

def ptp_vlog_path_delays_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.linuxptp_v1.getCuev1Ptp4lCliMpdVioLog()


###############################
# PTP current status
###############################

def ptp_current_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.linuxptp_v1.getCuev1Ptp4lCliCurrentPtpCurrent()


###############################
# PTP clock quality status
###############################

def ptp_clock_quality_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.linuxptp_v1.getCuev1Ptp4lCliDefaultPtpClockQuality()


###############################
# PTP parent status
###############################

def ptp_parent_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.linuxptp_v1.getCuev1Ptp4lCliParentPtpParent()


###############################
# PTP parent clock status
###############################

def ptp_parent_gm_clock_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    parent = ctx.linuxptp_v1.getCuev1Ptp4lCliParentPtpParent()
    return parent["grandmaster-clock-quality"]


###############################
# PTP time status
###############################

def ptp_time_get(ctx, rev, instance_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.linuxptp_v1.getCuev1Ptp4lCliTimePtpTimeProps()
