# Copyright (C) 2018-2020 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


from cue import utils


###############################
# System
###############################

def system_get(ctx, rev):
    if rev != "operational":
        return ctx.config_v1.getSystem(rev)

    response = {}
    response['hostname'] = ctx.env_v1.getHostname()
    response['timezone'] = ctx.env_v1.getTimezone()
    response['build'] = ctx.env_v1.getSystemBuild()
    response['uptime'] = ctx.env_v1.getSystemUptime()
    return response


def system_message_get(ctx, rev):
    if rev != "operational":
        sys = system_get(ctx, rev)
        return sys.get('message', {})

    # Fetch the operational value from the FU
    message = {}
    message['pre-login'] = ctx.env_v1.getPreloginMessage()
    message['post-login'] = ctx.env_v1.getPostloginMessage()
    return message


###############################
# System config
###############################

def system_config_get(ctx, rev):
    if rev == 'operational':
        # system config does not have a operational revision
        rev = 'applied'
    return ctx.config_v1.getSystemConfig(rev)


###############################
# Config apply
###############################

def config_apply_get(ctx, rev):
    if rev == 'operational':
        # System config does not have a operational revision
        rev = 'applied'
    return ctx.config_v1.getSystemConfigApply(rev)


###############################
# Apply ignores
###############################

def apply_ignores_get(ctx, rev):
    return config_apply_get(ctx, rev)["ignore"]


###############################
# Apply ignore
###############################

def apply_ignore_get(ctx, rev, ignore_id):
    ignores = apply_ignores_get(ctx, rev)
    try:
        return ignores[ignore_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# Snippets
###############################

def snippets_get(ctx, rev):
    if rev == 'operational':
        # Snippets do not have a operational revision
        rev = 'applied'
    return ctx.config_v1.getSnippets(rev)


###############################
# Global
###############################

def global_get(ctx, rev):
    system_mac = None

    if rev == 'operational':
        system_mac = ctx.platform_v1.getHardware().get('system-mac')

        # Can only get global applied config, not operational
        rev = 'applied'

    global_config = ctx.config_v1.getGlobal(rev)
    # If 'system-mac' is set to 'auto' we'll try to get it from eeprom
    if system_mac and global_config['system-mac'] == 'auto':
        global_config['system-mac'] = system_mac

    return global_config


###############################
# Reserved
###############################

def reserved_get(ctx, rev):
    if rev == 'operational':
        # Can only get reserved applied config, not operational
        rev = 'applied'

    return ctx.config_v1.getReserved(rev)


###############################
# Routing-table
###############################

def routing_table_get(ctx, rev):
    return reserved_get(ctx, rev)["routing-table"]


###############################
# PBR
###############################

def pbr_get(ctx, rev):
    return routing_table_get(ctx, rev)["pbr"]


###############################
# vlan
###############################

def vlan_get(ctx, rev):
    return reserved_get(ctx, rev)["vlan"]


###############################
# L3 VNI VLAN range
###############################

def l3vni_vlan_get(ctx, rev):
    return vlan_get(ctx, rev)["l3-vni-vlan"]


###############################
# ZTP
###############################

def ztp_script_get(ctx, rev):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.env_v1.getZtpScript()


def ztp_status_get(ctx, rev):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.env_v1.getZtpStatus()


def ztp_get(ctx, rev):
    ztp = {}
    ztp['status'] = ztp_status_get(ctx, rev)
    ztp['script'] = ztp_script_get(ctx, rev)
    return ztp


###############################
# Reboot reason
###############################

def reboot_get(ctx, rev):
    return {
        "reason": ctx.cue_v1.getRebootReason(rev),
        "history": ctx.cue_v1.getRebootHistory(rev),
    }


def reboot_reason_get(ctx, rev):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return utils.fast_merge(
        {
            # Default any non-existent keys to 'unknown'
            key: 'unknown'
            for key in ['gentime', 'reason', 'user']
        },
        ctx.platform_v1.getRebootReason()
    )


def reboot_history_get(ctx, rev):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.platform_v1.getRebootHistory()
