# Copyright (C) 2018-2021 Cumulus Networks, inc
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


from cue.exceptions import NotFound
from pydash import py_


###############################
# Vrfs
###############################
def vrfs_get(ctx, rev):
    if rev != "operational":
        return ctx.config_v1.getVrfs(rev)

    vrfs = py_.clone_deep(ctx.netlink_v1.getVrfs())

    # Fill in an empty router object for now.  Later, we'll start calling
    # FUnits to get real data.
    for vrf in vrfs.values():
        vrf["router"] = {}

    return vrfs


###############################
# Vrf
###############################
def vrf_get(ctx, rev, vrf_id):
    if rev != "operational":
        return ctx.config_v1.getVrf(rev, vrf_id)

    vrf = py_.clone_deep(ctx.netlink_v1.getVrf(vrf_id))

    # Fill in an empty router object for now.  Later, we'll start calling
    # FUnits to get real data.
    vrf["router"] = {}

    evpn_vni = ctx.frr_v1.getVrfVni(vrf_id)
    if evpn_vni and 'evpn' not in vrf:
        vrf["evpn"] = {}
        for vnik, _vniv in evpn_vni['vni'].items():
            vrf['evpn']['vni'] = {vnik: {}}

    return vrf


###############################
# Loopback
###############################
def loopback_get(ctx, rev, vrf_id):
    return vrf_get(ctx, rev, vrf_id)["loopback"]


###############################
# Router
###############################
def router_get(ctx, rev, vrf_id):
    return vrf_get(ctx, rev, vrf_id)["router"]


###############################
# BGP
###############################

def bgp_get(ctx, rev, vrf_id):
    if rev != "operational":
        return router_get(ctx, rev, vrf_id)["bgp"]

    # We will not show the operational state as of now.
    raise ctx.errors.OperationalNotSupported


###############################
# BGP/AddressFamily
###############################

def bgp_addr_family_get(ctx, rev, vrf_id):
    return bgp_get(ctx, rev, vrf_id)["address-family"]


###############################
# BGP/AddressFamily/ipv4-unicast
###############################

def addr_family_v4_ucast_get(ctx, rev, vrf_id):
    return bgp_addr_family_get(ctx, rev, vrf_id)["ipv4-unicast"]


###############################
# BGP/AddressFamily/ipv4-unicast/redistribute
###############################

def route_redistribution_get(ctx, rev, vrf_id):
    return addr_family_v4_ucast_get(ctx, rev, vrf_id)["redistribute"]


##############################################
# BGP/AddressFamily/ipv4-unicast/route-export
##############################################

def route_export_get(ctx, rev, vrf_id):
    return addr_family_v4_ucast_get(ctx, rev, vrf_id)["route-export"]


################################################
# BGP/AddressFamily/ipv4-unicast/route-import
################################################

def route_import_get(ctx, rev, vrf_id):
    return addr_family_v4_ucast_get(ctx, rev, vrf_id)["route-import"]


#######################################################
# BGP/AddressFamily/ipv4-unicast/route-import/from-vrf
######################################################

def route_import_from_vrf_get(ctx, rev, vrf_id):
    return route_import_get(ctx, rev, vrf_id)["from-vrf"]


##################################################################
# BGP/AddressFamily/ipv4-unicast/route-import/from-vrf/list
##################################################################

def route_import_from_vrf_list_get(ctx, rev, vrf_id):
    return route_import_from_vrf_get(ctx, rev, vrf_id)["list"]


################################################
# BGP/AddressFamily/ipv4-unicast/aggregate-routes
################################################

def aggregate_routes_get(ctx, rev, vrf_id):
    return addr_family_v4_ucast_get(ctx, rev, vrf_id)["aggregate-route"]


################################################################
# BGP/AddressFamily/ipv4-unicast/aggregate-routes/{aggregate-route-id}
################################################################

def aggregate_route_get(ctx, rev, vrf_id, aggregate_route_id):
    return aggregate_routes_get(ctx, rev, vrf_id)[aggregate_route_id]


################################################
# BGP/AddressFamily/ipv4-unicast/network
################################################

def static_networks_get(ctx, rev, vrf_id):
    return addr_family_v4_ucast_get(ctx, rev, vrf_id)["network"]


##################################################################
# BGP/AddressFamily/ipv4-unicast/network/{static-network-id}
####################################################################

def static_network_get(ctx, rev, vrf_id, static_network_id):
    return static_networks_get(ctx, rev, vrf_id)[static_network_id]


################################################
# BGP/AddressFamily/ipv4-unicast/multipaths
################################################

def multipaths_get(ctx, rev, vrf_id):
    return addr_family_v4_ucast_get(ctx, rev, vrf_id)["multipaths"]


################################################
# BGP/AddressFamily/ipv4-unicast/admin-distance
################################################

def admin_distance_get(ctx, rev, vrf_id):
    return addr_family_v4_ucast_get(ctx, rev, vrf_id)["admin-distance"]


######################################################
# BGP/AddressFamily/ipv4-unicast/route-export/to-evpn
#####################################################

def route_export_to_evpn_get(ctx, rev, vrf_id):
    return route_export_get(ctx, rev, vrf_id)["to-evpn"]


###############################
# BGP/AddressFamily/ipv4-unicast/redistribute/static
###############################

def static_get(ctx, rev, vrf_id):
    return route_redistribution_get(ctx, rev, vrf_id)["static"]


###############################
# BGP/AddressFamily/ipv4-unicast/redistribute/connected
###############################

def connected_get(ctx, rev, vrf_id):
    return route_redistribution_get(ctx, rev, vrf_id)["connected"]


###############################
# BGP/AddressFamily/ipv4-unicast/redistribute/kernel
###############################

def kernel_get(ctx, rev, vrf_id):
    return route_redistribution_get(ctx, rev, vrf_id)["kernel"]


###############################
# BGP/AddressFamily/ipv4-unicast/redistribute/ospf
###############################

def v4_ospf_get(ctx, rev, vrf_id):
    return route_redistribution_get(ctx, rev, vrf_id)["ospf"]


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib
###############################

def v4_loc_rib_get(ctx, rev, vrf_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.frr_v1.getVrfBgpAfIpv4UnicastRibTrans(vrf_id)


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route
###############################

def v4_loc_rib_routes_get(ctx, rev, vrf_id):
    v4_route = {}

    v4_rib = v4_loc_rib_get(ctx, rev, vrf_id)
    if v4_rib:
        v4_route = v4_rib['route']

    return v4_route


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route/{route-id}
###############################

def v4_loc_rib_route_get(ctx, rev, vrf_id, route_id):
    route_pfx = {}

    v4_routes = v4_loc_rib_routes_get(ctx, rev, vrf_id)
    if v4_routes:
        try:
            route_pfx = v4_routes[route_id]
        except KeyError:
            raise NotFound

    return route_pfx


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route/{route-id}/path
###############################

def v4_route_paths_get(ctx, rev, vrf_id, route_id):
    v4_paths = {}

    v4_route = v4_loc_rib_route_get(ctx, rev, vrf_id, route_id)
    if v4_route:
        v4_paths = v4_route['path']

    return v4_paths


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route/{route-id}/path/{path-id}
###############################

def v4_route_path_get(ctx, rev, vrf_id, route_id, path_id):
    v4_path = {}

    v4_paths = v4_route_paths_get(ctx, rev, vrf_id, route_id)
    if v4_paths:
        try:
            v4_path = v4_paths[path_id]
        except KeyError:
            raise NotFound

    return v4_path


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route/{route-id}/path/{path-id}
# /nexthop
###############################

def v4_route_path_nexthops_get(ctx, rev, vrf_id, route_id, path_id):
    v4_nxthps = {}

    v4_path = v4_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v4_path:
        v4_nxthps = v4_path['nexthop']

    return v4_nxthps


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route/{route-id}/path/{path-id}
# /nexthop/{nexthop-id}
###############################

def v4_route_path_nexthop_get(ctx, rev, vrf_id, route_id, path_id, nexthop_id):
    v4_nxthp = {}

    v4_nxthps = v4_route_path_nexthops_get(ctx, rev, vrf_id, route_id, path_id)
    if v4_nxthps:
        try:
            v4_nxthp = v4_nxthps[nexthop_id]
        except KeyError:
            raise NotFound

    return v4_nxthp


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route/{route-id}/path/{path-id}
# /peer
###############################

def v4_route_path_peer_get(ctx, rev, vrf_id, route_id, path_id):
    v4_peer = {}

    v4_path = v4_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v4_path:
        v4_peer = v4_path['peer']

    return v4_peer


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route/{route-id}/path/{path-id}
# /flags
###############################

def v4_route_path_flags_get(ctx, rev, vrf_id, route_id, path_id):
    v4_flags = {}

    v4_path = v4_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v4_path:
        v4_flags = v4_path['flags']

    return v4_flags


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route/{route-id}/path/{path-id}
# /bestpath
###############################

def v4_route_path_bestpath_get(ctx, rev, vrf_id, route_id, path_id):
    v4_bestpath = {}

    v4_path = v4_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v4_path:
        try:
            v4_bestpath = v4_path['bestpath']
        except KeyError:
            raise NotFound

    return v4_bestpath


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route/{route-id}/path/{path-id}
# /aspath
###############################

def v4_route_path_aspaths_get(ctx, rev, vrf_id, route_id, path_id):
    v4_aspaths = {}

    v4_path = v4_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v4_path:
        v4_aspaths = v4_path['aspath']

    return v4_aspaths


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route/{route-id}/path/{path-id}
# /community
###############################

def v4_route_path_comms_get(ctx, rev, vrf_id, route_id, path_id):
    v4_comms = {}

    v4_path = v4_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v4_path:
        try:
            v4_comms = v4_path['community']
        except KeyError:
            raise NotFound

    return v4_comms


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route/{route-id}/path/{path-id}
# /large-community
###############################

def v4_route_path_large_comms_get(ctx, rev, vrf_id, route_id, path_id):
    v4_large_comms = {}

    v4_path = v4_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v4_path:
        try:
            v4_large_comms = v4_path['large-community']
        except KeyError:
            raise NotFound

    return v4_large_comms


###############################
# BGP/AddressFamily/ipv4-unicast/loc-rib/route/{route-id}/path/{path-id}
# /ext-community
###############################

def v4_route_path_ext_comms_get(ctx, rev, vrf_id, route_id, path_id):
    v4_ext_comms = {}

    v4_path = v4_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v4_path:
        try:
            v4_ext_comms = v4_path['ext-community']
        except KeyError:
            raise NotFound

    return v4_ext_comms


###############################
# BGP/AddressFamily/ipv6-unicast
###############################

def addr_family_v6_ucast_get(ctx, rev, vrf_id):
    return bgp_addr_family_get(ctx, rev, vrf_id)["ipv6-unicast"]


######################################################
# BGP/AddressFamily/ipv6-unicast/aggregate-routes
######################################################

def v6_aggregate_routes_get(ctx, rev, vrf_id):
    return addr_family_v6_ucast_get(ctx, rev, vrf_id)["aggregate-route"]


################################################################
# BGP/AddressFamily/ipv6-unicast/aggregate-routes/{aggregate-route-id}
################################################################

def v6_aggregate_route_get(ctx, rev, vrf_id, aggregate_route_id):
    return v6_aggregate_routes_get(ctx, rev, vrf_id)[aggregate_route_id]


#####################################################################
# BGP/AddressFamily/ipv6-unicast/network
####################################################################

def v6_static_networks_get(ctx, rev, vrf_id):
    return addr_family_v6_ucast_get(ctx, rev, vrf_id)["network"]


################################################################
# BGP/AddressFamily/ipv6-unicast/network/{static-network_id}
################################################################

def v6_static_network_get(ctx, rev, vrf_id, static_network_id):
    return v6_static_networks_get(ctx, rev, vrf_id)[static_network_id]


######################################################
# BGP/AddressFamily/ipv6-unicast/route-import
######################################################

def v6_route_import_get(ctx, rev, vrf_id):
    return addr_family_v6_ucast_get(ctx, rev, vrf_id)["route-import"]


#######################################################
# BGP/AddressFamily/ipv6-unicast/route-import/from-vrf
######################################################

def v6_route_import_from_vrf_get(ctx, rev, vrf_id):
    return v6_route_import_get(ctx, rev, vrf_id)["from-vrf"]


###################################################################
# BGP/AddressFamily/ipv6-unicast/route-import/from-vrf/list
##################################################################

def v6_route_import_from_vrf_list_get(ctx, rev, vrf_id):
    return v6_route_import_from_vrf_get(ctx, rev, vrf_id)["list"]


######################################################
# BGP/AddressFamily/ipv6-unicast/multipaths
######################################################

def v6_multipaths_get(ctx, rev, vrf_id):
    return addr_family_v6_ucast_get(ctx, rev, vrf_id)["multipaths"]


######################################################
# BGP/AddressFamily/ipv6-unicast/admin-distance
######################################################

def v6_admin_distance_get(ctx, rev, vrf_id):
    return addr_family_v6_ucast_get(ctx, rev, vrf_id)["admin-distance"]


######################################################
# BGP/AddressFamily/ipv6-unicast/route-export
######################################################

def v6_route_export_get(ctx, rev, vrf_id):
    return addr_family_v6_ucast_get(ctx, rev, vrf_id)["route-export"]


######################################################
# BGP/AddressFamily/ipv6-unicast/route-export/to-evpn
######################################################

def v6_route_export_to_evpn_get(ctx, rev, vrf_id):
    return v6_route_export_get(ctx, rev, vrf_id)["to-evpn"]


######################################################
# BGP/AddressFamily/ipv6-unicast/redistribute
######################################################

def v6_route_redistribution_get(ctx, rev, vrf_id):
    return addr_family_v6_ucast_get(ctx, rev, vrf_id)["redistribute"]


###############################
# BGP/AddressFamily/ipv6-unicast/redistribute/static
###############################

def v6_static_get(ctx, rev, vrf_id):
    return v6_route_redistribution_get(ctx, rev, vrf_id)["static"]


###############################
# BGP/AddressFamily/ipv6-unicast/redistribute/connected
###############################

def v6_connected_get(ctx, rev, vrf_id):
    return v6_route_redistribution_get(ctx, rev, vrf_id)["connected"]


###############################
# BGP/AddressFamily/ipv6-unicast/redistribute/kernel
###############################

def v6_kernel_get(ctx, rev, vrf_id):
    return v6_route_redistribution_get(ctx, rev, vrf_id)["kernel"]


###############################
# BGP/AddressFamily/ipv6-unicast/redistribute/ospf
###############################

def v6_ospf_get(ctx, rev, vrf_id):
    return v6_route_redistribution_get(ctx, rev, vrf_id)["ospf6"]


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib
###############################

def v6_loc_rib_get(ctx, rev, vrf_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return ctx.frr_v1.getVrfBgpAfIpv6UnicastRibTrans(vrf_id)


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route
###############################

def v6_loc_rib_routes_get(ctx, rev, vrf_id):
    v6_route = {}

    v6_rib = v6_loc_rib_get(ctx, rev, vrf_id)
    if v6_rib:
        v6_route = v6_rib['route']

    return v6_route


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route/{route-id}
###############################

def v6_loc_rib_route_get(ctx, rev, vrf_id, route_id):
    route_pfx = {}

    v6_routes = v6_loc_rib_routes_get(ctx, rev, vrf_id)
    if v6_routes:
        try:
            route_pfx = v6_routes[route_id]
        except KeyError:
            raise NotFound

    return route_pfx


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route/{route-id}/path
###############################

def v6_route_paths_get(ctx, rev, vrf_id, route_id):
    v6_paths = {}

    v6_route = v6_loc_rib_route_get(ctx, rev, vrf_id, route_id)
    if v6_route:
        v6_paths = v6_route['path']

    return v6_paths


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route/{route-id}/path/{path-id}
###############################

def v6_route_path_get(ctx, rev, vrf_id, route_id, path_id):
    v6_path = {}

    v6_paths = v6_route_paths_get(ctx, rev, vrf_id, route_id)
    if v6_paths:
        try:
            v6_path = v6_paths[path_id]
        except KeyError:
            raise NotFound

    return v6_path


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route/{route-id}/path/{path-id}
# /nexthop
###############################

def v6_route_path_nexthops_get(ctx, rev, vrf_id, route_id, path_id):
    v6_nxthps = {}

    v6_path = v6_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v6_path:
        v6_nxthps = v6_path['nexthop']

    return v6_nxthps


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route/{route-id}/path/{path-id}
# /nexthop/{nexthop-id}
###############################

def v6_route_path_nexthop_get(ctx, rev, vrf_id, route_id, path_id, nexthop_id):
    v6_nxthp = {}

    v6_nxthps = v6_route_path_nexthops_get(ctx, rev, vrf_id, route_id, path_id)
    if v6_nxthps:
        try:
            v6_nxthp = v6_nxthps[nexthop_id]
        except KeyError:
            raise NotFound

    return v6_nxthp


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route/{route-id}/path/{path-id}
# /peer
###############################

def v6_route_path_peer_get(ctx, rev, vrf_id, route_id, path_id):
    v6_peer = {}

    v6_path = v6_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v6_path:
        v6_peer = v6_path['peer']

    return v6_peer


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route/{route-id}/path/{path-id}
# /flags
###############################

def v6_route_path_flags_get(ctx, rev, vrf_id, route_id, path_id):
    v6_flags = {}

    v6_path = v6_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v6_path:
        v6_flags = v6_path['flags']

    return v6_flags


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route/{route-id}/path/{path-id}
# /bestpath
###############################

def v6_route_path_bestpath_get(ctx, rev, vrf_id, route_id, path_id):
    v6_bestpath = {}

    v6_path = v6_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v6_path:
        try:
            v6_bestpath = v6_path['bestpath']
        except KeyError:
            raise NotFound

    return v6_bestpath


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route/{route-id}/path/{path-id}
# /aspath
###############################

def v6_route_path_aspaths_get(ctx, rev, vrf_id, route_id, path_id):
    v6_aspaths = {}

    v6_path = v6_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v6_path:
        v6_aspaths = v6_path['aspath']

    return v6_aspaths


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route/{route-id}/path/{path-id}
# /community
###############################

def v6_route_path_comms_get(ctx, rev, vrf_id, route_id, path_id):
    v6_comms = {}

    v6_path = v6_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v6_path:
        try:
            v6_comms = v6_path['community']
        except KeyError:
            raise NotFound

    return v6_comms


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route/{route-id}/path/{path-id}
# /large-community
###############################

def v6_route_path_large_comms_get(ctx, rev, vrf_id, route_id, path_id):
    v6_large_comms = {}

    v6_path = v6_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v6_path:
        try:
            v6_large_comms = v6_path['large-community']
        except KeyError:
            raise NotFound

    return v6_large_comms


###############################
# BGP/AddressFamily/ipv6-unicast/loc-rib/route/{route-id}/path/{path-id}
# /ext-community
###############################

def v6_route_path_ext_comms_get(ctx, rev, vrf_id, route_id, path_id):
    v6_ext_comms = {}

    v6_path = v6_route_path_get(ctx, rev, vrf_id, route_id, path_id)
    if v6_path:
        try:
            v6_ext_comms = v6_path['ext-community']
        except KeyError:
            raise NotFound

    return v6_ext_comms


###############################
# BGP/AddressFamily/l2vpn-evpn
###############################

def l2vpn_evpn_get(ctx, rev, vrf_id):
    return bgp_addr_family_get(ctx, rev, vrf_id)["l2vpn-evpn"]


###############################
# BGP/path-selection
###############################

def bgp_path_selection_get(ctx, rev, vrf_id):
    return bgp_get(ctx, rev, vrf_id)["path-selection"]


###############################
# BGP/path-selection/aspath
###############################

def bgp_path_selection_aspath_get(ctx, rev, vrf_id):
    return bgp_path_selection_get(ctx, rev, vrf_id)["aspath"]


###############################
# BGP/path-selection/med
###############################

def bgp_path_selection_med_get(ctx, rev, vrf_id):
    return bgp_path_selection_get(ctx, rev, vrf_id)["med"]


###############################
# BGP/path-selection/multipath
###############################

def bgp_path_selection_multipath_get(ctx, rev, vrf_id):
    return bgp_path_selection_get(ctx, rev, vrf_id)["multipath"]


###############################
# BGP/route-reflection
###############################

def bgp_route_reflection_get(ctx, rev, vrf_id):
    return bgp_get(ctx, rev, vrf_id)["route-reflection"]


###############################
# BGP/peer-groups
###############################

def bgp_peer_groups_get(ctx, rev, vrf_id):
    return bgp_get(ctx, rev, vrf_id)["peer-group"]


###############################
# BGP/peer-group
###############################

def bgp_peer_group_get(ctx, rev, vrf_id, peer_group_id):
    peer_groups = bgp_peer_groups_get(ctx, rev, vrf_id)
    try:
        return peer_groups[peer_group_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


##############################################
# BGP/peer-group/{peer-group-id}/local-as
##############################################

def bgp_local_as_get(ctx, rev, vrf_id, peer_group_id):
    return bgp_peer_group_get(ctx, rev, vrf_id, peer_group_id)["local-as"]


##############################################
# BGP/peer-group/{peer-group-id}/capabilities
##############################################

def bgp_capabilities_get(ctx, rev, vrf_id, peer_group_id):
    return bgp_peer_group_get(ctx, rev, vrf_id, peer_group_id)["capabilities"]


##############################################
# BGP/peer-group/{peer-group-id}/ttl-security
##############################################

def bgp_ttl_security_get(ctx, rev, vrf_id, peer_group_id):
    return bgp_peer_group_get(ctx, rev, vrf_id, peer_group_id)["ttl-security"]


#####################################
# BGP/peer-group/{peer-group-id}/bfd
####################################

def bgp_peer_group_bfd_get(ctx, rev, vrf_id, peer_group_id):
    return bgp_peer_group_get(ctx, rev, vrf_id, peer_group_id)["bfd"]


###############################
# BGP/peer-group/timers
###############################

def bgp_peer_group_timers_get(ctx, rev, vrf_id, peer_group_id):
    return bgp_peer_group_get(ctx, rev, vrf_id, peer_group_id)["timers"]


###############################
# BGP/peer-group/address-family
###############################

def bgp_peer_group_addr_family_get(ctx, rev, vrf_id, peer_group_id):
    return bgp_peer_group_get(ctx, rev, vrf_id,
                              peer_group_id)["address-family"]


##############################################
# BGP/peer-group/address-family/ipv4-unicast
##############################################

def peer_group_af_v4_ucast_get(ctx, rev, vrf_id, peer_group_id):
    return bgp_peer_group_addr_family_get(ctx, rev, vrf_id,
                                          peer_group_id)["ipv4-unicast"]


############################################################
# BGP/peer-group/address-family/ipv4-unicast/attribute-mod
###########################################################

def peer_group_v4_attribute_mod_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v4_ucast_get(ctx, rev, vrf_id,
                                      peer_group_id)["attribute-mod"]


#####################################################################
# BGP/peer-group/address-family/ipv4-unicast/community-advertise
#####################################################################

def peer_group_af_v4_comm_advts_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v4_ucast_get(ctx, rev, vrf_id,
                                      peer_group_id)["community-advertise"]


############################################################
# BGP/peer-group/address-family/ipv4-unicast/aspath
###########################################################

def peer_group_v4_aspath_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v4_ucast_get(ctx, rev, vrf_id,
                                      peer_group_id)["aspath"]


#################################################################
# BGP/peer-group/address-family/ipv4-unicast/aspath/allow-my-asn
#################################################################

def peer_group_v4_allow_my_as_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_v4_aspath_get(ctx, rev, vrf_id,
                                    peer_group_id)["allow-my-asn"]


############################################################
# BGP/peer-group/address-family/ipv4-unicast/prefix-limits
###########################################################

def peer_group_v4_prefix_limits_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v4_ucast_get(ctx, rev, vrf_id,
                                      peer_group_id)["prefix-limits"]


####################################################################
# BGP/peer-group/address-family/ipv4-unicast/prefix-limits/inbound
###################################################################

def peer_group_v4_prefix_limits_inbound_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_v4_prefix_limits_get(ctx, rev, vrf_id,
                                           peer_group_id)["inbound"]


############################################################
# BGP/peer-group/address-family/ipv4-unicast/policy
###########################################################

def peer_group_v4_policy_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v4_ucast_get(ctx, rev, vrf_id,
                                      peer_group_id)["policy"]


############################################################
# BGP/peer-group/address-family/ipv4-unicast/policy/inbound
###########################################################

def peer_group_v4_policy_inbound_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_v4_policy_get(ctx, rev, vrf_id,
                                    peer_group_id)["inbound"]


###############################################################
# BGP/peer-group/address-family/ipv4-unicast/policy/outbound
##############################################################

def peer_group_v4_policy_outbound_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_v4_policy_get(ctx, rev, vrf_id,
                                    peer_group_id)["outbound"]


#######################################################################
# BGP/peer-group/address-family/ipv4-unicast/default-route-origination
#######################################################################

def peer_group_v4_def_rt_origin_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v4_ucast_get(
        ctx, rev, vrf_id, peer_group_id)["default-route-origination"]


#######################################################################
# BGP/peer-group/address-family/ipv4-unicast/conditional-advertise
#######################################################################

def peer_group_af_v4_cond_advts_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v4_ucast_get(
        ctx, rev, vrf_id, peer_group_id)["conditional-advertise"]


##############################################
# BGP/peer-group/address-family/ipv6-unicast
##############################################

def peer_group_af_v6_ucast_get(ctx, rev, vrf_id, peer_group_id):
    return bgp_peer_group_addr_family_get(ctx, rev, vrf_id,
                                          peer_group_id)["ipv6-unicast"]


#########################################################
# BGP/peer-group/address-family/ipv6-unicast/aspath
##########################################################

def peer_group_af_v6_aspath_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v6_ucast_get(ctx, rev, vrf_id,
                                      peer_group_id)["aspath"]


#################################################################
# BGP/peer-group/address-family/ipv6-unicast/aspath/allow-my-asn
################################################################

def peer_group_af_v6_allow_my_as_get(
        ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v6_aspath_get(ctx, rev, vrf_id,
                                       peer_group_id)["allow-my-asn"]


#########################################################
# BGP/peer-group/address-family/ipv6-unicast/policy
##########################################################

def peer_group_af_v6_policy_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v6_ucast_get(ctx, rev, vrf_id,
                                      peer_group_id)["policy"]


##############################################################
# BGP/peer-group/address-family/ipv6-unicast/policy/inbound
##############################################################

def peer_group_af_v6_policy_inbound_get(
        ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v6_policy_get(ctx, rev, vrf_id,
                                       peer_group_id)["inbound"]


##############################################################
# BGP/peer-group/address-family/ipv6-unicast/policy/outbound
##############################################################

def peer_group_af_v6_policy_outbound_get(
        ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v6_policy_get(ctx, rev, vrf_id,
                                       peer_group_id)["outbound"]


###############################################################
# BGP/peer-group/address-family/ipv6-unicast/prefix-limits
###############################################################

def peer_group_af_v6_prefix_limits_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v6_ucast_get(ctx, rev, vrf_id,
                                      peer_group_id)["prefix-limits"]


#####################################################################
# BGP/peer-group/address-family/ipv6-unicast/prefix-limits/inbound
####################################################################

def peer_group_af_v6_prefix_limits_inbound_get(
        ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v6_prefix_limits_get(
        ctx, rev, vrf_id, peer_group_id)["inbound"]


###################################################################
# BGP/peer-group/address-family/ipv6-unicast/attribute-mod
##################################################################

def peer_group_af_v6_attribute_mod_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v6_ucast_get(ctx, rev, vrf_id,
                                      peer_group_id)["attribute-mod"]


########################################################################
# BGP/peer-group/address-family/ipv6-unicast/default-route-origination
#######################################################################

def peer_group_af_v6_def_rt_origin_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v6_ucast_get(
        ctx, rev, vrf_id, peer_group_id)["default-route-origination"]


#####################################################################
# BGP/peer-group/address-family/ipv6-unicast/community-advertise
#####################################################################

def peer_group_af_v6_comm_advts_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v6_ucast_get(ctx, rev, vrf_id,
                                      peer_group_id)["community-advertise"]


#######################################################################
# BGP/peer-group/address-family/ipv6-unicast/conditional-advertise
#######################################################################

def peer_group_af_v6_cond_advts_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_v6_ucast_get(
        ctx, rev, vrf_id, peer_group_id)["conditional-advertise"]


##############################################
# BGP/peer-group/address-family/l2vpn-evpn
##############################################

def peer_group_af_l2vpn_evpn_get(ctx, rev, vrf_id, peer_group_id):
    return bgp_peer_group_addr_family_get(ctx, rev, vrf_id,
                                          peer_group_id)["l2vpn-evpn"]


################################################################
# BGP/peer-group/address-family/l2vpn-evpn/attribute-mod
################################################################

def l2vpn_attribute_mod_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_l2vpn_evpn_get(
        ctx, rev, vrf_id, peer_group_id)["attribute-mod"]


################################################################
# BGP/peer-group/address-family/l2vpn-evpn/aspath
################################################################

def l2vpn_aspath_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_l2vpn_evpn_get(
        ctx, rev, vrf_id, peer_group_id)["aspath"]


################################################################
# BGP/peer-group/address-family/l2vpn-evpn/aspath/allow-my-asn
################################################################

def l2vpn_allow_my_as_get(ctx, rev, vrf_id, peer_group_id):
    return l2vpn_aspath_get(
        ctx, rev, vrf_id, peer_group_id)["allow-my-asn"]


################################################################
# BGP/peer-group/address-family/l2vpn-evpn/policy
################################################################

def l2vpn_policy_get(ctx, rev, vrf_id, peer_group_id):
    return peer_group_af_l2vpn_evpn_get(
        ctx, rev, vrf_id, peer_group_id)["policy"]


################################################################
# BGP/peer-group/address-family/l2vpn-evpn/policy/inbound
################################################################

def l2vpn_policy_inbound_get(ctx, rev, vrf_id, peer_group_id):
    return l2vpn_policy_get(
        ctx, rev, vrf_id, peer_group_id)["inbound"]


################################################################
# BGP/peer-group/address-family/l2vpn-evpn/policy/outbound
################################################################

def l2vpn_policy_outbound_get(ctx, rev, vrf_id, peer_group_id):
    return l2vpn_policy_get(
        ctx, rev, vrf_id, peer_group_id)["outbound"]


###############################
# BGP/route-export
###############################

def bgp_route_export_get(ctx, rev, vrf_id):
    return bgp_get(ctx, rev, vrf_id)["route-export"]


###############################
# BGP/route-export/to-evpn
###############################

def bgp_rt_export_to_evpn_get(ctx, rev, vrf_id):
    return bgp_route_export_get(ctx, rev, vrf_id)["to-evpn"]


###########################################
# BGP/route-export/to-evpn/route-target
##########################################

def bgp_rt_export_to_evpn_rt_targets_get(ctx, rev, vrf_id):
    return bgp_rt_export_to_evpn_get(ctx, rev, vrf_id)["route-target"]


###########################################
# BGP/route-export/to-evpn/route-target/{rt-id}
##########################################

def bgp_rt_export_to_evpn_rt_target_get(ctx, rev, vrf_id, rt_id):
    rts = bgp_rt_export_to_evpn_rt_targets_get(ctx, rev, vrf_id)
    try:
        return rts[rt_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


############################################
# vrf/{vrf-id}/router/bgp/route-import
############################################

def bgp_route_import_get(ctx, rev, vrf_id):
    return bgp_get(ctx, rev, vrf_id)["route-import"]


#######################################################
# vrf/{vrf-id}/router/bgp/route-import/from-evpn
#######################################################

def bgp_from_evpn_get(ctx, rev, vrf_id):
    return bgp_route_import_get(ctx, rev, vrf_id)["from-evpn"]


################################################################
# vrf/{vrf-id}/router/bgp/route-import/from-evpn/route-target
#################################################################

def bgp_from_evpn_rt_targets_get(ctx, rev, vrf_id):
    return bgp_from_evpn_get(ctx, rev, vrf_id)["route-target"]


################################################################
# vrf/{vrf-id}/router/bgp/route-import/from-evpn/route-target/{rt-id}
#################################################################

def bgp_from_evpn_rt_target_get(ctx, rev, vrf_id, rt_id):
    rts = bgp_from_evpn_rt_targets_get(ctx, rev, vrf_id)
    try:
        return rts[rt_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


############################################
# vrf/{vrf-id}/router/bgp/timers
############################################

def bgp_timers_get(ctx, rev, vrf_id):
    return bgp_get(ctx, rev, vrf_id)["timers"]


############################################
# vrf/{vrf-id}/router/bgp/confederation
############################################

def bgp_confederation_get(ctx, rev, vrf_id):
    return bgp_get(ctx, rev, vrf_id)["confederation"]


####################################################
# vrf/{vrf-id}/router/bgp/confederation/member-as
####################################################

def bgp_confederation_member_as_get(ctx, rev, vrf_id):
    return bgp_confederation_get(ctx, rev, vrf_id)["member-as"]


###############################
# BGP/peers
###############################

def bgp_neighbors_get(ctx, rev, vrf_id):
    if rev == "operational":
        return ctx.frr_v1.getVrfBgpNbrsTrans(vrf_id)
    try:
        return bgp_get(ctx, rev, vrf_id)["neighbor"]
    except KeyError:
        raise NotFound


###############################
# BGP/neighbor
###############################

def bgp_neighbor_get(ctx, rev, vrf_id, neighbor_id):

    if rev == "operational":
        return ctx.frr_v1.getVrfBgpNbrTrans(vrf_id, neighbor_id)

    neighbors = bgp_neighbors_get(ctx, rev, vrf_id)
    try:
        return neighbors[neighbor_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# BGP/neighbor/address-familty
###############################

def bgp_neighbor_addr_family_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_get(ctx, rev, vrf_id, neighbor_id)["address-family"]


###############################
# BGP/neighbor/{neighbor-id}/bfd
###############################

def bgp_neighbor_bfd_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_get(ctx, rev, vrf_id, neighbor_id)["bfd"]


###############################
# BGP/neighbor/{neighbor-id}/capabilities
###############################

def bgp_neighbor_capabilities_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_get(ctx, rev, vrf_id, neighbor_id)["capabilities"]


###############################
# BGP/neighbor/{neighbor-id}/local-as
###############################

def bgp_neighbor_local_as_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_get(ctx, rev, vrf_id, neighbor_id)["local-as"]


##################################
# BGP/neighbor/{neighbor-id}/ttl-security
##################################

def bgp_neighbor_ttl_security_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_get(ctx, rev, vrf_id, neighbor_id)["ttl-security"]


##################################
# BGP/neighbor/{neighbor-id}/nexthop
##################################

def bgp_neighbor_nexthop_get(ctx, rev, vrf_id, neighbor_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return bgp_neighbor_get(ctx, rev, vrf_id,
                            neighbor_id)["nexthop"]


##################################
# BGP/neighbor/{neighbor-id}/message-stats
##################################

def bgp_neighbor_message_stats_get(ctx, rev, vrf_id, neighbor_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return bgp_neighbor_get(ctx, rev, vrf_id,
                            neighbor_id)["message-stats"]


##################################
# BGP/neighbor/{neighbor-id}/ebgp-policy
##################################

def bgp_neighbor_ebgp_policy_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_get(ctx, rev, vrf_id,
                            neighbor_id)["ebgp-policy"]


#########################################
# BGP/neighbor/address-familty/ipv4-unicast
#########################################

def bgp_neighbor_af_v4ucast_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_addr_family_get(ctx, rev, vrf_id,
                                        neighbor_id)["ipv4-unicast"]


#######################################################
# BGP/neighbor/address-familty/ipv4-unicast/attribute-mod
######################################################

def bgp_neighbor_v4_attribute_mod_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v4ucast_get(ctx, rev, vrf_id,
                                       neighbor_id)["attribute-mod"]


#######################################################
# BGP/neighbor/address-familty/ipv4-unicast/aspath
######################################################

def bgp_neighbor_v4_aspath_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v4ucast_get(ctx, rev, vrf_id, neighbor_id)["aspath"]


############################################################
# BGP/neighbor/address-familty/ipv4-unicast/aspath/allow-my-asn
###########################################################

def bgp_neighbor_v4_allow_my_as_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_v4_aspath_get(ctx, rev, vrf_id,
                                      neighbor_id)["allow-my-asn"]


###################################################################
# BGP/neighbor/address-familty/ipv4-unicast/default-route-origination
####################################################################

def bgp_neighbor_v4_def_rt_origin_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v4ucast_get(
        ctx, rev, vrf_id, neighbor_id)["default-route-origination"]


#######################################################
# BGP/neighbor/address-familty/ipv4-unicast/policy
######################################################

def bgp_neighbor_v4_policy_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v4ucast_get(ctx, rev, vrf_id, neighbor_id)["policy"]


#########################################################
# BGP/neighbor/address-familty/ipv4-unicast/policy/inbound
#########################################################

def bgp_neighbor_v4_policy_inbound_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_v4_policy_get(ctx, rev, vrf_id, neighbor_id)["inbound"]


#########################################################
# BGP/neighbor/address-familty/ipv4-unicast/policy/outbound
#########################################################

def bgp_neighbor_v4_policy_outbound_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_v4_policy_get(ctx, rev, vrf_id,
                                      neighbor_id)["outbound"]


#######################################################
# BGP/neighbor/address-familty/ipv4-unicast/prefix-limits
######################################################

def bgp_neighbor_v4_prefix_limits_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v4ucast_get(ctx, rev, vrf_id,
                                       neighbor_id)["prefix-limits"]


###############################################################
# BGP/neighbor/address-familty/ipv4-unicast/prefix-limits/inbound
##############################################################

def bgp_neighbor_v4_prefix_limits_inbound_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_v4_prefix_limits_get(ctx, rev, vrf_id,
                                             neighbor_id)["inbound"]


#############################################################
# BGP/neighbor/address-familty/ipv4-unicast/community-advertise
############################################################

def bgp_neighbor_v4_comm_advertise_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v4ucast_get(
        ctx, rev, vrf_id, neighbor_id)["community-advertise"]


#############################################################
# BGP/neighbor/address-familty/ipv4-unicast/conditional-advertise
############################################################

def bgp_neighbor_af_v4_cond_advts_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v4ucast_get(
        ctx, rev, vrf_id, neighbor_id)["conditional-advertise"]


#############################################################
# BGP/neighbor/address-familty/ipv4-unicast/capabilities
############################################################

def bgp_neighbor_af_v4_caps_get(ctx, rev, vrf_id, neighbor_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return bgp_neighbor_af_v4ucast_get(
        ctx, rev, vrf_id, neighbor_id)["capabilities"]


#############################################################
# BGP/neighbor/address-familty/ipv4-unicast/graceful-restart
############################################################

def bgp_neighbor_af_v4_gr_get(ctx, rev, vrf_id, neighbor_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return bgp_neighbor_af_v4ucast_get(
        ctx, rev, vrf_id, neighbor_id)["graceful-restart"]


#########################################
# BGP/neighbor/address-familty/ipv6-unicast
#########################################

def bgp_neighbor_af_v6ucast_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_addr_family_get(ctx, rev, vrf_id,
                                        neighbor_id)["ipv6-unicast"]


#######################################################
# BGP/neighbor/address-familty/ipv6-unicast/attribute-mod
#######################################################

def bgp_neighbor_v6_attribute_mod_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v6ucast_get(
        ctx, rev, vrf_id, neighbor_id)["attribute-mod"]


#######################################################
# BGP/neighbor/address-familty/ipv6-unicast/aspath
#######################################################

def bgp_neighbor_v6_aspath_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v6ucast_get(
        ctx, rev, vrf_id, neighbor_id)["aspath"]


#############################################################
# BGP/neighbor/address-familty/ipv6-unicast/aspath/allow-my-asn
#############################################################

def bgp_neighbor_v6_allow_my_as_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_v6_aspath_get(
        ctx, rev, vrf_id, neighbor_id)["allow-my-asn"]


#######################################################
# BGP/neighbor/address-familty/ipv6-unicast/policy
#######################################################

def bgp_neighbor_v6_policy_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v6ucast_get(
        ctx, rev, vrf_id, neighbor_id)["policy"]


#######################################################
# BGP/neighbor/address-familty/ipv6-unicast/policy/inbound
#######################################################

def bgp_neighbor_v6_policy_inbound_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_v6_policy_get(
        ctx, rev, vrf_id, neighbor_id)["inbound"]


#######################################################
# BGP/neighbor/address-familty/ipv6-unicast/policy/outbound
#######################################################

def bgp_neighbor_v6_policy_outbound_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_v6_policy_get(
        ctx, rev, vrf_id, neighbor_id)["outbound"]


#######################################################
# BGP/neighbor/address-familty/ipv6-unicast/prefix-limits
#######################################################

def bgp_neighbor_v6_prefix_limits_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v6ucast_get(
        ctx, rev, vrf_id, neighbor_id)["prefix-limits"]


##############################################################
# BGP/neighbor/address-familty/ipv6-unicast/prefix-limits/inbound
##############################################################

def bgp_neighbor_v6_prefix_limits_inbound_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_v6_prefix_limits_get(
        ctx, rev, vrf_id, neighbor_id)["inbound"]


####################################################################
# BGP/neighbor/address-familty/ipv6-unicast/default-route-origination
###################################################################

def bgp_neighbor_v6_def_rt_origin_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v6ucast_get(
        ctx, rev, vrf_id, neighbor_id)["default-route-origination"]


###############################################################
# BGP/neighbor/address-familty/ipv6-unicast/community-advertise
###############################################################

def bgp_neighbor_v6_comm_advts_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v6ucast_get(
        ctx, rev, vrf_id, neighbor_id)["community-advertise"]


###############################################################
# BGP/neighbor/address-familty/ipv6-unicast/conditional-advertise
###############################################################

def bgp_neighbor_af_v6_cond_advts_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_v6ucast_get(
        ctx, rev, vrf_id, neighbor_id)["conditional-advertise"]


#############################################################
# BGP/neighbor/address-familty/ipv6-unicast/capabilities
############################################################

def bgp_neighbor_af_v6_caps_get(ctx, rev, vrf_id, neighbor_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return bgp_neighbor_af_v6ucast_get(
        ctx, rev, vrf_id, neighbor_id)["capabilities"]


#############################################################
# BGP/neighbor/address-familty/ipv6-unicast/graceful-restart
############################################################

def bgp_neighbor_af_v6_gr_get(ctx, rev, vrf_id, neighbor_id):
    ops = ctx.cue_v1._ops
    ops.pre_get_oper_only(rev)
    return bgp_neighbor_af_v6ucast_get(
        ctx, rev, vrf_id, neighbor_id)["graceful-restart"]


#########################################
# BGP/neighbor/address-familty/l2vpn-evpn
#########################################

def bgp_neighbor_af_l2vpn_evpn_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_addr_family_get(ctx, rev, vrf_id,
                                        neighbor_id)["l2vpn-evpn"]


######################################################
# BGP/neighbor/address-familty/l2vpn-evpn/attribute-mod
######################################################

def bgp_neighbor_attribute_mod_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_l2vpn_evpn_get(
        ctx, rev, vrf_id, neighbor_id)["attribute-mod"]


######################################################
# BGP/neighbor/address-familty/l2vpn-evpn/aspath
######################################################

def bgp_neighbor_aspath_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_l2vpn_evpn_get(
        ctx, rev, vrf_id, neighbor_id)["aspath"]


######################################################
# BGP/neighbor/address-familty/l2vpn-evpn/aspath/allow-my-asn
######################################################

def bgp_neighbor_allow_my_as_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_aspath_get(
        ctx, rev, vrf_id, neighbor_id)["allow-my-asn"]


######################################################
# BGP/neighbor/address-familty/l2vpn-evpn/policy
######################################################

def bgp_neighbor_policy_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_l2vpn_evpn_get(
        ctx, rev, vrf_id, neighbor_id)["policy"]


######################################################
# BGP/neighbor/address-familty/l2vpn-evpn/policy/inbound
######################################################

def bgp_neighbor_policy_inbound_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_policy_get(
        ctx, rev, vrf_id, neighbor_id)["inbound"]


######################################################
# BGP/neighbor/address-familty/l2vpn-evpn/policy/outbound
######################################################

def bgp_neighbor_policy_outbound_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_policy_get(
        ctx, rev, vrf_id, neighbor_id)["outbound"]


#############################################################
# BGP/neighbor/address-familty/l2vpn-evpn/capabilities
############################################################

def bgp_neighbor_caps_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_l2vpn_evpn_get(
        ctx, rev, vrf_id, neighbor_id)["capabilities"]


#############################################################
# BGP/neighbor/address-familty/l2vpn-evpn/graceful-restart
############################################################

def bgp_neighbor_gr_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_af_l2vpn_evpn_get(
        ctx, rev, vrf_id, neighbor_id)["graceful-restart"]


###############################
# BGP/neighbor/timers
###############################

def bgp_neighbor_timers_get(ctx, rev, vrf_id, neighbor_id):
    return bgp_neighbor_get(ctx, rev, vrf_id, neighbor_id)["timers"]


###############################
# vrf/route-table (rib)
###############################

def vrf_rib_get(ctx, rev, vrf_id, afi):
    if rev == "operational":
        if afi == 'ipv4':
            rib = ctx.frr_v1.getVrfIpv4RouteTable(vrf_id)
        elif afi == 'ipv6':
            rib = ctx.frr_v1.getVrfIpv6RouteTable(vrf_id)
        else:
            rib = {}

        return rib

    return {}


def vrf_ribs_get(ctx, rev, vrf_id):
    if rev == "operational":
        ribs = {}

        # Fetch IPv4 routes
        ipv4_routes = vrf_rib_get(ctx, rev, vrf_id, 'ipv4')
        if ipv4_routes:
            ribs['ipv4'] = ipv4_routes

        # Fetch IPv6 routes
        ipv6_routes = vrf_rib_get(ctx, rev, vrf_id, 'ipv6')
        if ipv6_routes:
            ribs['ipv6'] = ipv6_routes

        return ribs

    return {}


def vrf_rib_route_get(ctx, rev, vrf_id, afi):
    route = {}

    rib = vrf_rib_get(ctx, rev, vrf_id, afi)
    if rib:
        route = rib['route']

    return route


def vrf_rib_route_prefix_get(ctx, rev, vrf_id, afi, route_id):
    route_pfx = {}

    rib_route = vrf_rib_route_get(ctx, rev, vrf_id, afi)
    if rib_route:
        try:
            route_pfx = rib_route[route_id]
        except KeyError:
            raise NotFound

    return route_pfx


def vrf_rib_route_prefix_protocol_get(ctx, rev, vrf_id, afi, route_id):
    route_proto = {}

    rib_prefix = vrf_rib_route_prefix_get(ctx, rev, vrf_id, afi, route_id)
    if rib_prefix:
        route_proto = rib_prefix['protocol']

    return route_proto


def vrf_rib_route_prefix_protocol_id_get(ctx, rev, vrf_id,
                                         afi, route_id, protocol_id):
    rib_protocol = vrf_rib_route_prefix_protocol_get(ctx, rev,
                                                     vrf_id,
                                                     afi, route_id)
    route_proto_id = {}
    if rib_protocol:
        try:
            route_proto_id = rib_protocol[protocol_id]
        except KeyError:
            raise NotFound

    return route_proto_id


def vrf_rib_route_prefix_entries_get(ctx, rev, vrf_id,
                                     afi, route_id, protocol_id):
    route_entry = {}
    rib_protocol = vrf_rib_route_prefix_protocol_id_get(ctx, rev,
                                                        vrf_id, afi,
                                                        route_id, protocol_id)
    if rib_protocol:
        route_entry = rib_protocol['entry-index']

    return route_entry


def vrf_rib_route_prefix_entry_get(ctx, rev, vrf_id, afi,
                                   route_id, protocol_id, entry_index):
    route_entry = {}
    rib_protocol = vrf_rib_route_prefix_entries_get(ctx, rev,
                                                    vrf_id, afi,
                                                    route_id, protocol_id)
    if rib_protocol:
        try:
            route_entry = rib_protocol[entry_index]
        except KeyError:
            raise NotFound

    return route_entry


def vrf_rib_route_prefix_entry_flags_get(ctx, rev, vrf_id, afi,
                                         route_id, protocol_id, entry_index):
    entry_flags = {}
    route_entry = vrf_rib_route_prefix_entry_get(ctx, rev,
                                                 vrf_id, afi,
                                                 route_id, protocol_id,
                                                 entry_index)
    if route_entry:
        entry_flags = route_entry['flags']

    return entry_flags


def vrf_rib_route_prefix_vias_get(ctx, rev, vrf_id, afi,
                                  route_id, protocol_id, entry_index):
    rib_vias = {}
    rib_entry = vrf_rib_route_prefix_entry_get(ctx, rev,
                                               vrf_id, afi, route_id,
                                               protocol_id, entry_index)
    if rib_entry:
        rib_vias = rib_entry['via']

    return rib_vias


def vrf_rib_route_prefix_via_get(ctx, rev, vrf_id, afi,
                                 route_id, protocol_id, entry_index,
                                 via_id):
    rib_via = {}
    rib_vias = vrf_rib_route_prefix_vias_get(ctx, rev,
                                             vrf_id, afi, route_id,
                                             protocol_id, entry_index)
    if rib_vias:
        rib_via = rib_vias[via_id]

    return rib_via


def vrf_rib_route_prefix_via_flags_get(ctx, rev, vrf_id, afi,
                                       route_id, protocol_id,
                                       entry_index, via_id):
    rib_via_flags = {}
    rib_via = vrf_rib_route_prefix_via_get(ctx, rev, vrf_id, afi,
                                           route_id, protocol_id,
                                           entry_index, via_id)
    if rib_via:
        rib_via_flags = rib_via['flags']

    return rib_via_flags


def vrf_rib_route_prefix_resolved_vias_get(ctx, rev, vrf_id, afi,
                                           route_id, protocol_id,
                                           entry_index, via_id):
    rib_resolved_vias = {}
    rib_via = vrf_rib_route_prefix_via_get(ctx, rev, vrf_id, afi,
                                           route_id, protocol_id,
                                           entry_index, via_id)
    if rib_via:
        try:
            rib_resolved_vias = rib_via['resolved-via']
        except KeyError:
            raise NotFound

    return rib_resolved_vias


def vrf_rib_route_prefix_resolved_via_get(ctx, rev, vrf_id, afi,
                                          route_id, protocol_id,
                                          entry_index, via_id,
                                          resolved_via_id):
    rib_resolved_via = {}
    rib_via = vrf_rib_route_prefix_resolved_vias_get(ctx, rev, vrf_id,
                                                     afi, route_id,
                                                     protocol_id,
                                                     entry_index,
                                                     via_id)
    if rib_via:
        try:
            rib_resolved_via = rib_via[resolved_via_id]
        except KeyError:
            raise NotFound

    return rib_resolved_via


###############################
# Static/routes
###############################

def static_routes_get(ctx, rev, vrf_id):
    if rev != "operational":
        return router_get(ctx, rev, vrf_id)["static"]

    # We will not show the operational state as of now.
    raise ctx.errors.OperationalNotSupported


###############################
# Static/route
###############################

def static_route_get(ctx, rev, vrf_id, route_id):
    static_routes = static_routes_get(ctx, rev, vrf_id)
    try:
        return static_routes[route_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# Static/route/vias
###############################

def static_route_vias_get(ctx, rev, vrf_id, route_id):
    static_route = static_route_get(ctx, rev, vrf_id, route_id)
    return static_route.get("via", {})


###############################
# Static/route/via
###############################

def static_route_via_get(ctx, rev, vrf_id, route_id, via_id):
    next_hops = static_route_vias_get(ctx, rev, vrf_id, route_id)
    try:
        return next_hops[via_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# Static/route/via/flag
###############################

def static_route_via_flag_get(ctx, rev, vrf_id, route_id, via_id):
    next_hop = static_route_via_get(
        ctx, rev, vrf_id, route_id, via_id)
    return next_hop.get("flag", {})


##################################
# Static/route/distances
##################################

def static_route_distances_get(ctx, rev, vrf_id, route_id):
    static_route = static_route_get(ctx, rev, vrf_id, route_id)
    return static_route.get("distance", {})


###############################
# Static/route/distance
###############################

def static_route_distance_get(ctx, rev, vrf_id, route_id, distance_id):
    distances = static_route_distances_get(ctx, rev, vrf_id, route_id)
    try:
        return distances[distance_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


#########################################
# Static/route/distance/vias
########################################

def static_route_distance_vias_get(ctx, rev, vrf_id, route_id, distance_id):

    distance = static_route_distance_get(
        ctx, rev, vrf_id, route_id, distance_id)
    return distance.get("via", {})


#########################################
# Static/route/distance/via
########################################

def static_route_distance_via_get(
        ctx, rev, vrf_id, route_id, distance_id, via_id):
    nhs = static_route_distance_vias_get(
        ctx, rev, vrf_id, route_id, distance_id)
    try:
        return nhs[via_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


##############################################
# Static/route/alternatepath/via/flag
##############################################

def static_route_distance_via_flag_get(
        ctx, rev, vrf_id, route_id, distance_id, via_id):
    next_hop = static_route_distance_via_get(
        ctx, rev, vrf_id, route_id, distance_id, via_id)
    return next_hop.get("flag", {})


###############################
# IP
###############################

def ip_get(ctx, rev, vrf_id):
    return loopback_get(ctx, rev, vrf_id)["ip"]


###############################
# IP Addresses
###############################

def ip_addresses_get(ctx, rev, vrf_id):
    return ip_get(ctx, rev, vrf_id)["address"]


###############################
# IP Address
###############################

def ip_address_get(ctx, rev, vrf_id, ip_prefix_id):
    addresses = ip_addresses_get(ctx, rev, vrf_id)
    try:
        return addresses[ip_prefix_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# Evpn
###############################
def vrf_evpn_get(ctx, rev, vrf_id):
    output = vrf_get(ctx, rev, vrf_id)
    try:
        return output["evpn"]
    except KeyError:
        raise NotFound


###############################
# L3 VNIs
###############################

def vrf_vnis_get(ctx, rev, vrf_id):
    output = vrf_evpn_get(ctx, rev, vrf_id)
    try:
        return output["vni"]
    except KeyError:
        raise NotFound


###############################
# L3 VNI
###############################

def vrf_vni_get(ctx, rev, vrf_id, vni_id):
    vnis = vrf_vnis_get(ctx, rev, vrf_id)
    try:
        return vnis[vni_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# PTP
###############################

def ptp_get(ctx, rev, vrf_id):
    if rev == 'operational':
        # PTP under bridge.vlan is purely a config concept.  Return the last
        # applied config as the operational state.
        rev = 'applied'

    return vrf_get(ctx, rev, vrf_id)["ptp"]


# ###############################
# # PTP Timers
# ###############################
#
# def ptp_timers_get(ctx, rev, vrf_id):
#    return ptp_get(ctx, rev, vrf_id)["timers"]
#
#
###############################
# OSPF
###############################

def ospf_get(ctx, rev, vrf_id):
    if rev != "operational":
        return router_get(ctx, rev, vrf_id)["ospf"]

    # We will not show the operational state as of now.
    raise ctx.errors.OperationalNotSupported


###############################
# OSPF/Areas
###############################

def ospf_areas_get(ctx, rev, vrf_id):
    return ospf_get(ctx, rev, vrf_id)["area"]


###############################
# OSPF/area
###############################

def ospf_area_get(ctx, rev, vrf_id, area_id):
    areas = ospf_areas_get(ctx, rev, vrf_id)
    try:
        return areas[area_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


##################################
# OSPF/area/{area-id}/filter-list
##################################

def ospf_area_filter_list_get(ctx, rev, vrf_id, area_id):
    return ospf_area_get(ctx, rev, vrf_id, area_id)["filter-list"]


###############################
# OSPF/area/{area-id}/networks
###############################

def ospf_area_networks_get(ctx, rev, vrf_id, area_id):
    return ospf_area_get(ctx, rev, vrf_id, area_id)["network"]


####################################
# OSPF/area/{area-id}/networks/network
####################################

def ospf_area_network_get(ctx, rev, vrf_id, area_id, network_id):
    networks = ospf_area_networks_get(ctx, rev, vrf_id, area_id)
    try:
        return networks[network_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# OSPF/area/{area-id}/ranges
###############################

def ospf_area_ranges_get(ctx, rev, vrf_id, area_id):
    return ospf_area_get(ctx, rev, vrf_id, area_id)["range"]


####################################
# OSPF/area/{area-id}/ranges/range
####################################

def ospf_area_range_get(ctx, rev, vrf_id, area_id, range_id):
    ranges = ospf_area_ranges_get(ctx, rev, vrf_id, area_id)
    try:
        return ranges[range_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# OSPF/default-originate
###############################

def ospf_default_originate_get(ctx, rev, vrf_id):
    return ospf_get(ctx, rev, vrf_id)["default-originate"]


###############################
# OSPF/distance
###############################

def ospf_distance_get(ctx, rev, vrf_id):
    return ospf_get(ctx, rev, vrf_id)["distance"]


###############################
# OSPF/log
###############################

def ospf_log_get(ctx, rev, vrf_id):
    return ospf_get(ctx, rev, vrf_id)["log"]


###############################
# OSPF/max-metric
###############################

def ospf_max_metric_get(ctx, rev, vrf_id):
    return ospf_get(ctx, rev, vrf_id)["max-metric"]


###############################
# OSPF/route-redistribute
###############################

def ospf_route_redistribute_get(ctx, rev, vrf_id):
    return ospf_get(ctx, rev, vrf_id)["redistribute"]


#################################
# OSPF/route-redistribute/static
#################################

def ospf_redistribute_static_get(ctx, rev, vrf_id):
    return ospf_route_redistribute_get(ctx, rev, vrf_id)["static"]


####################################
# OSPF/route-redistribute/connected
####################################

def ospf_redistribute_connected_get(ctx, rev, vrf_id):
    return ospf_route_redistribute_get(ctx, rev, vrf_id)["connected"]


##################################
# OSPF/route-redistribute/kernel
##################################

def ospf_redistribute_kernel_get(ctx, rev, vrf_id):
    return ospf_route_redistribute_get(ctx, rev, vrf_id)["kernel"]


##################################
# OSPF/route-redistribute/bgp
##################################

def ospf_redistribute_bgp_get(ctx, rev, vrf_id):
    return ospf_route_redistribute_get(ctx, rev, vrf_id)["bgp"]


###############################
# OSPF/timers
###############################

def ospf_timers_get(ctx, rev, vrf_id):
    return ospf_get(ctx, rev, vrf_id)["timers"]


###############################
# OSPF/lsa-timers
###############################

def ospf_lsa_timers_get(ctx, rev, vrf_id):
    return ospf_timers_get(ctx, rev, vrf_id)["lsa"]


###############################
# OSPF/spf-timers
###############################

def ospf_spf_timers_get(ctx, rev, vrf_id):
    return ospf_timers_get(ctx, rev, vrf_id)["spf"]


###############################
# PIM
###############################

def pim_get(ctx, rev, vrf_id):
    if rev != "operational":
        return router_get(ctx, rev, vrf_id)["pim"]

    # We will not show the operational state as of now.
    raise ctx.errors.OperationalNotSupported


###############################
# PIM/timers
###############################

def pim_timers_get(ctx, rev, vrf_id):
    return pim_get(ctx, rev, vrf_id)["timers"]


###############################
# PIM/ecmp
###############################

def pim_ecmp_get(ctx, rev, vrf_id):
    return pim_get(ctx, rev, vrf_id)["ecmp"]


###############################
# PIM/msdp-mesh-groups
###############################

def pim_msdp_mesh_groups_get(ctx, rev, vrf_id):
    return pim_get(ctx, rev, vrf_id)["msdp-mesh-group"]


###############################
# PIM/msdp-mesh-group
###############################

def pim_msdp_mesh_group_get(ctx, rev, vrf_id, msdp_mesh_group_id):
    groups = pim_msdp_mesh_groups_get(ctx, rev, vrf_id)
    try:
        return groups[msdp_mesh_group_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# PIM/msdp-mesh-group/mesh-members
###############################

def pim_mesh_mbrs_get(ctx, rev, vrf_id, msdp_mesh_group_id):
    return pim_msdp_mesh_group_get(ctx, rev, vrf_id,
                                   msdp_mesh_group_id)["member-address"]


###############################
# PIM/msdp-mesh-group/mesh-member
###############################

def pim_mesh_mbr_get(ctx, rev, vrf_id, msdp_mesh_group_id, mesh_member_id):
    mbrs = pim_mesh_mbrs_get(ctx, rev, vrf_id, msdp_mesh_group_id)
    try:
        return mbrs[mesh_member_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# PIM/address-family
###############################

def pim_af_get(ctx, rev, vrf_id):
    return pim_get(ctx, rev, vrf_id)["address-family"]


###############################
# PIM/address-family/ipv4-unicast
###############################

def pim_af_ipv4_unicast_get(ctx, rev, vrf_id):
    return pim_af_get(ctx, rev, vrf_id)["ipv4-unicast"]


###############################
# PIM/address-family/ipv4-unicast/spt-switchover
###############################

def pim_spt_switchover_get(ctx, rev, vrf_id):
    return pim_af_ipv4_unicast_get(ctx, rev, vrf_id)["spt-switchover"]


###############################
# PIM/address-family/ipv4-unicast/rps
###############################

def pim_rps_get(ctx, rev, vrf_id):
    return pim_af_ipv4_unicast_get(ctx, rev, vrf_id)["rp"]


###############################
# PIM/address-family/ipv4-unicast/rp
###############################

def pim_rp_get(ctx, rev, vrf_id, rp_id):
    rps = pim_rps_get(ctx, rev, vrf_id)
    try:
        return rps[rp_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# PIM/address-family/ipv4-unicast/rp/group-ranges
###############################

def pim_rp_grp_ranges_get(ctx, rev, vrf_id, rp_id):
    return pim_rp_get(ctx, rev, vrf_id, rp_id)["group-range"]


###############################
# PIM/address-family/ipv4-unicast/rp/group-range
###############################

def pim_rp_grp_range_get(ctx, rev, vrf_id, rp_id, group_range_id):
    grp_ranges = pim_rp_grp_ranges_get(ctx, rev, vrf_id, rp_id)
    try:
        return grp_ranges[group_range_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist
