# Copyright (C) 2018, 2019 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


from . import templates
import logging

logger = logging.getLogger(__name__)


def prepare_unit(ctx):
    """
    Prepare the unit for lifecycle events.
    """
    # Tell the render engine about our templates
    ctx.render.register_package(templates)

    # Register for events
    ctx.events.verify_config.register("env_v1", handle_verify_config)
    ctx.events.ready_apply.register("env_v1", handle_ready_apply)
    ctx.events.check_health.register("env_v1", handle_check_health)


def config_pre_login_banner(evt, system):
    pre_login_msg = system['message'].get('pre-login', None)
    # Generate the scripts and config files
    evt.render(templates, "etc-prelogin-banner",
               message=pre_login_msg)

    # Update the config files for post login banner
    pre_login_file = '/etc/issue.net'
    evt.stage_file("etc-prelogin-banner", pre_login_file)

    # Update the pre-login banner config file in sshd_config
    snippets = evt.config_v1.getSnippets()
    sshd_config_snippet = snippets.get("sshd_config", "")
    evt.render(
        templates,
        "etc-sshd-config",
        pre_login_filename=pre_login_file,
        sshd_config_snippet=sshd_config_snippet,
    )
    sshd_staged = evt.stage_file("etc-sshd-config", "/etc/ssh/sshd_config")

    if sshd_staged:
        evt.schedule_restart("sshd")


def handle_verify_config(evt):
    """
    Handle a verify_config event.
    """


def config_post_login_banner(evt, system):
    post_login_msg = system['message'].get('post-login', None)
    # Generate the scripts and config files
    evt.render(templates, "etc_motd",
               message=post_login_msg)

    # Update the config files for post login banner
    evt.stage_file("etc_motd", "/etc/motd")


def handle_ready_apply(evt):
    """
    Prepare to apply the config.  Generate config files and scripts.  Also,
    look for issues in the config that would prevent us from applying it.
    """
    system = evt.config_v1.getSystem()
    hostname_value = system["hostname"]
    # We will always set the dhcp_lease_hostname to "no" when we set/unset
    # static hostname from nvue.
    configure_dhcp_lease_hostname = "no"

    if hostname_value != evt._operational_hostname:
        evt.report_warning((
            f"Warning: current hostname `{evt._operational_hostname}`"
            f" will be replaced with `{hostname_value}`"
        ))

    # Generate the scripts and config files
    evt.render(templates, "etc_hostname", hostname=hostname_value)
    evt.render(templates, "apply_hostname.sh")
    evt.render(templates, "etc_hosts",
               hostname=hostname_value)
    evt.render(templates, "etc_dhcp_exit_hook_hostname",
               configure=configure_dhcp_lease_hostname)

    # Make sure we can update our config files.
    evt.stage_file("etc_hosts", "/etc/hosts")
    evt.stage_file(
        "etc_dhcp_exit_hook_hostname",
        "/etc/dhcp/dhclient-exit-hooks.d/dhcp-sethostname"
    )
    if evt.stage_file("etc_hostname", "/etc/hostname"):
        evt.schedule_run_script("apply_hostname.sh")
        # Make sure hostnamectl is up-to-date with what we did.
        evt.schedule_restart("systemd-hostnamed.service")

    timezone_value = system.get('timezone')
    if timezone_value is not None:
        # Generate the scripts and config files
        evt.render(templates, "etc_timezone",
                   timezone=timezone_value)
        evt.render(templates, "apply_timezone.sh",
                   timezone=timezone_value)
        # Make sure we can update our config files.
        tz_staged = evt.stage_file("etc_timezone", "/etc/nvue-timezone")

        if tz_staged:
            evt.schedule_run_script("apply_timezone.sh")

    # Configure system post-login and pre-login banners
    system = evt.config_v1.getSystem()
    if 'message' in system:
        config_post_login_banner(evt, system)

    # Configure pre-login banner
    system = evt.config_v1.getSystem()
    if 'message' in system:
        config_pre_login_banner(evt, system)


def handle_check_health(evt):
    """
    Handle an check_health event.

    Use evt.report_error to report any found errors.
    Use evt.report_warning to report any found warnings.

    Warnings *should* clear on their own, given some time, usually indicating
    that a component is still "coming up."  Errors *will not* clear on their
    own.

    Args:
        evt: A CheckHealthEvent instance.
    """
