# Copyright (C) 2018, 2019 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


from textwrap import dedent
from pathlib import Path
import logging
import os
import json


logger = logging.getLogger(__name__)


# translate the pretty keys to something a bit more predictable
def _normalize_key(key):
    return key.lower().replace(" ", "-")


def ztp_script_get(ctx):
    ztp_scripts = []
    ztp_scripts.extend([
        Path(f) for f in ctx.fileops.list_files('/var/lib/cumulus/ztp')
        if f.name.startswith('ztp_script')
    ])
    list_of_files = sorted(
        ztp_scripts,
        key=lambda x: os.path.getmtime(x)
    )

    script = {}
    if list_of_files:
        # Get the latest file sorted according to time
        file = list_of_files[-1]
        script['location'] = str(file)
        # We are seeing issue with multi-line output display on NVUE:RM-2848042
        # We will comment out the native API until we can fix it in NVUE
        # script['content'] = ctx.fileops.load_file(file, default='')

    return script


def ztp_status_get(ctx):
    output = ctx.sh.sudo("/usr/sbin/ztp", "-j")
    json_out = json.loads(str(output))
    for _key, _val in json_out.items():
        if _key == 'state':
            json_out[_key] = _normalize_key(_val)
    return json_out


def proc_uptime_get(ctx):
    output = ctx.fileops.load_file("/proc/uptime", default='')
    return output.strip().split(' ')[0]


def lsb_release_get(ctx):
    output = ctx.fileops.load_file("/etc/lsb-release", default='')
    out = {}
    for line in output.split('\n'):
        line_list = line.split('=')
        if len(line_list) != 2:
            continue
        key, value = line_list
        out[key.strip()] = value.strip().replace('"', '')
    return out


def hostnamectl_get(ctx):
    logger.info("Running 'hostnamectl'")
    output = ctx.sh.sudo.hostnamectl()
    lines = output.splitlines()

    result = {}

    for line in lines:
        split = line.split(":", 1)

        key = _normalize_key(split[0].strip())
        value = split[1].strip()

        result[key] = value

    return result


def timedatectl_get(ctx):
    output = ctx.sh.sudo.timedatectl()
    lines = output.splitlines()

    result = {}

    for line in lines:
        split = line.split(":", 1)
        if len(split) != 2:
            continue

        key = _normalize_key(split[0].strip())
        value = split[1].strip()

        result[key] = value

    return result


"""
These are defined in `man dpkg-query` under `--showformat`:

"This option is used to specify  the  format  of  the  output  --show  will
produce  (short  option  since  dpkg 1.13.1).  The format is a string that
will be output for each package listed."
"""
_SHOWFORMAT_SPECIFIERS = [
    "Architecture",
    "Bugs",
    "Conffiles",
    "Config-Version",
    "Conflicts",
    "Breaks",
    "Depends",
    "Description",
    "Enhances",
    "Essential",
    "Filename",
    "Homepage",
    "Installed-Size",
    "MD5sum",
    "MSDOS-Filename",
    "Maintainer",
    "Origin",
    "Package",
    "Pre-Depends",
    "Priority",
    "Provides",
    "Recommends",
    "Replaces",
    "Section",
    "Size",
    "Source",
    "Status",
    "Suggests",
    "Tag",
    "Triggers-Awaited",
    "Triggers-Pending",
    "Version",
    "binary:Package",
    "binary:Synopsis",
    "binary:Summary",
    "db:Status-Abbrev",
    "db:Status-Want",
    "db:Status-Status",
    "db:Status-Eflag",
    "db-fsys:Files",
    "db-fsys:Last-Modified",
    "source:Package",
    "source:Version",
    "source:Upstream-Version",
]


def dpkg_query_show_get(ctx):
    format_fields = [specifier + "==${" + specifier + "}"
                     for specifier in _SHOWFORMAT_SPECIFIERS]
    format_string = "\\t".join(format_fields) + "\\n\\n\\t"

    output = ctx.sh.sudo("dpkg-query", "--show", "--showformat", format_string)

    entries = output.split("\n\n\t")[:-1]
    result = []

    for entry in entries:
        item = {}
        split_entry = entry.split("\t")
        for row in split_entry:
            key, _, value = row.partition("==")
            # some fields need extra parsing
            if key == "Conffiles" or key == "db-fsys:Files":
                value = dedent(value).splitlines()
            elif key == "Description":
                summary_and_detail = value.split("\n", 1)
                summary = summary_and_detail[0]
                detail = (
                    # detail is indented a single space when we get it
                    dedent(summary_and_detail[1])
                    if len(summary_and_detail) > 1
                    else None
                )
                value = {
                    "summary": summary,
                    "detail": detail
                }

            item[key] = value

        result.append(item)

    return result


def load_file_sshd_config_get(ctx):
    output = ctx.fileops.load_file("/etc/ssh/sshd_config", default='')
    return output


def load_file_motd_config_get(ctx):
    output = ctx.fileops.load_file("/etc/motd", default='')
    return output


def load_file_pre_login_banner_config_get(ctx, file_name):
    output = ctx.fileops.load_file(file_name, default='')
    return output
