# Copyright (C) 2018, 2019 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from datetime import datetime


def platform_get(ctx):
    return {
        "software": software_get(ctx),
    }


def hostname_get(ctx):
    try:
        # If transient hostname is displayed, that means the operational
        # hostname is different from what's in /etc/hostname.
        return ctx.env_v1.getHostnamectl()['transient-hostname']
    except KeyError:
        # When the kernel is using the hostname from /etc/hostname, it doesn't
        # display a transient hostname.
        return ctx.env_v1.getHostnamectl()['static-hostname']


def system_ztp_script_get(ctx):
    return ctx.env_v1.getSystemZtpScript()


def system_uptime_get(ctx):
    uptime = ctx.env_v1.getSystemProcUptime()
    # Remove the decimals from the uptime string
    return int(float(uptime))


def system_ztp_status_get(ctx):
    ztp_status = ctx.env_v1.getSystemZtpStatus()
    status = {}
    date = ztp_status.get('date', '')
    if date:
        dateformat = datetime.strptime(date, '%a %b %d %H:%M:%S %Y %Z')
        status['date'] = dateformat.strftime('%Y-%m-%dT%H:%M:%SZ')
    status['version'] = ztp_status.get('version', '')
    status['url'] = ztp_status.get('url', '')
    state = ztp_status.get('state', None)
    if state:
        status['state'] = state
    status['method'] = ztp_status.get('method', '')
    result = ztp_status.get('result', None)
    if result:
        status['result'] = result
    return status


def system_build_get(ctx):
    lsb_release_info = ctx.env_v1.getSystemLsbRelease()
    return lsb_release_info.get('DISTRIB_DESCRIPTION', '')


def timezone_get(ctx):
    date_time = ctx.env_v1.getTimedatectl()
    timezone = date_time.get('time-zone', '')
    if timezone != '':
        return timezone.split(" ", 1)[0]
    else:
        return None


def software_get(ctx):
    return {
        "installed": installed_packages_get(ctx)
    }


def installed_packages_get(ctx):
    native = ctx.env_v1.getDpkgQueryShow()
    result = {}
    for item in native:
        package = item["Package"]
        result[package] = {
            "package": package,
            "version": item["Version"],
            "description": item["Description"]["summary"],
        }
    return result


def installed_package_get(ctx, installed_id):
    return installed_packages_get(ctx).get(installed_id)


###############################
# System Message
###############################

def post_login_message_get(ctx):
    msg = ctx.env_v1.getMotdConfigFile()
    return msg


def pre_login_message_get(ctx):
    msg = ctx.env_v1.getSshConfigFile()
    banner_file = ''
    for line in msg.splitlines():
        line = line.strip()
        if line.startswith("Banner"):
            banner_file = line.split()[1].strip()
            break

    banner_msg = ''
    if banner_file:
        banner_msg = \
            ctx.env_v1.getPreLoginBannerConfigFile(banner_file)

    return banner_msg
