# Copyright (C) 2018, 2019 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


# Translate the native format into something close to the public format.  This
# is the abstration layer between the public object model and the
# implementation.

from pydash import py_
import logging
from cue.exceptions import NotFound
from collections import defaultdict
logger = logging.getLogger("cue_frr_v1")


nested_dict = lambda: defaultdict(nested_dict) # noqa


def interface_by_name_get(ctx, interface_id):
    native_out = ctx.frr_v1.getNativeInterfaceByName(interface_id)

    # `show interface <name> json` only returns one interface
    interface_id = list(native_out.keys())[0]
    if_data = native_out[interface_id]

    xlated = nested_dict()

    if "flags" in if_data:
        # Remove "<" & ">"
        flags = if_data["flags"][1:-1]

        for f in flags.split(","):
            if f == "UP":
                xlated["link"]["flag"]["up"] = {}
            elif f == "LOOPBACK":
                xlated["link"]["flag"]["loopback"] = {}
            elif f == "BROADCAST":
                xlated["link"]["flag"]["broadcast"] = {}
            elif f == "MULTICAST":
                xlated["link"]["flag"]["multicast"] = {}

    # mtu excluding loopbacks
    if "mtu" in if_data and if_data["mtu"] <= 9216:
        xlated["link"]["mtu"] = if_data["mtu"]

    if "hardwareAddress" in if_data:
        xlated["link"]["mac"] = if_data["hardwareAddress"]

    if "description" in if_data:
        xlated["description"] = if_data["description"]

    if "type" in if_data:
        # loopback
        if if_data["type"] == "Loopback":
            xlated["type"] = "loopback"

        elif if_data["type"] == "Ethernet":
            # bond
            if "interfaceType" in if_data and \
                    if_data["interfaceType"] == "bond":
                xlated["type"] = "bond"

            # svi
            elif "interfaceType" in if_data and \
                    if_data["interfaceType"] == "Vlan":
                xlated["type"] = "svi"

            # swp
            elif interface_id.startswith("swp"):  # lol
                xlated["type"] = "swp"

            # eth
            else:
                xlated["type"] = "eth"

        # TODO handle peerlink and sub ifaces

    if "vlanId" in if_data:
        xlated["vlan"] = if_data["vlanId"]

    # MH
    if "evpnMh" in if_data and if_data["evpnMh"] != {}:
        xlated["evpn"]["multihoming"]["segment"]["enable"] = "on"
        if "esId" in if_data["evpnMh"]:
            xlated["evpn"]["multihoming"]["segment"]["identifier"] = \
                if_data["evpnMh"]["esId"]
        if "esSysmac" in if_data["evpnMh"]:
            xlated["evpn"]["multihoming"]["segment"]["mac-address"] = \
                if_data["evpnMh"]["esSysmac"]
        if "uplink" in if_data["evpnMh"] and if_data["evpnMh"]["uplink"] == \
           "up":
            xlated["evpn"]["multihoming"]["uplink"] = "on"

    if "parentInterface" in if_data:
        xlated["parent"] = if_data["parentInterface"]

    if "administrativeStatus" in if_data:
        if if_data["administrativeStatus"] == "up":
            xlated["link"]["state"]["up"] = {}

    if "neighborIpAddresses" in if_data:
        for neigh in if_data["neighborIpAddresses"]:
            # strip subnet
            neigh_addr = neigh.split("/")[0]

            if ":" in neigh:  # v6
                xlated["ip"]["neighbor"]["ipv6"][neigh_addr] = {}
            else:  # v4
                xlated["ip"]["neighbor"]["ipv4"][neigh_addr] = {}

    # te
    if "trafficEngineeringLinkParameters" in if_data:
        xlated["link"]["traffic-engineering"] = {
            "metric":
            if_data["trafficEngineeringLinkParameters"]["teMetric"],
            "max-bandwidth":
            if_data["trafficEngineeringLinkParameters"]["maximumBandwidth"],
            "max-reservable-bandwidth":
            if_data[
                "trafficEngineeringLinkParameters"
            ][
                "maximumReservableBandwidth"
            ],
            "unreserved-bandwidth":
            if_data["trafficEngineeringLinkParameters"]["unreservedBandwidth"],
            "available-bandwidth":
            if_data["trafficEngineeringLinkParameters"]["availableBandwidth"],
            "residual-bandwidth":
            if_data["trafficEngineeringLinkParameters"]["residualBandwidth"],
            "utilized-bandwidth":
            if_data["trafficEngineeringLinkParameters"]["utilizedBandwidth"],
        }

    return xlated


def vrf_vni_get(ctx, vrf_id):
    vrf_vni = {}
    vni = {}
    found = False

    native_out = ctx.frr_v1.getNativeVrfVni()
    # native_out = ctx.frr_v1.native.vrf_vni_get()

    for vrf in native_out['vrfs']:
        if vrf['vrf'] == vrf_id:
            found = True
            break
    if found:
        vni_id = str(vrf['vni'])
        vni[vni_id] = {}
        vni[vni_id]['routermac'] = vrf['routerMac']
        vni[vni_id]['state'] = vrf['state'].lower()
        vni[vni_id]['svi'] = vrf['sviIntf']
        if vrf['vxlanIntf'] != 'None':
            vni[vni_id]['vxlanif'] = vrf['vxlanIntf']
        vrf_vni['vni'] = vni

        # vrf_vni['vni'] = {vni_id: {}}
        # vrf_vni['vni'][vni_id]['svi'] = vrf['sviIntf']

    return vrf_vni


def _vrf_route_table_parse(native_out):
    route = {}
    routes = dict()

    for prefix, val in native_out.items():
        if prefix not in route:
            route[prefix] = {}

        idx = 0
        for p_entry in val:
            entry = {}
            proto = p_entry['protocol']

            if py_.get(route[prefix], 'protocol') is None:
                route[prefix]['protocol'] = dict()

            if py_.get(route[prefix], 'protocol.' + proto) is None:
                route[prefix]['protocol'][proto] = dict()

            if 'entry-index' not in route[prefix]['protocol'][proto]:
                route[prefix]['protocol'][proto]['entry-index'] = dict()
                idx = 0
            entry = route[prefix]['protocol'][proto]['entry-index']

            # create a new entry under the proto.
            # for multiple entries under the same proto increment index.
            idx += 1
            e_idx = str(idx)
            entry[e_idx] = dict()

            entry[e_idx]['distance'] = p_entry['distance']
            entry[e_idx]['metric'] = p_entry['metric']
            entry[e_idx]['uptime'] = p_entry['uptime']

            table = py_.get(p_entry, 'table')
            if table:
                entry[e_idx]['table-id'] = table
            nhg = py_.get(p_entry, 'internalNextHopNum')
            if nhg:
                entry[e_idx]['nexthop-group-id'] = nhg

            tag = py_.get(p_entry, 'tag')
            if tag:
                entry[e_idx]['tag'] = tag

            entry[e_idx]['flags'] = {}
            f_selected = py_.get(p_entry, 'selected')
            if f_selected:
                entry[e_idx]['flags']['selected'] = {}
            f_fibselected = py_.get(p_entry, 'destSelected')
            if f_fibselected:
                entry[e_idx]['flags']['fib-selected'] = {}
            f_installed = py_.get(p_entry, 'installed')
            if f_installed:
                entry[e_idx]['flags']['installed'] = {}
            f_queued = py_.get(p_entry, 'queued')
            if f_queued:
                entry[e_idx]['flags']['queued'] = {}
            f_offloaded = py_.get(p_entry, 'offloaded')
            if f_offloaded:
                entry[e_idx]['flags']['offloaded'] = {}
            f_failed = py_.get(p_entry, 'failed')
            if f_failed:
                entry[e_idx]['flags']['failed'] = {}

            if 'via' not in entry[e_idx]:
                # create empty nexthop list
                entry[e_idx]['via'] = {}
            # Iterate nexthops and fill in nexthops under an entry.
            for n_nexthop in p_entry['nexthops']:
                nexthop = {}
                nexthop['flags'] = {}
                is_resolver = False

                # Find the nhop key
                if py_.get(n_nexthop, 'blackhole'):
                    nhopk = 'blackhole'
                if py_.get(n_nexthop, 'reject'):
                    nhopk = 'reject'
                ip = py_.get(n_nexthop, 'ip')
                if ip:
                    nhopk = ip
                ifname = py_.get(n_nexthop, 'interfaceName')
                if ifname and ip is None:
                    nhopk = ifname

                # check for a condition to have a route entry
                # points to same nexthop
                # with different interface
                # if nhopk in entry[e_idx]['via']:
                #     nexthop = entry[e_idx]['via'][nhopk]

                if py_.get(n_nexthop, 'recursive') is not None:
                    re_nhopk = ip  # nhopk

                # nexthop has resolver flag
                if py_.get(n_nexthop, 'resolver') is not None and \
                        nhopk in entry[e_idx]['via'][re_nhopk]['resolved-via']:
                    is_resolver = True
                    nexthop = \
                        entry[e_idx]['via'][re_nhopk]['resolved-via'][nhopk]

                # set nexthop type based on nhop key
                if nhopk == 'blackhole':
                    nexthop['type'] = 'blackhole'
                if nhopk == 'reject':
                    nexthop['type'] = 'reject'
                if py_.get(n_nexthop, 'ip') is not None:
                    nexthop['type'] = 'ip-address'
                # nexthop is type interface
                if ifname and ip is None:
                    nexthop['type'] = 'interface'

                if ifname and ip:
                    if 'interface' not in nexthop:
                        nexthop['interface'] = {}
                    nexthop['interface'].update({ifname: {}})

                # nexthop has recursive flag
                if py_.get(n_nexthop, 'recursive') is not None:
                    # set recursive flag
                    nexthop['flags']['recursive'] = {}
                    # create resolver dict
                    nexthop['resolved-via'] = dict()
                if py_.get(n_nexthop, 'resolver') is not None:
                    is_resolver = True
                if py_.get(n_nexthop, 'active') is not None:
                    nexthop['flags']['active'] = {}
                if py_.get(n_nexthop, 'onLink') is not None:
                    nexthop['flags']['onlink'] = {}
                if py_.get(n_nexthop, 'fib') is not None:
                    nexthop['flags']['installed'] = {}
                if py_.get(n_nexthop, 'duplicate') is not None:
                    nexthop['flags']['duplicate'] = {}
                if py_.get(n_nexthop, 'directlyConnected') is not None:
                    nexthop['flags']['directly-connected'] = {}
                if py_.get(n_nexthop, 'unreachable'):
                    nexthop['flags']['unreachable'] = {}

                weight = py_.get(n_nexthop, 'weight')
                if weight:
                    nexthop['weight'] = weight

                nvrf = py_.get(n_nexthop, 'vrf')
                if nvrf:
                    nexthop['vrf'] = nvrf

                # Append the nexthop to either via or resolved_via dict.
                if is_resolver is False:
                    entry[e_idx]['via'].update({nhopk: nexthop})
                else:
                    entry[e_idx]['via'][re_nhopk]['resolved-via']\
                        .update({nhopk: nexthop})

    routes['route'] = route
    return routes


def vrf_v4_route_table_get(ctx, vrf_id):

    native_out = ctx.frr_v1.getVrfIpRoute(vrf_id)

    return _vrf_route_table_parse(native_out)


def vrf_v6_route_table_get(ctx, vrf_id):

    native_out = ctx.frr_v1.getVrfIpv6Route(vrf_id)

    return _vrf_route_table_parse(native_out)


def _vrf_bgp_route_table_parse(native_out):
    route = {}
    routes = dict()
    if not native_out:
        raise NotFound
    val = native_out['routes']
    if not val:
        raise NotFound
    pre_index = 0
    for _r_key, r_value in val.items():
        pre = py_.get(r_value[pre_index], 'prefix')
        if pre is not None:
            prefix = pre
        else:
            prefix = _r_key
        route[prefix] = {}
        path_index = 0

        # create a dict for path
        route[prefix]['path'] = dict()
        for r_entry in r_value:
            if 'advertisedTo' in r_entry:
                advertised_to = {}
                for akey in r_entry['advertisedTo']:
                    advertised_to[akey] = {}
                    advertised_to[akey]['hostname'] = \
                        r_entry['advertisedTo'][akey]['hostname']
                route[prefix]['advertised-to'] = advertised_to
                continue

            elif 'aspath' in r_entry:
                path_index += 1
                p_index = str(path_index)

                # multiple paths per route
                route[prefix]['path'][p_index] = {}
                path_entry = route[prefix]['path'][p_index]

                # parse the aspath list
                path_entry['aspath'] = dict()
                for apath in r_entry['aspath']['segments']:
                    for path in apath['list']:
                        p = str(path)
                        path_entry['aspath'][p] = {}

                # parse the bestpath
                bestpath = py_.get(r_entry, 'bestpath')
                if bestpath:
                    path_entry['bestpath'] = {}
                    bestpathfromas = \
                        py_.get(r_entry['bestpath'], 'bestpathFromAs')
                    if bestpathfromas:
                        path_entry['bestpath']['from-as'] = \
                            r_entry['bestpath']['bestpathFromAs']
                    overall = py_.get(r_entry['bestpath'], 'overall')
                    if overall is not None:
                        if overall:
                            path_entry['bestpath']['overall'] = 'on'
                        else:
                            path_entry['bestpath']['overall'] = 'off'
                    selectionreason = \
                        py_.get(r_entry['bestpath'], 'selectionReason')
                    if selectionreason:
                        path_entry['bestpath']['selection-reason'] = \
                            r_entry['bestpath']['selectionReason']

                origin = py_.get(r_entry, 'origin')
                if origin:
                    path_entry['origin'] = origin
                valid = py_.get(r_entry, 'valid')
                if valid is not None:
                    if valid:
                        path_entry['valid'] = 'on'
                    else:
                        path_entry['valid'] = 'off'
                weight = py_.get(r_entry, 'weight')
                if weight:
                    path_entry['weight'] = weight
                med = py_.get(r_entry, 'metric')
                if med is not None:
                    path_entry['metric'] = med
                lp = py_.get(r_entry, 'locPrf')
                if lp:
                    path_entry['local-pref'] = lp

                extendedcommunity = py_.get(r_entry, 'extendedCommunity')
                if extendedcommunity:
                    path_entry['ext-community'] = dict()
                    contents = r_entry['extendedCommunity']['string'].split()
                    for c in contents:
                        path_entry['ext-community'][c] = {}

                community = py_.get(r_entry, 'community')
                if community:
                    path_entry['community'] = dict()
                    contents = r_entry['community']['string'].split()
                    for c in contents:
                        path_entry['community'][c] = {}

                largecommunity = py_.get(r_entry, 'largeCommunity')
                if largecommunity:
                    path_entry['large-community'] = dict()
                    contents = r_entry['largeCommunity']['string'].split()
                    for c in contents:
                        path_entry['large-community'][c] = {}

                epoch = py_.get(r_entry, 'lastUpdate')
                if epoch:
                    path_entry['last-update'] = epoch['epoch']

                # create empty flags list
                path_entry['flags'] = {}
                multipath = py_.get(r_entry, 'multipath')
                if multipath:
                    path_entry['flags']['multipath'] = {}
                stale = py_.get(r_entry, 'stale')
                if stale:
                    path_entry['flags']['stale'] = {}
                suppressed = py_.get(r_entry, 'suppressed')
                if suppressed:
                    path_entry['flags']['suppressed'] = {}
                removed = py_.get(r_entry, 'removed')
                if removed:
                    path_entry['flags']['removed'] = {}

                path_entry['peer'] = {}
                r_peer = r_entry['peer']
                rid = py_.get(r_peer, 'routerId')
                if rid:
                    path_entry['peer']['router-id'] = rid
                hname = py_.get(r_peer, 'hostname')
                if hname:
                    path_entry['peer']['hostname'] = hname
                interface = py_.get(r_peer, 'interface')
                if interface:
                    path_entry['peer']['interface'] = interface
                tp = py_.get(r_peer, 'type')
                if tp:
                    path_entry['peer']['type'] = tp
                    path_entry['path-from'] = tp

                peer_id = py_.get(r_peer, 'peerId')
                if peer_id:
                    path_entry['peer']['id'] = peer_id

                indx = 0
                # create a list of nexthops and add it to path
                path_entry['nexthop'] = {}
                for n_nexthop in r_entry['nexthops']:
                    nexthop = {}

                    afi = py_.get(n_nexthop, 'afi')
                    if afi:
                        nexthop['afi'] = afi

                    # hostname = py_.get(n_nexthop, 'hostname')
                    # if hostname:
                    #     nexthop['hostname'] = hostname

                    used = py_.get(n_nexthop, 'used')
                    if used is not None:
                        if used:
                            nexthop['used'] = 'on'
                        else:
                            nexthop['used'] = 'off'

                    scope = py_.get(n_nexthop, 'scope')
                    if scope:
                        nexthop['scope'] = scope

                    metric = py_.get(n_nexthop, 'metric')
                    if metric is not None:
                        nexthop['metric'] = n_nexthop['metric']

                    ip = py_.get(n_nexthop, 'ip')
                    if ip:
                        nexthop['ip'] = ip

                    accessible = py_.get(n_nexthop, 'accessible')
                    if accessible is not None:
                        if accessible:
                            nexthop['accessible'] = 'on'
                        else:
                            nexthop['accessible'] = 'off'

                    indx += 1
                    idx = str(indx)
                    path_entry['nexthop'][idx] = nexthop
                    continue

                path_entry['nexthop-count'] = indx
                route[prefix]['path-count'] = path_index

    routes['route'] = route
    return routes


def vrf_bgp_v4_unicast_route_table_get(ctx, vrf_id):

    native_out = ctx.frr_v1.getBgpAfUnicastRouteNtv(vrf_id, "ipv4")

    return _vrf_bgp_route_table_parse(native_out)


def vrf_bgp_v6_unicast_route_table_get(ctx, vrf_id):

    native_out = ctx.frr_v1.getBgpAfUnicastRouteNtv(vrf_id, "ipv6")

    return _vrf_bgp_route_table_parse(native_out)


def _translate_gr_info(info, addr_fam):
    """
    maps to (address_family, endOfRibSend)  and (address_family, endOfRibRecv)
    """
    gr_restart = {}
    end_of_rib = py_.get(info, 'endOfRibSend')
    tx_eof_rib = py_.get(end_of_rib, addr_fam)
    if tx_eof_rib:
        gr_restart['tx-eof-rib'] = 'on'
    else:
        gr_restart['tx-eof-rib'] = 'off'

    end_of_rib = py_.get(info, 'endOfRibRecv')
    rx_eof_rib = py_.get(end_of_rib, addr_fam)
    if rx_eof_rib:
        gr_restart['rx-eof-rib'] = 'on'
    else:
        gr_restart['rx-eof-rib'] = 'off'
    return gr_restart


def _translate_addr_cap(neighbor_cap_info, graceful_restart_info, addr_fam):
    addr_capabilities = {}
    """
    maps to gracefulRestart:fBit:TBD
    addr_capabilities['tx-restart-f-bit'] =
    """
    """
    maps to multiprotocol extension, the capability assignment is based on
    whether the value has advertised or recieved
    """
    neigh_cap_info = py_.get(neighbor_cap_info, "multiprotocolExtensions")
    if neigh_cap_info is None:
        return addr_capabilities
    info = py_.get(neigh_cap_info, addr_fam)
    if info is None:
        return addr_capabilities
    advertised = py_.get(info, 'advertised')
    advertisedandreceived = py_.get(info, 'advertisedAndReceived')
    received = py_.get(info, 'received')
    """
    if either of the configs is present the capability is true
    """
    if advertised or advertisedandreceived:
        addr_capabilities['tx-mpbgp'] = 'on'
    else:
        addr_capabilities['tx-mpbgp'] = 'off'

    if received or advertisedandreceived:
        addr_capabilities['rx-mpbgp'] = 'on'
    else:
        addr_capabilities['rx-mpbgp'] = 'off'
    """
    address family graceful restart info
    """
    gr_info = neighbor_cap_info['gracefulRestart']
    if gr_info == 'advertisedAndReceived':
        addr_capabilities['rx-graceful-restart'] = 'on'
    else:
        addr_capabilities['rx-graceful-restart'] = 'off'
    """
    maps to tx graceful restart info that is absent:TBD
    addr_capabilities['tx-graceful-restart'] = 'False'
    """
    """
    graceful restart f-bit
    """
    addr_fam_info = py_.get(graceful_restart_info, addr_fam)
    if addr_fam_info:
        f_bit = graceful_restart_info[addr_fam]['fBit']
        if f_bit:
            addr_capabilities['rx-restart-f-bit'] = 'on'
        else:
            addr_capabilities['rx-restart-f-bit'] = 'off'
    """
    tx-restart-f-bit: TBD support in json missing
    addr_capabilities['tx-restart-f-bit'] =
    """

    """
    maps to addPath:{address_family}{...}
    """
    addpath = neighbor_cap_info['addPath'][addr_fam]
    rx_advrcv = py_.get(addpath, 'rxAdvertisedAndReceived')
    tx_advrcv = py_.get(addpath, 'txAdvertisedAndRecieved')
    if rx_advrcv:
        addr_capabilities['rx-addpath'] = 'on'
    else:
        addr_capabilities['rx-addpath'] = 'off'
    if tx_advrcv:
        addr_capabilities['tx-addpath'] = 'on'
    else:
        addr_capabilities['tx-addpath'] = 'off'
    return addr_capabilities


def _translate_policy(info):
    """
    policy is mapped to routeMapForOutgoingAdvertisements
    and routeMapForIncomingAdvertisements
    """
    policy = {}
    inbound_policy = {}
    outbound_policy = {}
    val = py_.get(info, 'incomingUpdatePrefixFilterList')
    if val:
        inbound_policy['prefix-list'] = val
    val = py_.get(info, 'incomingUpdateAsPathFilterList')
    if val:
        inbound_policy['aspath-list'] = val
    val = py_.get(info, 'routeMapForIncomingAdvertisements')
    if val:
        inbound_policy['route-map'] = val
    outbound_policy = {}
    val = py_.get(info, 'outgoingUpdatePrefixFilterList')
    if val:
        outbound_policy['prefix-list'] = val
    val = py_.get(info, 'outgoingUpdateAsPathFilterList')
    if val:
        outbound_policy['aspath-list'] = val
    val = py_.get(info, 'selectiveUnsuppressRouteMap')
    if val:
        outbound_policy['unsuppress-map'] = val
    val = py_.get(info, 'routeMapForOutgoingAdvertisements')
    if val:
        outbound_policy['route-map'] = val
    policy['inbound'] = inbound_policy
    policy['outbound'] = outbound_policy
    return policy


def _translate_attr_mode(info):
    """
    attribute mode maps to the attr-unchanged flags
    unchangedAsPathPropogatedToNbr, unchangedNextHopPropogatedToNbr
    unchangedMedPropogatedToNbr
    """
    attr_mod = {}
    attr_val = py_.get(info, 'unchangedNextHopPropogatedToNbr')
    if attr_val:
        attr_mod['nexthop'] = 'on'
    else:
        attr_mod['nexthop'] = 'off'
    attr_val = py_.get(info, 'unchangedMedPropogatedToNbr')
    if attr_val:
        attr_mod['med'] = 'on'
    else:
        attr_mod['med'] = 'off'
    attr_val = py_.get(info, 'unchangedAsPathPropogatedToNbr')
    if attr_val:
        attr_mod['aspath'] = 'on'
    else:
        attr_mod['aspath'] = 'off'
    return attr_mod


def _translate_route_policy(info):
    """
    applicable to ipv4 and ipv6 route map information
    if the default route map config is present the assign the
    route map and set enable to true.
    """
    default_route_origination = {}
    def_route_map = py_.get(info, 'defaultRouteMap')
    if def_route_map:
        default_route_origination['policy'] = def_route_map
        default_route_origination['enable'] = 'on'
    return default_route_origination


def _translate_prefix_limits(info):
    """
    ipv4/ipv6 prefix limit configuration
    """
    prefix_limits = {}
    inbound = {}
    threshold = py_.get(info, 'prefixAllowedWarningThresh')
    if threshold:
        inbound['warning-threshold'] = threshold
    allowedmax = py_.get(info, 'prefixAllowedMaxWarning')
    if allowedmax:
        inbound['warning-only'] = 'on'
    else:
        inbound['warning-only'] = 'off'

    intervals = py_.get(info, 'prefixAllowedRestartIntervalMsecs')
    if intervals:
        inbound['reestablish-wait'] = intervals
    allowed_max = py_.get(info, 'prefixAllowedMax')
    if allowed_max:
        inbound['maximum'] = allowed_max
    prefix_limits['inbound'] = inbound
    return prefix_limits


def _translate_cond_adv(info):
    """
    condtional advertise configuration
    """
    conditional_adv = {}
    conditional_adv['non-exist-map'] = info['condition']
    conditional_adv['exist-map'] = info['conditionMap']
    conditional_adv['advertise-map'] = info['advertiseMap']
    conditional_adv['enable'] = 'on'
    return conditional_adv


def _translate_community_adv(info):
    """
    community advertise attributes
    """
    community_adv = {}
    community_adv_info = info['commAttriSentToNbr']
    if 'large' in community_adv_info:
        community_adv['large'] = 'on'
    else:
        community_adv['large'] = 'off'
    if 'extendedAndStandard' in community_adv_info:
        community_adv['regular'] = 'on'
    else:
        community_adv['regular'] = 'off'
    if 'extended' in community_adv_info:
        community_adv['extended'] = 'on'
    else:
        community_adv['extended'] = 'off'
    return community_adv


def _translate_timer(info):
    """
    timers{} holds peer timer info and is sub-tree of neighbor
    """
    timers = {}
    timers['hold'] = info['bgpTimerHoldTimeMsecs']
    timers['keepalive'] = info['bgpTimerKeepAliveIntervalMsecs']
    timers['connection-retry'] = info['connectRetryTimer']
    return timers


def _translate_nexthop(info):
    """
    nexthop{} holds the nh info and is subtree of neighbor
    """
    nexthop = {}
    next_hop = py_.get(info, 'nexthop')
    if next_hop:
        nexthop['ipv4'] = info['nexthop']
        nexthop['ipv6-global'] = info['nexthopGlobal']
        nexthop['ipv6-local'] = info['nexthopLocal']
    return nexthop


def _translate_ebgp(info):
    """
    aligning to ebgp-requires-policy
    """
    ebgp = {}
    outbound = py_.get(info, 'outboundEbgpRequiresPolicy')
    if outbound:
        ebgp['outbound'] = {}
    else:
        ebgp['outbound'] = {}
    inbound = py_.get(info, 'inboundEbgpRequiresPolicy')
    if inbound:
        ebgp['inbound'] = {}
    else:
        ebgp['inbound'] = {}
    return ebgp


def _translate_msg_stats(stats_info):
    msg_stats = {}
    msg_stats['input-queue'] = stats_info['depthInq']
    msg_stats['output-queue'] = stats_info['depthOutq']
    msg_stats['tx-opens'] = stats_info['opensSent']
    msg_stats['rx-opens'] = stats_info['opensRecv']
    msg_stats['tx-keepalives'] = stats_info['keepalivesSent']
    msg_stats['rx-keepalives'] = stats_info['keepalivesRecv']
    msg_stats['tx-route-refreshes'] = stats_info['routeRefreshRecv']
    msg_stats['rx-route-refreshes'] = stats_info['routeRefreshRecv']
    msg_stats['tx-total'] = stats_info['totalSent']
    msg_stats['rx-total'] = stats_info['totalRecv']
    return msg_stats


def _translate_peer_capabilities(info):
    """
    capabilites{} will hold the neigh capabilities as subtree of neighbor
    """
    capabilities = {}
    neighbor_cap_info = py_.get(info, 'neighborCapabilities')
    if neighbor_cap_info:
        asn_info = neighbor_cap_info['4byteAs']
        if 'advertised' in asn_info:
            capabilities['tx-asn32'] = 'on'
        else:
            capabilities['tx-asn32'] = 'off'
        if 'received' in asn_info:
            capabilities['rx-asn32'] = 'on'
        else:
            capabilities['rx-asn32'] = 'off'
        ext_nh_info = neighbor_cap_info['extendedNexthop']
        ext_nh = {}
        if 'advertised' in ext_nh_info:
            ext_nh['advertised'] = 'on'
            capabilities['extended-nexthop'] = 'on'
        else:
            ext_nh['advertised'] = 'off'
            capabilities['extended-nexthop'] = 'off'
        if 'received' in ext_nh_info:
            ext_nh['received'] = 'on'
            capabilities['rx-extended-nexthop'] = 'on'
        else:
            ext_nh['received'] = 'on'
            capabilities['rx-extended-nexthop'] = 'off'
        refresh_info = neighbor_cap_info['routeRefresh']
        if 'advertised' in refresh_info:
            capabilities['tx-route-refresh'] = 'on'
        else:
            capabilities['tx-route-refresh'] = 'off'
        if 'Received' in refresh_info:
            capabilities['rx-route-refresh'] = 'on'
        else:
            capabilities['rx-route-refresh'] = 'off'
        """
        Rx GR is True if:
          - gracefulRestart has value "advertisedAndReceived" OR
          - gracefulRestartCapability has value "received”
        """
        gr_info = neighbor_cap_info['gracefulRestart']
        gr_cap_info = py_.get(neighbor_cap_info, 'gracefulRestartCapability')
        if (gr_info == 'advertisedAndReceived'):
            capabilities['rx-graceful-restart'] = 'on'
        elif (gr_cap_info and 'received' in gr_cap_info):
            capabilities['rx-graceful-restart'] = 'on'
        else:
            capabilities['rx-graceful-restart'] = 'off'
        if (gr_info == 'advertisedAndReceived'):
            capabilities['tx-graceful-restart'] = 'on'
        elif (gr_cap_info and 'advertised' in gr_cap_info):
            capabilities['tx-graceful-restart'] = 'on'
        else:
            capabilities['tx-graceful-restart'] = 'off'
        graceful_restart_info = info['gracefulRestartInfo']
        if graceful_restart_info['rBit']:
            capabilities['rx-restart-r-bit'] = 'on'
        else:
            capabilities['rx-restart-r-bit'] = 'off'
    return capabilities


def _translate_ttl_security(info):
    """
    ttl-security maps to externalBgpNbrMaxHopsAway
    """
    ttl_security = {}
    hop_info = py_.get(info, 'externalBgpNbrMaxHopsAway')
    if hop_info:
        ttl_security['hops'] = hop_info
        ttl_security['enable'] = 'on'
    else:
        ttl_security['enable'] = 'off'
    return ttl_security


def _parse_and_populate_l2vpnevpn(l2vpn_info, graceful_restart_info,
                                  neighbor_cap_info):

    l2vpn_evpn = {}
    """
    graceful restart info
    """
    l2vpn_evpn['graceful-restart'] = _translate_gr_info(graceful_restart_info,
                                                        'l2VpnEvpn')
    """
    address capabilties
    """
    l2vpn_evpn['capabilities'] = _translate_addr_cap(neighbor_cap_info,
                                                     graceful_restart_info,
                                                     'l2VpnEvpn')
    """
    attribute mode
    """
    l2vpn_evpn['attribute-mod'] = _translate_attr_mode(l2vpn_info)
    """
    policy
    """
    l2vpn_evpn['policy'] = _translate_policy(l2vpn_info)
    """
    aspath:TBD
    """
    """
    Other config's
    """
    update_grp = py_.get(l2vpn_info, 'updateGroupId')
    if update_grp:
        l2vpn_evpn['update-group'] = update_grp
    rx_prefix = py_.get(l2vpn_info, 'acceptedPrefixCounter')
    if rx_prefix:
        l2vpn_evpn['rx-prefix'] = rx_prefix
    tx_prefix = py_.get(l2vpn_info, 'sentPrefixCounter')
    if tx_prefix:
        l2vpn_evpn['tx-prefix'] = tx_prefix
    soft_config = py_.get(l2vpn_info, 'inboundSoftConfigPermit')
    if soft_config:
        l2vpn_evpn['soft-reconfiguration'] = 'on'
    route_config = py_.get(l2vpn_info, 'routeServerClient')
    if route_config:
        l2vpn_evpn['route-server-client'] = 'on'
    route_config = py_.get(l2vpn_info, 'routeReflectorClient')
    if route_config:
        l2vpn_evpn['route-reflector-client'] = 'on'
    return l2vpn_evpn


def _parse_and_populate_ipv4unicast(ipv4_info, graceful_restart_info,
                                    neighbor_cap_info):

    ipv4_unicast = {}
    """
    graceful restart info
    """
    ipv4_unicast['graceful-restart'] = _translate_gr_info(
     graceful_restart_info, 'ipv4Unicast')
    """
    address capabilties
    """
    ipv4_unicast['capabilities'] = _translate_addr_cap(neighbor_cap_info,
                                                       graceful_restart_info,
                                                       'ipv4Unicast')
    """
    attribute mode
    """
    ipv4_unicast['attribute-mod'] = _translate_attr_mode(ipv4_info)
    """
    policy
    """
    ipv4_unicast['policy'] = _translate_policy(ipv4_info)
    """
    aspath:TBD
    """
    """
    default route policy
    """
    def_sent = py_.get(ipv4_info, 'defaultSent')
    def_notsent = py_.get(ipv4_info, 'defaultNotSent')
    if def_sent or def_notsent:
        ipv4_unicast['default-route-origination'] = \
         _translate_route_policy(ipv4_info)
    """
    prefix limitation
    """
    ipv4_unicast['prefix-limits'] = _translate_prefix_limits(ipv4_info)
    """
    community advertise
    """
    ipv4_unicast['community-advertise'] = _translate_community_adv(ipv4_info)
    """
    conditional advertise
    """
    adv_info = py_.get(ipv4_info, 'advertiseMap')
    if adv_info:
        ipv4_unicast['conditional-advertise'] = _translate_cond_adv(adv_info)
    """
    Other config's
    """
    update_grp = py_.get(ipv4_info, 'updateGroupId')
    if update_grp:
        ipv4_unicast['update-group'] = update_grp
    rx_prefix = py_.get(ipv4_info, 'acceptedPrefixCounter')
    if rx_prefix:
        ipv4_unicast['rx-prefix'] = rx_prefix
    tx_prefix = py_.get(ipv4_info, 'sentPrefixCounter')
    if tx_prefix:
        ipv4_unicast['tx-prefix'] = tx_prefix
    soft_config = py_.get(ipv4_info, 'inboundSoftConfigPermit')
    if soft_config:
        ipv4_unicast['soft-reconfiguration'] = 'on'
    route_config = py_.get(ipv4_info, 'routeServerClient')
    if route_config:
        ipv4_unicast['route-server-client'] = 'on'
    route_config = py_.get(ipv4_info, 'routeReflectorClient')
    if route_config:
        ipv4_unicast['route-reflector-client'] = 'on'
    return ipv4_unicast


def _parse_and_populate_ipv6unicast(ipv6_info, graceful_restart_info,
                                    neighbor_cap_info):
    ipv6_unicast = {}
    """
    graceful restart info
    """
    ipv6_unicast['graceful-restart'] = _translate_gr_info(
     graceful_restart_info, 'ipv6Unicast')
    """
    address capabilties
    """
    ipv6_unicast['capabilities'] = _translate_addr_cap(neighbor_cap_info,
                                                       graceful_restart_info,
                                                       'ipv6Unicast')
    """
    attribute mode
    """
    ipv6_unicast['attribute-mod'] = _translate_attr_mode(ipv6_info)
    """
    policy
    """
    ipv6_unicast['policy'] = _translate_policy(ipv6_info)
    """
    aspath:TBD
    """
    """
    default route policy
    """
    def_sent = py_.get(ipv6_info, 'defaultSent')
    def_notsent = py_.get(ipv6_info, 'defaultNotSent')
    if def_sent or def_notsent:
        ipv6_unicast['default-route-origination'] = \
         _translate_route_policy(ipv6_info)
    """
    prefix limitations
    """
    ipv6_unicast['prefix-limits'] = _translate_prefix_limits(ipv6_info)
    """
    community advertise
    """
    ipv6_unicast['community-advertise'] = _translate_community_adv(ipv6_info)
    """
    conditional advertise
    """
    adv_info = py_.get(ipv6_info, 'advertiseMap')
    if adv_info:
        ipv6_unicast['conditional-advertise'] = _translate_cond_adv(adv_info)
    """
    community advertise
    """
    """
    Other config's
    """
    update_grp = py_.get(ipv6_info, 'updateGroupId')
    if update_grp:
        ipv6_unicast['update-group'] = update_grp
    rx_prefix = py_.get(ipv6_info, 'acceptedPrefixCounter')
    if rx_prefix:
        ipv6_unicast['rx-prefix'] = rx_prefix
    tx_prefix = py_.get(ipv6_info, 'sentPrefixCounter')
    if tx_prefix:
        ipv6_unicast['tx-prefix'] = tx_prefix
    soft_config = py_.get(ipv6_info, 'inboundSoftConfigPermit')
    if soft_config:
        ipv6_unicast['soft-reconfiguration'] = 'on'
    route_config = py_.get(ipv6_info, 'routeServerClient')
    if route_config:
        ipv6_unicast['route-server-client'] = 'on'
    route_config = py_.get(ipv6_info, 'routeReflectorClient')
    if route_config:
        ipv6_unicast['route-reflector-client'] = 'on'
    return ipv6_unicast


def _bgp_parsed_info(native_out, neigh):
    """
    _bgp_parsed_info will  parse the data from the json file and
    populate a dictionary to return.
    """
    peer_name = str(neigh)
    neighbor = {}

    info = native_out[peer_name]
    """""""""""""""""""""""""""""""""""
    neighbor holds top level peer config
    """""""""""""""""""""""""""""""""""
    neighbor['bgp-version'] = info['bgpVersion']
    neighbor['remote-router-id'] = info['remoteRouterId']
    neighbor['state'] = info['bgpState'].lower()
    timer_up_str = py_.get(info, 'bgpTimerUpString')
    if timer_up_str:
        neighbor['uptime'] = timer_up_str
    timer_up_msec = py_.get(info, 'bgpTimerUpMsec')
    if timer_up_msec:
        neighbor['uptime-msec'] = timer_up_msec
    neighbor['connections-established'] = info['connectionsEstablished']
    neighbor['connections-dropped'] = info['connectionsDropped']
    neighbor['last-reset-reason'] = info['lastResetDueTo']

    bgp_connection = py_.get(info, 'bgpConnection')
    if bgp_connection == 'sharedNetwork':
        neighbor['connection-type'] = 'shared-network'
    elif bgp_connection == 'nonSharedNetwork':
        neighbor['connection-type'] = 'non-shared-network'

    neighbor['last-reset-timer'] = info['lastResetTimerMsecs']
    neighbor['last-reset-code'] = info['lastResetCode']
    local_host = py_.get(info, 'hostLocal')
    if local_host:
        neighbor['local-ip'] = info['hostLocal']
    local_port = py_.get(info, 'portLocal')
    if local_port:
        neighbor['local-port'] = local_port
    foreign_host = py_.get(info, 'hostForeign')
    if foreign_host:
        neighbor['remote-ip'] = foreign_host
    foreign_port = py_.get(info, 'portForeign')
    if foreign_port:
        neighbor['remote-port'] = foreign_port
    """
    fill in the capability from the host info
    """
    neighbor_cap_info = py_.get(info, 'neighborCapabilities')
    if neighbor_cap_info:
        host_name_info = py_.get(neighbor_cap_info, 'hostName')
        if host_name_info:
            neighbor['local-hostname'] = host_name_info['advHostName']
            neighbor['local-domain'] = host_name_info['advDomainName']
            neighbor['remote-hostname'] = host_name_info['rcvHostName']
            neighbor['remote-domain'] = host_name_info['rcvDomainName']
    if info['remoteAs'] != 0:
        neighbor['remote-as'] = info['remoteAs']
    """
    peer timer info
    """
    neighbor['timers'] = _translate_timer(info)

    """
    peer nexthop info
    """
    neighbor['nexthop'] = _translate_nexthop(info)

    """
    peer ebgp policy TBD: fix needed in OM
    """
    neighbor['ebgp-policy'] = _translate_ebgp(info)

    """
    msg_stats{} info
    """
    stats_info = info['messageStats']
    neighbor['message-stats'] = _translate_msg_stats(stats_info)

    """
    capabilites{} info
    """
    neighbor['capabilities'] = _translate_peer_capabilities(info)

    """
    ttl-security info
    """
    neighbor['ttl-security'] = _translate_ttl_security(info)
    """
    local as info
    """
    local_as = {}
    local_as_info = info['localAs']
    if local_as_info:
        local_as['asn'] = local_as_info

    address_family = {}
    addr_family_info = info['addressFamilyInfo']
    graceful_restart_info = info['gracefulRestartInfo']
    """
    Retrieve and fill in the l2vpnEvpn information
    """
    l2vpn_info = py_.get(addr_family_info, 'l2VpnEvpn')
    if l2vpn_info:
        address_family['l2vpn-evpn'] = _parse_and_populate_l2vpnevpn(
         l2vpn_info, graceful_restart_info, neighbor_cap_info)

    """
    Retrieve and fill in the ipv4 unicast information
    """
    ipv4_info = py_.get(addr_family_info, 'ipv4Unicast')
    if ipv4_info:
        address_family['ipv4-unicast'] = _parse_and_populate_ipv4unicast(
         ipv4_info, graceful_restart_info, neighbor_cap_info)

    """
    Retrieve and fill in the ipv6 unicast information
    """
    ipv6_info = py_.get(addr_family_info, 'ipv6Unicast')
    if ipv6_info:
        address_family['ipv6-unicast'] = _parse_and_populate_ipv6unicast(
         ipv6_info, graceful_restart_info, neighbor_cap_info)

    neighbor['address-family'] = address_family
    return neighbor


def bgp_neighbor_trans_get(ctx, vrf_id, neighbor_id):
    """
    bpg_neighbors_get will take get the json info from the native api call
    and call the parse api.
    """
    nbr_trans_out = {}
    native_out = ctx.frr_v1.getNativeBgpVrfNeighbor(vrf_id, neighbor_id)
    prefix = (','.join(list(native_out.keys())))
    if prefix != neighbor_id:
        raise NotFound
    else:
        nbr_trans_out = _bgp_parsed_info(native_out, neighbor_id)

    return nbr_trans_out


def bgp_neighbors_trans_get(ctx, vrf_id):
    """
    bpg_neighbors_get will take get the json info from the native api call
    and call the parse api.
    """
    nbr_trans_out = {}
    native_out = ctx.frr_v1.getBgpNeighborsByVrf(vrf_id)

    if native_out:
        for prefix in native_out.keys():
            nbr_trans_out[prefix] = {}
            nbr_trans_out[prefix] = _bgp_parsed_info(native_out, prefix)

    return nbr_trans_out
