#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of Nvidia Corporation
# and its affiliates (the "Company") and all right, title, and interest
# in and to the software product, including all associated intellectual
# property rights, are and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
# Copyright (C) 2021 Cumulus Networks, inc.
# All Rights reserved.


from . import templates
from pydash import py_
import logging
import ipaddress
import cue_linuxptp_v1.show_ptp4l_cli as show_ptp4l_cli
logger = logging.getLogger(__name__)


def prepare_unit(ctx):
    """
    Prepare the unit for lifecycle events.
    """
    # Tell the render engine about our templates
    ctx.render.register_package(templates)

    # Register for events
    ctx.events.verify_config.register("linuxptp_v1", handle_verify_config)
    ctx.events.ready_apply.register("linuxptp_v1", handle_ready_apply)
    ctx.events.check_health.register("linuxptp_v1", handle_check_health)


def handle_verify_config(evt):
    """
    Handle a verify_config event.

    Use evt.config_v1 to access the pending configuration's state.
    Use evt.report_issue to report any found issues.

    Args:
        evt: A VerifyConfigEvent instance.
    """


def format_acceptable_masters_clocks(ptps):
    """
    The config file expects clock-ids in a different format.
    """
    for ptp in ptps.values():
        new_masters = {}
        for clock_id, acc in ptp["acceptable-master"].items():
            parts = clock_id.split(":")
            new_clock_id = "{}{}{}.{}{}.{}{}{}".format(*parts)
            new_masters[new_clock_id] = acc
        ptp["acceptable-master"] = new_masters


def find_svi_interface(all_ifaces, domain_name, vid):
    svi_name = None
    # For now, if we're dealing with br_default, just generate the svi's name.
    # Long term, we shouldn't be assuming a naming convention, but for that to
    # work, we need base-interface and vlan to be populated.
    if domain_name == "br_default":
        svi_name = "vlan" + str(vid)
    else:
        # If not br-default, look for a matching svi.
        for iface_name, iface in all_ifaces.items():
            if iface["type"] != "svi":
                continue
            if (domain_name == iface.get("base-interface")
                    and str(vid) == str(iface.get("vlan"))):
                svi_name = iface_name
                break

    if svi_name:
        try:
            iface = all_ifaces[svi_name]

            # Found a match.  Get it's first IP address.
            if iface["ip"]["address"]:
                ip_prefix = next(iter(iface["ip"]["address"]))
                ip_addr = str(ipaddress.ip_interface(ip_prefix).ip)

                # Got an IP address, too.  Return it.
                return svi_name, ip_addr
            else:
                return None
        except (KeyError, IndexError, ValueError):
            pass

    return None


def find_active_vlan(domains, domain_name, vlans):
    # Find one that's active.
    for vid, vlan in vlans.items():
        try:
            vlan = domains[domain_name]["vlan"][vid]
            candidate = vlan["ptp"]
            if candidate["enable"] == "on":
                return vid, candidate
        except KeyError:
            pass

    return None


def inherit_interface_enable(ifaces, all_ifaces, vrfs, domains):
    """
    Determine if the interface should be enabled.
    """
    for iface in ifaces.values():
        if_ptp = iface["ptp"]

        # If we cannot inherit, then assume our parent is disabled.
        parent = {"enable": "off"}

        # If the iface has an IP address, call it an L3 interface.  Try to
        # inherit from the VRF.
        vrf = iface["ip"]["vrf"]

        if "bridge" in iface:
            # It's an L2 interface, try to inherit from VLAN.
            for domain_name, domain in iface["bridge"]["domain"].items():
                # If the interface is in "all" VLANs, loop over the
                # domain's vlans.  Otherwise, loop over the interface's
                # vlans.
                # Only Trunk port has vlan, access port has 'access'
                if "vlan" in domain:
                    if "all" in domain["vlan"]:
                        vlans = domains[domain_name]["vlan"]
                    else:
                        vlans = domain["vlan"]
                    if_ptp["_l2-mode"] = 'trunk'
                elif "access" in domain:
                    vlan = str(domain["access"])
                    vlans = {}
                    vlans[vlan] = domains[domain_name]["vlan"][vlan]
                    if_ptp["_l2-mode"] = 'access'

                # Look for an active VLAN in this bridge
                active_vlan = find_active_vlan(domains, domain_name, vlans)
                if active_vlan is None:
                    continue
                vid, parent = active_vlan

                # While we know where we found a match, fill in the SVI
                # props
                svi_tup = find_svi_interface(all_ifaces, domain_name, vid)
                if svi_tup:
                    svi_name, ip_addr = svi_tup
                    if_ptp["_vlan-interface"] = svi_name
                    if_ptp["_source-ip"] = ip_addr
        else:
            # Otherwise, it is an L3 interfce
            if vrf in vrfs:
                parent = vrfs[vrf]["ptp"]

        # If our parent is disabled, then we're disabled.
        if parent["enable"] == "off":
            if_ptp["enable"] = "off"


def find_active_interfaces(ifaces):
    """
    Filter ifaces for ones with PTP enabled.
    """
    enabled_ifaces = {}
    for iface_name, iface in ifaces.items():
        if_ptp = py_.get(iface, "ptp", {"enable": "off"})

        # If explicitly on (or inherited on), add it to the list of
        # interface instances
        if if_ptp["enable"] == "on":
            if iface["type"] != 'svi':
                enabled_ifaces[iface_name] = iface

    return enabled_ifaces


def is_software(evt):
    """
    Return True if this looks like a hardware platform.
    """
    hw = evt.platform_v1.getHardware()
    if (hw.get("manufacturer") == "Cumulus" and hw.get("model") == "VX"):
        return True
    else:
        return False


def handle_ready_apply(evt):
    """
    Prepare to apply the config.  Generate config files and scripts.  Also,
    look for issues in the config that would prevent us from applying it.

    Use evt.config_v1 to access the pending configuration's state.
    Use evt.render to render a template with the given variables.
    Use evt.stage_file to prepare a rendered file to be copied over to its
        ultimate destination.

    Args:
        evt: A ReadyApplyEvent instance.
    """
    ptps = evt.config_v1.getPtps()
    if_ptps = None

    # Check if running in HW or SW
    sw = is_software(evt)

    # For now, assume a single instance, named "1".
    if "1" in ptps and ptps["1"]["enable"] == "on":
        format_acceptable_masters_clocks(ptps)
        ptp = ptps["1"]

        # Find the enabled PTP interfaces
        all_ifaces = evt.config_v1.getInterfaces()
        ifaces = find_active_interfaces(all_ifaces)

        # Inherit values
        vrfs = evt.config_v1.getVrfs()
        domains = evt.config_v1.getBridgeDomains()
        inherit_interface_enable(ifaces, all_ifaces, vrfs, domains)

        # Some of our interfaces might be disabled now.  Filter again.
        ifaces = find_active_interfaces(ifaces)

        # Pull out just the PTP service
        if_ptps = {
            iface_name: iface["ptp"]
            for iface_name, iface in ifaces.items()
        }
        logger.debug(if_ptps)
    if if_ptps:
        # Reload ptp when ptp interfaces exist
        evt.render(
            templates, "ptp4l.conf",
            py_=py_,
            running=True,
            simulation=sw,
            ptp_conf=ptp,
            if_ptps=if_ptps
        )
        if evt.stage_file("ptp4l.conf", "/etc/ptp4l.conf"):
            evt.schedule_enable("ptp4l")
            evt.schedule_reload_or_restart("ptp4l")
            if not sw:
                evt.schedule_enable("phc2sys")
                evt.schedule_reload_or_restart("phc2sys")

        show_ptp4l_cli.ptp_domain_read = 0
    else:
        # If disabled or no interfaces configured, render a config for a
        # stopped service
        evt.render(templates, "ptp4l.conf", running=False)
        if evt.stage_file("ptp4l.conf", "/etc/ptp4l.conf"):
            evt.schedule_disable("ptp4l")
            evt.schedule_stop("ptp4l")
            if not sw:
                evt.schedule_disable("phc2sys")
                evt.schedule_stop("phc2sys")


def handle_check_health(evt):
    """
    Handle a check_health event.

    Use evt.report_error to report any found errors.
    Use evt.report_warning to report any found warnings.

    Warnings *should* clear on their own, given some time, usually indicating
    that a component is still "coming up."  Errors *will not* clear on their
    own.

    Args:
        evt: A CheckHealthEvent instance.
    """
