# Copyright (C) 2018, 2019 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


# Translate the native format into something close to the public format.  This
# is the abstration layer between the public object model and the
# implementation.


from cue.exceptions import NotFound
from cue.utils import format_checker


#############################
# Neighbors
#############################
def nbrs_get(ctx):
    """
    Get neighbors for all interface and translate to OM.
    """
    # Get the LLDP data
    nbr_list = ctx.lldp_v1.getLldpCtls()
    nbr_data = {}
    for intf in nbr_list.keys():
        nbr_data[intf] = _make_nbr(nbr_list[intf])
    return nbr_data


def nbr_get(ctx, interface_id):
    """
    Get a neighbor for an interface, swp_id and translate to OM.
    """
    try:
        # Get the LLDP data
        nbr_list = ctx.lldp_v1.getLldpCtl(interface_id)

        return _make_nbr(nbr_list)
    except NotFound:
        # We can't tell if the interface is unknown or if it doesn't
        # have neighbors.  Leave the "unknown" case to the upper layers
        # of CUE and assume the "no neighbors" case.
        return {}


def _make_nbr(nbr_list):
    """
    Translate the nbr object from lldpctl lldp neighbor OM
    """
    trans_nbr_list = {}

    for nbr in nbr_list:
        # Pluck out what we need, defaulting as appropriate.

        age = ''
        if "age" in nbr:
            # convert age day, HH:MM:SS format to seconds
            age_str = nbr["age"]
            split = "days," if "days" in age_str else "day,"
            day, time_str = age_str.split(split)
            h, m, s = time_str.split(':')
            age = int(day) * 24 * 3600 + int(h) * 3600 + int(m) * 60 + int(s)

        # lldp neighbor chassis related properties
        neigh_id = ""
        chassis_id = "00:00:00:00:00:00"
        sys_name = ""
        sys_descr = ""
        capability = {}
        chassis_mgmt_ipv6 = ''
        chassis_mgmt_ipv4 = ''

        try:
            neigh_id = nbr["chassis"][0]["name"][0]["value"]

            if format_checker.is_mac_address(
                nbr["chassis"][0]["id"][0]["value"]
            ):
                chassis_id = nbr["chassis"][0]["id"][0]["value"]
            else:
                # The chassis id is not in proper format
                chassis_id = "00:00:00:00:00:00"
            sys_name = nbr["chassis"][0]["name"][0]["value"]
            sys_descr = nbr["chassis"][0]["descr"][0]["value"]

            for ip in nbr["chassis"][0]["mgmt-ip"]:
                if format_checker.is_ipv4_unicast(ip["value"]):
                    chassis_mgmt_ipv4 = ip["value"]
                elif not format_checker.is_ipv4_multicast(ip["value"]):
                    chassis_mgmt_ipv6 = ip["value"]

            cap_bridge = False
            cap_router = False
            for cap in nbr["chassis"][0]["capability"]:
                if cap["type"] == "Bridge" and cap["enabled"]:
                    cap_bridge = True
                if cap["type"] == "Router" and cap["enabled"]:
                    cap_router = True

            if cap_bridge:
                capability["is-bridge"] = {}
            if cap_router:
                capability["is-router"] = {}
        except KeyError:
            # Just show the configurations fetched and default for others
            pass

        trans_nbr_chassis = {
            "chassis-id": chassis_id,
            "system-name": sys_name,
            "system-description": sys_descr,
            "capability": capability,
        }
        if chassis_mgmt_ipv4:
            trans_nbr_chassis['management-address-ipv4'] = \
                chassis_mgmt_ipv4
        if chassis_mgmt_ipv6:
            trans_nbr_chassis['management-address-ipv6'] = \
                chassis_mgmt_ipv6

        # lldp port related properties
        pmd_autoneg = {}
        auto_neg = {}
        ttl = 0
        port_id = ""
        port_id_type = ""
        port_descr = ""
        try:
            ttl = int(nbr["port"][0]["ttl"][0]["value"])
            port_id = nbr["port"][0]["id"][0]["value"]
            port_id_type = nbr["port"][0]["id"][0]["type"]
            port_descr = nbr["port"][0]["descr"][0]["value"]

            auto_neg = nbr["port"][0]["auto-negotiation"][0]
            pmd_autoneg["autoneg"] = {}
            if auto_neg["supported"]:
                pmd_autoneg["autoneg"]["is-supported"] = {}
            if auto_neg["enabled"]:
                pmd_autoneg["autoneg"]["is-enabled"] = {}
            idx = 0
            pmd_autoneg["advertised"] = {}
            for adv_cap in auto_neg.get("advertised", []):
                if str(idx) not in pmd_autoneg["advertised"]:
                    pmd_autoneg["advertised"][str(idx)] = {}
                pmd_autoneg["advertised"][str(idx)]["type"] = adv_cap["type"]
                pmd_autoneg["advertised"][str(idx)]["flags"] = {}
                if adv_cap["hd"]:
                    pmd_autoneg["advertised"][str(idx)]["flags"]["hd"] = {}
                if adv_cap["fd"]:
                    pmd_autoneg["advertised"][str(idx)]["flags"]["fd"] = {}
                idx += 1
            pmd_autoneg["mau-oper-type"] = auto_neg["current"][0]["value"]
        except KeyError:
            # Just show the configurations fetched and default for others
            pass

        trans_nbr_port = {
            "ttl": ttl,
            "name": port_id,
            "type": port_id_type,
            "description": port_descr,
            "pmd-autoneg": pmd_autoneg,
        }

        # lldp med related properties
        inventory = {}
        med_capability = {}
        dev_type = ''
        if 'lldp-med' in nbr:
            dev_type = nbr["lldp-med"][0]["device-type"][0]["value"]
            for cap in nbr["lldp-med"][0]["capability"]:
                if cap["type"] == "Capabilities" and cap["available"]:
                    med_capability["capabilities"] = {}
                if cap["type"] == "Policy" and cap["available"]:
                    med_capability["policy"] = {}
                if cap["type"] == "Location" and cap["available"]:
                    med_capability["location"] = {}
                if cap["type"] == "MDI/PSE" and cap["available"]:
                    med_capability["mdi-pse"] = {}
                if cap["type"] == "MDI/PD" and cap["available"]:
                    med_capability["mdi-pd"] = {}
                if cap["type"] == "Inventory" and cap["available"]:
                    med_capability["inventory"] = {}

            if "inventory" in nbr["lldp-med"][0]:
                nbr_lldp_inv = nbr["lldp-med"][0]["inventory"][0]
                inventory["software-revision"] = \
                    nbr_lldp_inv["software"][0]["value"]
                inventory["firmware-revision"] = \
                    nbr_lldp_inv["firmware"][0]["value"]
                inventory["serial-number"] = \
                    nbr_lldp_inv["serial"][0]["value"]
                inventory["manufacturer"] = \
                    nbr_lldp_inv["manufacturer"][0]["value"]
                inventory["model"] = \
                    nbr_lldp_inv["model"][0]["value"]

        trans_nbr_lldp_med = {
            "device-type": dev_type,
            "capability": med_capability,
            "inventory": inventory,
        }

        # LLDP Neighbor Bridge Info
        bridge = {'vlan': {}}
        if "vlan" in nbr:
            for vlan in nbr['vlan']:
                if vlan['pvid'] is True:
                    bridge['untagged'] = int(vlan['vlan-id'])
                else:
                    bridge['vlan'][vlan['vlan-id']] = {}

        trans_nbr = {
            "age": str(age),
            "chassis": trans_nbr_chassis,
            "port": trans_nbr_port,
            "lldp-med": trans_nbr_lldp_med,
            "bridge": bridge
        }

        if neigh_id:
            trans_nbr_list[neigh_id] = trans_nbr

    # Put it in neighbor OM format and return it
    return trans_nbr_list


#############################
# Global Show LLDP Config
#############################
def config_get(ctx):
    """
    Get the global LLDP config
    """

    config = ctx.lldp_v1.getLldpcliConfig()

    if config:
        result = {
            'tx-interval': config['tx-delay'],
            'tx-hold-multiplier': config['tx-hold'],
            'dot1-tlv': 'on' if config['dot1-tlv'] == 'yes' else 'off'
        }
    else:
        result = {}

    return result
