#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of Nvidia Corporation
# and its affiliates (the "Company") and all right, title, and interest
# in and to the software product, including all associated intellectual
# property rights, are and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
# Copyright (C) 2018, 2019 Cumulus Networks, inc.
# All Rights reserved.


from . import templates


def get_state(config):
    res = ''
    try:
        response = config['admin-state']
        if response == 'enable':
            res = 'enable'
        elif response == 'disable':
            res = 'disable'
    except KeyError:
        return res

    return res


def get_provision(config):
    res = ''
    try:
        response = config['linecard']['provision']
        if response == 'NONE':
            res = 'DISABLE'
        else:
            res = response
    except KeyError:
        return res

    return res


def prepare_unit(ctx):
    """
    Prepare the unit for lifecycle events.
    """
    # Tell the render engine about our templates
    ctx.render.register_package(templates)

    # Register for events
    ctx.events.verify_config.register("platform_config_v1",
                                      handle_verify_config)
    ctx.events.ready_apply.register("platform_config_v1",
                                    handle_ready_apply)
    ctx.events.check_health.register("platform_config_v1",
                                     handle_check_health)


def handle_verify_config(evt):
    """
    Handle a verify_config event.

    Use evt.config_v1 to access the pending configuration's state.
    Use evt.report_issue to report any found issues.

    Args:
        evt: A VerifyConfigEvent instance.
    """


def handle_ready_apply(evt):
    """
    Prepare to apply the config.  Generate config files and scripts.  Also,
    look for issues in the config that would prevent us from applying it.

    Use evt.config_v1 to access the pending configuration's state.
    Use evt.render to render a template with the given variables.
    Use evt.stage_file to prepare a rendered file to be copied over to its
        ultimate destination.

    Args:
        evt: A ReadyApplyEvent instance.
    """
    # Get the platform object from config_v1.
    # You can call whatever API is most granular for your needs.
    platform = evt.config_v1.getPlatform()
    state = ''
    provision = ''

    compobj = platform['hardware']['component']

    for i in compobj.keys():
        if compobj[i]['type'] == 'linecard':
            comp_name = i
            slot = comp_name.strip('module')
            state = get_state(compobj[i])
            provision = get_provision(compobj[i])

    # Generate the scripts.
    # This is where you will render your script(s) from the templates.
    # These scripts will call chassismgrctl to change the state for each
    # component.
    if state != '':
        evt.render(templates, 'run_chassismgrctl.sh',
                   linecard=comp_name, slot=slot, state=state)
    elif provision != '':
        evt.render(templates, 'run_chassismgrctl.sh',
                   linecard=comp_name, slot=slot, provision=provision)
    else:
        evt.render(templates, 'run_chassismgrctl.sh',
                   platform=platform)

    # Tell CUED to run the script
    evt.schedule_run_script('run_chassismgrctl.sh')


def handle_check_health(evt):
    """
    Handle an check_health event.

    Use evt.report_error to report any found errors.
    Use evt.report_warning to report any found warnings.

    Warnings *should* clear on their own, given some time, usually indicating
    that a component is still "coming up."  Errors *will not* clear on their
    own.

    Args:
        evt: A CheckHealthEvent instance.
    """
