# Copyright (C) 2018, 2019 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


# Translate the native format into something close to the public format.  This
# is the abstration layer between the public object model and the
# implementation.

from cue import exceptions
from pydash import py_


def convert_memory_bytes(val, default=-1):
    if not isinstance(val, str):
        return default

    val = val.strip().upper()
    mem_bytes = default

    try:
        if val.endswith('G'):
            gb_val = float(val.split('G', 1)[0].strip())
            mem_bytes = int(gb_val * 1024.0 * 1024.0 * 1024.0)
        elif val.endswith('GB'):
            gb_val = float(val.split('GB', 1)[0].strip())
            mem_bytes = int(gb_val * 1024.0 * 1024.0 * 1024.0)
        elif val.endswith('KB'):
            kb_val = float(val.split('KB', 1)[0].strip())
            mem_bytes = int(kb_val * 1024.0)
        elif val.endswith('K'):
            kb_val = float(val.split('K', 1)[0].strip())
            mem_bytes = int(kb_val * 1024.0)
        elif val.endswith('MB'):
            mb_val = float(val.split('MB', 1)[0].strip())
            mem_bytes = int(mb_val * 1024.0 * 1024.0)
        elif val.endswith('M'):
            mb_val = float(val.split('M', 1)[0].strip())
            mem_bytes = int(mb_val * 1024.0 * 1024.0)
        elif val.endswith('B'):
            mem_bytes = int(val.split('B', 1)[0].strip())
        else:
            mem_bytes = int(val)
    except ValueError:
        mem_bytes = default

    return mem_bytes


def hardware_get(ctx):
    # Platform string from platform-detect
    response = ctx.platform_v1.getPlatformString()
    if not response:
        raise exceptions.NotFound

    vend = response.get('manufacturer')
    mod = response.get('model')
    mod = mod.strip('\n')
    res = {'manufacturer': vend.title(),
           'model': mod.upper()}

    # Platform information from file
    platforms = ctx.platform_v1.getPlatforms()
    soc_vendor = py_.get(platforms, f'{vend},{mod}.0.soc.vendor')
    soc_model = py_.get(platforms, f'{vend},{mod}.0.soc.model')
    if soc_vendor:
        res['asic-vendor'] = soc_vendor
    if soc_model:
        res['asic-model'] = soc_model

    cpu_info = py_.get(platforms, f'{vend},{mod}.0.cpu')
    if cpu_info and cpu_info['arch']:
        res['cpu'] = f'{cpu_info["arch"]} {cpu_info["vendor"]} '\
                     f'{cpu_info["model"]} {cpu_info["model_id"]} '\
                     f'{cpu_info["speed"]}'

    port_layout = py_.get(platforms, f'{vend},{mod}.0.ports.layout')
    if port_layout:
        res['port-layout'] = port_layout

    # System Memory and Harddisk
    mem_info = ctx.platform_v1.getMemory()
    tot_mem = py_.get(mem_info, 'total')
    if tot_mem:
        res['memory'] = convert_memory_bytes(tot_mem)

    disk_info = ctx.platform_v1.getDisk()
    size = py_.get(disk_info, 'size')
    if size:
        res['disk-size'] = convert_memory_bytes(size)

    # System MAC
    system_mac = ctx.platform_v1.getSystemMac()
    if system_mac:
        res['system-mac'] = system_mac

    # System eeprom data
    eeprom = ctx.platform_v1.getEeprom()
    if py_.get(eeprom, 'eeprom'):
        eeprom_data = eeprom['eeprom']
        base_mac = py_.get(eeprom_data, "tlv.Base MAC Address.value")
        if base_mac:
            res['base-mac'] = base_mac
        manufacturer = py_.get(eeprom_data, "tlv.Manufacturer.value")
        if manufacturer:
            res['manufacturer'] = manufacturer
        onie_ver = py_.get(eeprom_data, "tlv.ONIE Version.value")
        if onie_ver:
            res['onie-version'] = onie_ver
        part_num = py_.get(eeprom_data, "tlv.Part Number.value")
        if part_num:
            res['part-number'] = part_num
        plat_name = py_.get(eeprom_data, "tlv.Platform Name.value")
        if plat_name:
            res['platform-name'] = plat_name
        prod_name = py_.get(eeprom_data, "tlv.Product Name.value")
        if prod_name:
            res['product-name'] = prod_name
        ser_num = py_.get(eeprom_data, "tlv.Serial Number.value")
        if ser_num:
            res['serial-number'] = ser_num

    return res


###############################
# Breakout modes
###############################

def breakout_modes_get(ctx, component_id, port_id):
    if component_id != "device":
        # Currently, we only support simple, single-switch platforms and not
        # chassis, modules, etc.  Therefore, all ports are on the root
        # component, which we all "device."  If the request is for something
        # other than "device", then raise a NotFound for now.
        raise exceptions.NotFound

    res = {}
    port_modes = ctx.platform_v1.getPortModes()
    if not port_modes:
        raise exceptions.NotFound

    if port_id not in port_modes:
        # Unknown port
        raise ctx.errors.ItemDoesNotExist

    # Loop through the modes for the requested port
    modes = port_modes[port_id]
    for mode in modes:
        # Parse the mode to get the speed and breakout.
        if ('x' not in mode
                and '/' not in mode
                and mode != 'loopback'
                and mode != 'disabled'):
            breakout = '1x'
        else:
            breakout = mode

        # Add the new mode.
        py_.merge(res, {breakout: {}})

    return res


###############################
# Components
###############################

def components_get(ctx):
    linecards = ctx.platform_v1.getLinecardInfo()
    if not linecards:
        response = ctx.platform_v1.getPlatformString()
        if response:
            mod = response.get('model')
            mod = mod.strip('\n')
            res = {'type': 'switch',
                   'model': mod}
            dev = {'device': res}
            return dev
        else:
            raise exceptions.NotFound

    return linecards_get(linecards)


def component_get(ctx, component_id):
    linecard = linecard_get(ctx, component_id)
    if linecard is None:
        components = ctx.platform_v1.getComponents()
        if components:
            return components[component_id]
        else:
            raise ctx.errors.ItemDoesNotExist

    return linecard


def component_admin_state_get(ctx, component_id):
    if component_id[0:6] != 'module':
        return None

    comp_state = ctx.platform_v1.getComponentAdminStateInfo()
    slot = component_id.strip("module")
    res = comp_state[slot]
    return res["state"]


###############################
# Ports
###############################

def ports_get(ctx, component_id):
    res = {}
    ports = ctx.platform_v1.getPortModes()
    for port in ports:
        breakout = breakout_modes_get(ctx, "device", port)
        mode = {}
        mode["breakout-mode"] = breakout
        res[port] = mode
    return res


def port_get(ctx, component_id, port_id):
    port = {}
    port["breakout-mode"] = breakout_modes_get(ctx, "device", port_id)
    return port


###############################
# Fans
###############################

def fans_get(ctx):
    res = {}
    fans = ctx.platform_v1.getFanInfo()
    for fan in fans:
        name = fan.get('name')
        speed = fan.get('input')
        state = fan.get('state').lower()
        minspeed = fan.get('min')
        maxspeed = fan.get('max')
        py_.merge(res, {name: {"speed": speed,
                               "state": state,
                               "min-speed": minspeed,
                               "max-speed": maxspeed}})
    return res


def fan_get(ctx, fan_id):
    fans = fans_get(ctx)
    try:
        return fans[fan_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# Sensors
###############################
def temp_sensors_get(ctx):
    res = {}
    sensors = ctx.platform_v1.getSensorsInfo()
    for sensor in sensors:
        name = sensor.get('description')
        state = sensor.get('state').lower()
        temp = str(sensor.get('input'))
        mintemp = sensor.get('min')
        maxtemp = sensor.get('max')
        crittemp = sensor.get('crit')
        py_.merge(res, {name: {"temp": temp,
                               "state": state,
                               "min": mintemp,
                               "max": maxtemp,
                               "crit": crittemp}})
    return res


def temp_sensor_get(ctx, sensor_id):
    temp_sensors = temp_sensors_get(ctx)
    try:
        return temp_sensors[sensor_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# PSUs
###############################
def psus_get(ctx):
    res = {}
    psus = ctx.platform_v1.getPsuInfo()
    for psu in psus:
        name = psu.get('description')
        state = psu.get('state').lower()
        py_.merge(res, {name: {"state": state}})

    return res


def psu_get(ctx, psu_id):
    psus = psus_get(ctx)
    try:
        return psus[psu_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# LEDs
###############################
def leds_get(ctx):
    res = {}
    leds = ctx.platform_v1.getLedInfo()
    for led in leds:
        name = led.get('name')
        color = led.get('color')
        py_.merge(res, {name: {"color": color}})

    return res


def led_get(ctx, led_id):
    leds = leds_get(ctx)
    try:
        return leds[led_id]
    except KeyError:
        raise ctx.errors.ItemDoesNotExist


###############################
# Linecard
###############################
def linecards_get(linecards):
    res = {}
    for slot in linecards:
        res[slot] = {}
        res[slot]['state'] = linecards[slot]['state']
        res[slot]['serial'] = linecards[slot]['serial']
        res[slot]['type'] = 'linecard'
        res[slot]['model'] = linecards[slot]['card_type']

    return res


def linecard_get(ctx, component_id):
    # It doesn't look like a linecard's name.
    if component_id[0:6] != 'module':
        return None

    components = ctx.platform_v1.getLinecardInfo()
    linecards = linecards_get(components)
    if component_id in linecards:
        return linecards[component_id]

    # It's a linecard (by name), but not one we have.
    raise ctx.errors.ItemDoesNotExist


###############################
# Capabilities
###############################
def ifupdown2_capabilities_get(ctx):
    policies = ctx.platform_v1.getIfupdown2Policy()

    # check if ifupdown2 is in SVD mode
    single_vxlan_device_enabled = {
        "on": True,
        "yes": True,
        "enabled": True
    }.get(
        policies.get("vxlan", {})
        .get("module_globals", {})
        .get("single-vxlan-device"), False)

    return {
        "single-vxlan-device": 'on' if single_vxlan_device_enabled else 'off'
    }


###############################
# Reboot reason
###############################

def reboot_reason_get(ctx):
    trans_res = {}
    nat_res = ctx.platform_v1.getNativeRebootReason()
    key_mapping = {
        'reason': 'cause',
        'gentime': 'gen_time',
        'user': 'user',
    }
    for trans_key, nat_key in key_mapping.items():
        if nat_key in nat_res:
            if nat_res[nat_key] == 'null':
                trans_res[trans_key] = 'None'
            else:
                trans_res[trans_key] = nat_res[nat_key]

    return trans_res


def reboot_history_get(ctx):
    res = {}
    response = ctx.platform_v1.getNativeRebootHistory()
    for item in response.keys():
        reason = response[item].get('cause')
        gentime = response[item].get('gen_time')
        user = response[item].get('user')
        idx = str(item)
        py_.merge(res, {idx: {"reason": reason,
                              "gentime": gentime,
                              "user": user}})

    return res
