# Copyright 2021 NVIDIA Corporation


from . import templates
from . import native
from cue.utils.format_checker import patterns
from .validate_ports_errors import translate_port_error, get_ifname
import re
import logging
import time
from pydash import py_
from cue.utils import yaml_load
from pathlib import Path

logger = logging.getLogger(__name__)
swp_pattern = patterns['swp']


def _get_ifaces(evt):
    """
    Wrapping the API call so we can mock this function in
    test_lifecycle.py
    """
    ifaces = py_.get(evt.config_v1.getRoot(), 'interface', {})
    return ifaces


def _validate_ports(evt, port_modes, temp_ports_conf):
    cmd = "/usr/cumulus/bin/validate-ports -f {0}".format(temp_ports_conf)
    try:
        evt.sh.bash('-c', cmd)
        return []
    except evt.sh.ErrorReturnCode_1 as e:
        errors = str(e.stdout).replace("\\n", "\n")
        # if blocked port in error message then try disabling
        # the blocking port and rendering again
        ret = []
        for line in errors.splitlines():
            # TODO CUE-3192: Improve error reporting
            ret.append(translate_port_error(evt, line))
        return ret
    except Exception:
        evt.report_issue("Internal Error: Unable to validate ports")
        return []


def _get_default_ports_conf_fn(evt):
    platform_string = evt.platform_v1.getPlatformString()
    if not platform_string:
        logger.debug("No platform string")
        return ""

    mfg = platform_string['manufacturer']
    platform = platform_string['model']

    # read platform specific default ports.conf file to create portmap
    fn = ("/usr/share/cumulus-platform/{0}/{1}/etc/cumulus/ports.conf"
          .format(mfg, platform))
    return fn


speed_map = {
    "100M": 100,
    "1G": 1000,
    "10G": 10000,
    "25G": 25000,
    "40G": 40000,
    "50G": 50000,
    "100G": 100000,
    "200G": 200000}

breakout_map = {
    "1x": "",
    "2x": "2x",
    "4x": "4x",
    "/4": "/4",
    "/2": "/2",
    "disabled": "disabled",
    "loopback": "loopback"}


class PortConf:
    platform_port_info = None

    def __init__(self, evt):

        self.allowed_speeds = {}
        self.evt = evt
        self.port_modes = {}
        self.default_ports_conf_fn = _get_default_ports_conf_fn(self.evt)
        self.default_portmap = native.default_portmap_get(self.evt)

        self.port_modes = self.evt.platform_v1.getPortModes()
        if not self.port_modes:
            return

        for port in self.port_modes:
            speed_list = []
            for mode in self.port_modes[port]:
                for key in breakout_map:
                    mode = mode.replace(key, '')
                speed = speed_map.get(mode, None)
                if (speed):
                    speed_list.append(speed)
            self.allowed_speeds[port] = sorted(list(set(speed_list)))

    def get_platform_port_info(self):
        if not PortConf.platform_port_info:
            with open(Path(__file__).parent / 'platform_port_info.yaml',
                      'r') as f:
                PortConf.platform_port_info = yaml_load(f)
        return PortConf.platform_port_info

    def get_default_ports_conf_fn(self):
        return self.default_ports_conf_fn

    def get_port_modes(self):
        return self.port_modes

    def get_port_bandwidth(self, port, iface_speed):
        # If port is not in portmap, or cant be mapped then use
        # interface speed for port_bandwidth
        if port not in self.default_portmap.keys():
            return iface_speed  # pragma: no cover TODO CUE-3046
        mapped_speed = speed_map.get(iface_speed, None)
        if not mapped_speed:
            return iface_speed

        # Otherwise, use the iface speed to select the best
        # possible bandwidth from list of allowed speeds for the port
        if str(port) in self.allowed_speeds:
            for allowed_speed in self.allowed_speeds[str(port)]:
                if mapped_speed <= allowed_speed:
                    mapped_speed = allowed_speed
                    break
        else:  # pragma: no cover
            return iface_speed

        bandwidth = str(allowed_speed)

        if len(bandwidth) <= 3:  # pragma: no cover
            return bandwidth + 'M'
        else:
            return bandwidth[:-3] + "G"

    def get_platform_operational_portmap(self):
        return native.portmap_get(self.evt)

    def get_platform_default_portmap(self):
        return self.default_portmap


def prepare_unit(ctx):
    """
    Prepare the unit for lifecycle events.
    """
    # Tell the render engine about our templates
    ctx.render.register_package(templates)

    # Register for events
    ctx.events.verify_config.register("ports_v1", handle_verify_config)
    ctx.events.ready_apply.register("ports_v1", handle_ready_apply)
    ctx.events.check_health.register("ports_v1", handle_check_health)


def handle_verify_config(evt):
    """
    Handle a verify_config event.

    Use evt.config_v1 to access the pending configuration's state.
    Use evt.report_issue to report any found issues.

    Args:
        evt: A VerifyConfigEvent instance.
    """
    pf_string = evt.platform_v1.getPlatformString()
    model = py_.get(pf_string, 'model')
    if not model:
        logger.warning('Could not get model from platform string: '
                       '%s', pf_string)

    poer_handler = PortConf(evt)
    # Always start with a platform specific default portmap
    # portmap dictionary keys represent the port labels in ports.conf
    # portmap dictionary values represent the ports mode in ports.conf
    portmap = poer_handler.get_platform_default_portmap()
    if portmap is None:
        # logger.warning("Internal Error: Cannot create platform portmap")
        # evt.report_issue("Internal Error: Cannot create platform portmap")
        return  # pragma: no cover TODO CUE-3046

    # Populate configured iface_speeds and iface_breakouts from
    # the Object Model
    iface_speeds = {}
    iface_breakouts = {}
    ifaces = _get_ifaces(evt)
    for iface_name, iface in sorted(ifaces.items()):
        m = re.match(swp_pattern, iface_name)
        if m:
            # Use the port as the key for speeds and breakouts to make
            # it easier to index via portmap keys later

            # 3p4
            iface_name = m.group('suffix').strip('p')
        try:
            if iface["link"]["speed"] != "auto":
                # {'3p4': '10G'}
                iface_speeds[iface_name] = iface["link"]["speed"]
        except Exception:
            pass
        try:
            # {'3p4': '4x25G'}
            iface_breakouts[iface_name] = iface["link"]["breakout"]
        except Exception:
            pass

    # Dont allow breakout config for ports that are not configurable
    for iface in sorted(ifaces.keys()):
        m = re.match(swp_pattern, iface)
        if m:
            # Use the port as the key for speeds and breakouts to make
            # it easier to index via portmap keys later
            iface = m.group('suffix').strip('p')
        if iface not in portmap and iface_breakouts.get(iface, "1x") != "1x":
            msg = "Breakout config is not valid for port {0}".format(iface)
            logger.warning(msg)
            evt.report_issue(msg)

    if (
            model and model in
            poer_handler.get_platform_port_info().get('4x_disable')
       ):
        requires_disable = True
    else:
        requires_disable = False

    # Override the default port settings with configured speeds and/or
    # breakouts
    warnings = []
    for port in portmap.keys():
        # Check if platform is one of the mellanox platforms that requires
        # subsequent EVEN ports to be disabled

        iface_breakout = iface_breakouts.get(port)
        iface_speed = iface_speeds.get(port, portmap.get(port))
        if iface_breakout and iface_breakout != '1x':
            portmap[port] = iface_breakout

            if '4x' in iface_breakout and requires_disable:
                # Check subsequent port
                valid = _verify_4x_breakout(evt, portmap, port,
                                            iface_breakout, model, warnings)
                if not valid:
                    return False

        elif iface_speed:
            portmap[port] = poer_handler.get_port_bandwidth(port, iface_speed)

    # Create a temporary ports.conf file from portmap so it can be validated
    """ Write a temporary ports.conf and return the filename """
    ts = time.strftime("%Y%m%d-%H%M%S")
    temp_ports_conf = "/tmp/ports.conf_{0}".format(ts)
    data = ""
    for port, mode in portmap.items():
        data += "{0}={1}\n".format(port, mode)
    try:
        with evt.fileops.open_file(temp_ports_conf, 'w') as ports_conf:
            ports_conf.write(("# See {}\n# for additional details about this "
                             "file.").format(
                                 poer_handler.get_default_ports_conf_fn()))
            ports_conf.write(data)
    except Exception:  # pragma: no cover TODO CUE-3046
        msg = "Internal Error: Unable to create {0}".format(temp_ports_conf)  # pragma: no cover TODO CUE-3046 # noqa
        logger.error(msg)  # pragma: no cover TODO CUE-3046
        evt.report_issue(msg)  # pragma: no cover TODO CUE-3046
        return False  # pragma: no cover TODO CUE-3046

    # Use existing validate-ports script to perform final validation of
    # potential ports.conf file.
    errors = _validate_ports(evt, poer_handler.get_port_modes(),
                             temp_ports_conf)

    for error, _ in errors:
        evt.report_issue(error)

    # Carry the portmap forward, so we don't have to repeat the portmap logic
    # in handle_ready_apply()
    evt.carry_forward["portmap"] = portmap
    evt.carry_forward["verify_config_warnings"] = warnings
    try:
        evt.fileops.remove_file("{0}".format(temp_ports_conf))
    except Exception:  # pragma: no cover TODO CUE-3046
        pass  # pragma: no cover TODO CUE-3046
    if errors:
        return False
    else:
        return True


def handle_ready_apply(evt):
    """
    Prepare to apply the config.  Generate config files and scripts.  Also,
    look for issues in the config that would prevent us from applying it.

    Use evt.config_v1 to access the pending configuration's state.
    Use evt.render to render a template with the given variables.
    Use evt.stage_file to prepare a rendered file to be copied over to its
        ultimate destination.

    Args:
        evt: A ReadyApplyEvent instance.
    """
    poer_handler = PortConf(evt)
    warnings = evt.carry_forward["verify_config_warnings"]
    portmap = evt.carry_forward["portmap"]

    for warning in warnings:
        evt.report_warning(warning)

    # Generate the final config file
    header = ("# See {}\n# for additional details about this "
              "file.").format(poer_handler.get_default_ports_conf_fn())
    evt.render(templates, "ports.conf",
               header=header, ports_dictionary=portmap)

    staged = evt.stage_file("ports.conf", "/etc/cumulus/ports.conf")

    # No need to reload config file if there are no configurable ports
    if not portmap:
        return  # pragma: no cover TODO CUE-3046

    operational_portmap = poer_handler.get_platform_operational_portmap()
    asic_vendor = py_.get(evt.platform_v1.getHardware(), 'asic-vendor')
    if staged and operational_portmap != portmap:
        if asic_vendor == 'Broadcom':
            evt.schedule_restart('switchd')
        else:
            evt.schedule_reload('switchd')


def handle_check_health(evt):
    """
    Handle a check_health event.

    Use evt.report_error to report any found errors.
    Use evt.report_warning to report any found warnings.

    Warnings *should* clear on their own, given some time, usually indicating
    that a component is still "coming up."  Errors *will not* clear on their
    own.

    Args:
        evt: A CheckHealthEvent instance.
    """


def _verify_4x_breakout(evt, portmap, port, mode, model, warnings):
    """
    Verify that the 4x breakout configuration is valid on Mellanox
    platforms that require the subsequent EVEN numbered port to be disabled.

    First it checks that the port is not an even port.  If it is, then
    configuration is invalid so it returns False.

    Then it gets the interface and link.breakout configuration for
    the subsequent interface...

    There are 3 cases to handle:
        case 1)
            link.breakout is disabled.  Theres nothing to do.  Return True.
        case 2)
            The interface config is None. In this case we can disable the
            subsequent port automatically but we still
            need to warn the user that this is happening, and tell them how
            they can stop the warnings. Return True.
        case 3)
            Neither of the above is true, meaning there is explicit
            configuration on the interface and/or a breakout mode is
            configured other than `disabled`.
            In either case, report an issue to tell the user what the problem.
            is.  Return False.
    """
    prefix = None
    if 'p' in port:
        prefix, port_num = port.split('p')
    else:
        port_num = port

    port_num = int(port_num)

    # Check if port is even, in which case we know its an invalid config
    if port_num % 2 == 0:
        evt.report_issue((f'{model} does not support configuring 4x breakout'
                          f' modes on even numbered ports. '
                          f'interface.{get_ifname(port)}.link.breakout is set'
                          f' to {mode}.')
                         )
        return False

    # Increment port number to get subsequent port
    port_num = port_num + 1

    # Add the module prefix back to port num
    if prefix:
        next_port = prefix + 'p' + str(port_num)
    else:
        next_port = str(port_num)

    # If subsequent port is not in the portmap its not a valid port
    # This should never be the case unless the port is even though.
    if next_port not in portmap:
        logger.error('%s is not in the portmap: %s', next_port, str(portmap))
        evt.report_issue((f"Internal server error.  {next_port} is not in the"
                          "portmap"))
        return False

    # Turn port number in to an interface name
    next_ifname = get_ifname(next_port)

    # Get interface.<subsequent>.link.breakout from the subsequent interface.
    # If the interface has config then the link.breakout is either configured
    # manually or it defaults to 1x, therefore if this is None or disabled then
    # its not an error.
    interfaces = evt.config_v1.getInterfaces()
    config_breakout = py_.get(interfaces,
                              f'{next_ifname}.link.breakout')

    if config_breakout == 'disabled':
        return True
    elif next_ifname not in interfaces:
        warnings.append((f'interface.{get_ifname(port)}.link.breakout is '
                         f'{mode}. Platform {model} requires the subsequent '
                         f' port '
                         f'to be disabled when a 4x breakout mode is '
                         f'configured.\nAutomatically disabling '
                         f'{next_ifname}. To avoid this warning, disable '
                         f'{next_ifname} via:\n'
                         f' nv set interface {next_ifname} link breakout '
                         f'disabled'))
        portmap[next_port] = 'disabled'
        return True
    else:
        evt.report_issue((f'interface.{get_ifname(port)}.link.breakout is '
                          f'{mode}. {model} requires the subsequent port '
                          f'to be disabled when a 4x breakout mode is '
                          f'configured.  Disable {next_ifname} via:\n'
                          f' nv set interface {next_ifname} link breakout '
                          f'disabled'))
        return False
